struct System{

    /*--------------------------------------------------------------------------------------*/
    /*  System is of size LxL, giving LxL nodes. The first L nodes (row 1) are part of the  */
    /*  lower electrode, while the last L nodes (row L) are the nodes situated below the    */
    /*  upper electrode. Each node is connected to four neighbours by bonds, and are seen   */
    /*  as having two bonds going out (upwards) and two coming in (from below).             */
    /*  max and min_voltage is the voltage at the top and bottom electrodes respectively    */
    /*  b is the slope of the current in a bond after the threshold has been exceeded,      */
    /*  that is b = tan(theta)                                                              */
    /*                                                                                      */
    /*  threshold stores the thresholds of all the bonds in a (L*L)x2 array, with vertical  */
    /*  bonds going out of a node stored in the first column (0) and outgoing diagonal      */
    /*	bonds are stored in the second column (1).                                          */
    /*	conductance stores the conductance of                                               */
    /*  the bonds in the same way as threshold.                                             */
    /*  voltage stores the voltage of all nodes in an array of size (L*L).                  */
    /*  updated stores bonds that have passed the threshold as 1 and all others as 0.       */
    /*  perm_updated stores bonds that have broken permanently as 1 and all others as 0.    */
    /*  pathscape stores the lowest total threshold of spanning paths for each node.        */
    /*  pathscape_bonds stores the bonds connecting a node to the pathscape (or something... */
    /*--------------------------------------------------------------------------------------*/    

    int L;
    int nthreads; // Number of threads, for OpenMP
    int num_updated_bonds;
    double max_voltage;
    double min_voltage;
    double a, b;
    double total_current;
    double **threshold;
    double **conductance;
    double **max_dV;
    double **max_i;
    int **updated;
    int **perm_updated;
    double *voltage; 
    double *pathscape;
    int **pathscape_bonds;    
};



double walltime (void); //Get system time

void allocSys(struct System *sys, int size);
void initSys(struct System *sys, double theta, int seed, double beta, int threads);
void freeSys(struct System *sys);

void updateDV(struct System *sys);
int updateConductance(struct System *sys);
int updateConductanceSingleElasticPhi(struct System *sys, double phi);
int updateConductancePlastic23(struct System *sys);

int CG(struct System *sys);
void breakBonds(struct System *sys);

double HKBonds(struct System *sys);
int HKnpb(int **updated, int *nodes, int L);
int HKV(int **updated, int L);
int HK(int **updated, int *nodes, int L);
int **getFracture(int **updated, int L);
int **getBackbone(int **updated, int L);
int burntBonds(struct System *sys);
void removeArms(double **conductance, int L);
double getWidth(int **bonds, int L);
int **shortestPath(double **threshold, int L);
int **getShortestPath(int *predecessor, int min_i, int min_j, int L);
int **shortestPathOLD(double **threshold, int L);
int **getShortestPathOLD(int *predecessor, int min_i, int min_j, int L);

void compare(int max_runs, double b_min, double b_max, double b_step, double dV, double beta, char *fname);
void checkCurrent(double b_min, double b_max, double b_step, double dV);
void printToPS(struct System *sys, char* filename);
void checkPrint();
void checkPrintTrapping();
void pathscapeTest(int size, int runs);
void fwritePathscape(struct System *sys, int run);
void fwriteBackbone(int size, int run, double theta, int **backbone, int **sc, int **burn_class, double phi);
void breakFuseNetwork(struct System *sys, int run, double theta);
void breakFuseFast(struct System *sys);
double vtRatio(struct System *sys);
double vtRatioAvg(struct System *sys);
