
#pragma OPENCL EXTENSION cl_khr_global_int32_base_atomics : enable
#pragma OPENCL EXTENSION cl_khr_local_int32_base_atomics : enable

// __kernel void calculateFracturescapeCL(__global double/**/ *nodes1, __global double/**/ *nodes2, 
// 				       const __constant double/**/ *vthresholds, 
// 				       const __constant double/**/ *dthresholds,
// 				       __global int *bonds, int size, int i_row, int i_col, ushort rw,
// 				       __global double/**/ *gsum, __local double/**/ *lsum){

typedef struct Params{    
    int size;
    int i_row;
    int i_col;
    int col;
    int row;
}Params;


static void
global_sync(volatile __global int *flags, __global int *counter)
{
    const size_t thread_id = get_local_id(0);
    const size_t workgroup_id = get_group_id(0);
 
    if (thread_id == 0) {
        flags[workgroup_id] = 1;
    }
 
    if (workgroup_id == 0) {
        if (thread_id < get_num_groups(0)) {
            while (flags[thread_id] != 1) ;
        }
        barrier(CLK_GLOBAL_MEM_FENCE);
 
        if (thread_id < get_num_groups(0)) {
            flags[thread_id] = 0;
        }
    }
 
    if (thread_id == 0) {
        while (flags[workgroup_id] != 0) ;
    }
    barrier(CLK_GLOBAL_MEM_FENCE);


}




__kernel void calculateFracturescapeCL(__global double/**/ *nodes1, __global double/**/ *nodes2, 
				       __global int *bonds1, __global int *bonds2,
				       __global double/**/ *vthresholds, 
				       __global double/**/ *dthresholds,
				       __local double/**/ *lsum,
				       __constant struct Params *par, __local int *rw){

    // v/dthresholds should possibly go to global memory. Will not have much use for 
    // broadcast, and might not have simple direct-addressing patterns. If that's
    // not the case, there is no need to use the constant space, which might not
    // have room for it. Params, on the other hand, can make use of the broadcast
    // in the AMD hardware.
    
    // System of lxl, i.e. (l+1)xl nodes. Make sure that the L and N in the kernel
    // is l+3 and 2*l+2

    // lsum is of size # work units per work group

    // NOTE: Currently, this only works if the number of work items per WG = 2*L.
    //       Due to this, only system sizes up to L = 128 are supported, with L = 2^n.

    // gid == row-1, lid == column-1
    int l = par->size;
    //int L = l+3;
    int N = 2*l+2;
    int gid = get_group_id(0);
    int lid = get_local_id(0);
    int lsize = get_local_size(0);
    int num_wg = get_num_groups(0);

    int stride = 2*l*(l+1)/lsize;
    int row = lid/(2*l/stride) + 1;
    int scol = stride * (lid % (2*l/stride)) + 1;
    int col = scol;
    col = lid+1;
    scol = lid+1;
    row = 1;
    int n = row*N+col;
    int i_node = (par->i_row+1)*N+par->i_col+1; // Shift up and right due to "walls"
    //int i = par->col;
    //i_node = par->col ? i_node+l : i_node;
    int first_col = gid < num_wg/2;
    i_node = first_col ? i_node : i_node+l;
    int loc_counter = 0;

    double/**/s1, s2, s3, s4, m, local_sum, sum, old_sum;
    short t;
    
    old_sum = 0;
    sum = 0;
    if(lid == 0){
    	*rw = 1;
    }

    barrier(CLK_LOCAL_MEM_FENCE);
    //int j;
    int run;
    while(*rw){
	if(first_col){
	    local_sum = 0;
	    //j = 0;
	    run = 1;
	    row = 1;
	    //row = row_start;
	    while(row < l+2){

		//for(int j=0; j<stride; j++){
		//col = scol+j;
		n = row*N+col;

		// if not i_node
		s1 = nodes1[ row*N+col-N-1 ] + vthresholds[ row*N+col-N ];
		s2 = nodes1[ row*N+col-N   ] + dthresholds[ row*N+col-N ];
		s3 = nodes1[ row*N+col+N   ] + dthresholds[ row*N+col   ];
		s4 = nodes1[ row*N+col+N+1 ] + vthresholds[ row*N+col+1 ];

		t = (s1 < s2) ? 1 : 2;
		m = (s1 < s2) ? s1 : s2;
		t = (s3 < m) ? 3 : t;
		m = (s3 < m) ? s3 : m;
		t = (s4 < m) ? 4 : t;
		m = (s4 < m) ? s4 : m;
		barrier(CLK_GLOBAL_MEM_FENCE);
		nodes1[ n ] = m * (n != i_node);
		bonds1[ n ] = (nodes1[ n ] > 0) * t;
		local_sum += nodes1[ n ];
		barrier(CLK_GLOBAL_MEM_FENCE);
		//j++;
		row++;
		//run = (j < l+1);

	    }
	    lsum[lid] = local_sum;
	    //barrier(CLK_LOCAL_MEM_FENCE);
	    if(lid == 0){
		sum = 0;
		for(int k=0; k<lsize; k++)
		    sum += lsum[k];
		loc_counter += !(fabs(sum-old_sum) > 1e-10);
		loc_counter *= !(fabs(sum-old_sum) > 1e-10);
		*rw = (loc_counter < 10);
		//*rw = (fabs(sum-old_sum) > 1e-10);
		old_sum = sum;
		// if(fabs(sum-old_sum) < 1e-6){
		//     *rw = 0;
		// }
		// else
		//     old_sum = sum;
		
	    }
	    barrier(CLK_LOCAL_MEM_FENCE);
	}
	else{
	    local_sum = 0;
	    //row = row_start;
	    //while(row < row_stop){//l+2){
	    for(int j=1; j<l+2; j++){
		//col = scol+j;
		row = j;
		n = row*N+col;
		// if not i_node
		s1 = nodes2[ row*N+col-N-1 ] + vthresholds[ row*N+col-N ];
		s2 = nodes2[ row*N+col-N   ] + dthresholds[ row*N+col-N ];
		s3 = nodes2[ row*N+col+N   ] + dthresholds[ row*N+col   ];
		s4 = nodes2[ row*N+col+N+1 ] + vthresholds[ row*N+col+1 ];

		t = (s1 < s2) ? 1 : 2;
		m = (s1 < s2) ? s1 : s2;
		t = (s3 < m) ? 3 : t;
		m = (s3 < m) ? s3 : m;
		t = (s4 < m) ? 4 : t;
		m = (s4 < m) ? s4 : m;
		barrier(CLK_GLOBAL_MEM_FENCE);
		nodes2[ n ] = m * (n != i_node);
		bonds2[ n ] = (nodes2[ n ] > 0) * t;
		local_sum += nodes2[ n ];
		barrier(CLK_GLOBAL_MEM_FENCE);
		//row++;
	    }
	    lsum[lid] = local_sum;
	    barrier(CLK_LOCAL_MEM_FENCE);
	    if(lid == 0){
		sum = 0;
		for(int k=0; k<lsize; k++)
		    sum += lsum[k];
		loc_counter += !(fabs(sum-old_sum) > 1e-10);
		loc_counter *= !(fabs(sum-old_sum) > 1e-10);
		*rw = (loc_counter < 10);
		//*rw = (fabs(sum-old_sum) > 1e-10);
		old_sum = sum;

		// if(fabs(sum-old_sum) < 1e-10)
		//     *rw = 0;
		// else
		//     old_sum = sum;
		
	    }
	    barrier(CLK_LOCAL_MEM_FENCE);
	}
    }
    
}
