#define _XOPEN_SOURCE  // for popen(), pclose()

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/time.h>
#include "network.h"
#include "statistics.h"
#include "testfuncs.h"
#include <gsl/gsl_rng.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_randist.h>
#include <math.h>
#include <limits.h>
#include <float.h>
#include <unistd.h>
#include <string.h>
#include <sys/dir.h>
#include <errno.h>
#include <sys/stat.h>

#ifdef _OPENMP
#  include <omp.h>
#endif

//#include <omp.h>

#define SIZE 16  // Default size for memory allocation. Should generally not be used.


/*  min and max for integers. For doubles/floats, use fmin/fminf  */
#define min(a,b) ((a < b) ? a : b)
#define max(a,b) ((a > b) ? a : b)
/* #define min(a,b) ({ __typeof__ (a) _a = (a); __typeof__ (b) _b = (b); _a < _b ? _a : _b; }) */
/* #define max(a,b) ({ __typeof__ (a) _a = (a); __typeof__ (b) _b = (b); _a > _b ? _a : _b; }) */


double walltime(void){

    /************************************************************/
    /*  Returns the time in seconds with microsecond precision  */
    /************************************************************/

    static struct timeval t;
    gettimeofday ( &t, NULL );
    return ( t.tv_sec + 1e-6 * t.tv_usec );
}



void allocSys(struct System *sys, int size){

    // TODO: Update comment, it's using deprecated arrays
    /*---------------------------------------------------------------------------*/
    /*  Allocate memory for matrices and vectors in the struct sys               */
    /*  N = LxL inner nodes, total of 2N*2 thresholds                            */
    /*  Conductance matrix is of size NxN, stored as Nx4 as each row has four    */
    /*  non-zero entries, except for the N first and last rows, which have two.  */
    /*  conductance stores for bonds between inner nodes. outer_conductance      */
    /*  stores for bonds between inner nodes (first and last row) and bus-bars.  */
    /*  NOTE: Have a total of 2*L*L bonds. As num_updated_bonds is an int, it    */
    /*        can have a maximum value of 32767. Consider changing to long int   */
    /*        or unsigned if runs are done on arrays larger than L=128.          */
    /*---------------------------------------------------------------------------*/


    int L, N;

    assert(size > 1);
    L = size;
    N = L*L;

    sys->L = L;
    sys->num_updated_bonds = 0;
    sys->threshold = calloc(N, sizeof(sys->threshold));
    sys->conductance = calloc(N, sizeof(sys->conductance));
    sys->max_dV = calloc(N, sizeof(sys->max_dV));
    sys->max_i = calloc(N, sizeof(sys->max_i));
    sys->updated = calloc(N, sizeof(sys->updated));
    sys->perm_updated = calloc(N, sizeof(sys->perm_updated));
    sys->voltage = calloc(N, sizeof(sys->voltage));
    sys->pathscape = calloc(L*(L+1), sizeof(sys->pathscape)); 
    sys->pathscape_bonds = calloc(L*(L+1), sizeof(sys->pathscape_bonds)); 
    for(int i=0; i<N; i++){
	sys->threshold[i] = calloc(2, sizeof(sys->threshold[i]));
	sys->conductance[i] = calloc(2, sizeof(sys->conductance[i]));
	sys->max_dV[i] = calloc(2, sizeof(sys->max_dV[i]));
	sys->max_i[i] = calloc(2, sizeof(sys->max_i[i]));
	sys->updated[i] = calloc(2, sizeof(sys->updated[i]));
	sys->perm_updated[i] = calloc(2, sizeof(sys->perm_updated[i]));
    }
    for(int i=0; i<L*(L+1); i++)
	sys->pathscape_bonds[i] = calloc(2, sizeof(sys->pathscape_bonds[i])); 
}
void initSys(struct System *sys, double theta, int seed, double beta, int threads){

    /*  Initiate system with threshold, guess initial voltages, fill matrix/vectors  */
    /*  Conductance = unity, conductance matrix is D-matrix                          */
    /*  b is given as theta in degrees                                               */
    /*  beta decides the probability distribution used for the thresholds.           */
    /*  beta = 0: uniform dist. beta > 0: hyperbolic dist., beta denotes limits      */

    int L = sys->L;
    int N = L*L;    
    double voltage_step, guess, slope;
    gsl_rng *r = gsl_rng_alloc(gsl_rng_mt19937);
    gsl_rng_set(r, seed);

    slope = tan(theta*M_PI/180);

    sys->max_voltage = 1.0;
    sys->min_voltage = 0.0;
    sys->total_current = sys->max_voltage-sys->min_voltage;
    sys->a = 1;
    sys->b = slope;
    sys->num_updated_bonds = 0;
    sys->nthreads = threads > 0 ? threads : 1;
    /*  Initialize thresholds  */

    //double mu = 1/1.5;
    // Beta decides disorder. beta = 0.01 => weak disorder,
    // beta = 1 => intermediate, beta = 100 => strong disorder
    // beta > 2 is probably not a good idea here, as the thresholds
    // get up to 100 in that case => needs to run a long time to get large enough
    // voltage drops to break the network
    // From "Fracturing highly disordered materials" by Oliveira et. al. (July 2012)

    /* if(beta > 0){    // hyperbolic distribution */
    /* 	double R; */
    /* 	for(int j=0; j<2; j++) */
    /* 	    for(int i=0; i<N; i++){ */
    /* 		R = gsl_rng_uniform(r);   */
    /* 		R = (gsl_rng_uniform(r) > 0.5) ? R : -R; // random number between -1 and 1 */
    /* 		sys->threshold[i][j] = pow(10, beta*R); */
    /* 	    } */
    /* } */

    /* else    // uniform distribution */
	for(int j=0; j<2; j++)
	    for(int i=0; i<N; i++)
		sys->threshold[i][j] = gsl_rng_uniform(r);
    
    /*  L inner nodes, L+1 voltage differences  between bus bars  */

    voltage_step = (sys->max_voltage-sys->min_voltage)/L;
    guess = sys->min_voltage;
    for(int i=0; i<L; i++){
	for(int j=0; j<L; j++){
	    sys->voltage[i*L + j] = guess;
	}
	guess += voltage_step;
    }


    /*---------------------------------------------------------------------------*/
    /*  Conductance[i][1] contains the conductance of the bond between node i    */
    /*  and node i+L (above). conductance[i][2] is the conductance of the bond   */
    /*  between	node i and node i+L+1 (above and to the right).                  */
    /*---------------------------------------------------------------------------*/

    for(int i=0; i<N; i++){
	sys->conductance[i][0] = 1;
	sys->conductance[i][1] = 1;
	sys->max_dV[i][0] = voltage_step;
	sys->max_dV[i][1] = voltage_step;
	sys->max_i[i][0] = 0;
	sys->max_i[i][1] = 0;
	sys->updated[i][0] = 0;
	sys->updated[i][1] = 0;
	sys->perm_updated[i][0] = 0;
	sys->perm_updated[i][1] = 0;
    }
    for(int i=0; i<L*(L+1); i++)
	sys->pathscape[i] = DBL_MAX;

    gsl_rng_free(r);


}



void freeSys(struct System *sys){
    for(int i=0; i<sys->L*(sys->L+1); i++)
	free(sys->pathscape_bonds[i]);
    for(int i=0; i<sys->L*sys->L; i++){
	free(sys->threshold[i]);
	free(sys->conductance[i]);
	free(sys->max_dV[i]);
	free(sys->max_i[i]);
	free(sys->updated[i]);
	free(sys->perm_updated[i]);
    }
    free(sys->threshold);
    free(sys->conductance);
    free(sys->max_dV);
    free(sys->max_i);
    free(sys->updated);
    free(sys->perm_updated);
    free(sys->voltage);
    free(sys->pathscape);
    free(sys->pathscape_bonds);
}



int updateConductance(struct System *sys){
 
    int L = sys->L;
    int updated = 0;
    double dV, t, tempc;
    int **old_updated = calloc(L*L, sizeof(old_updated));
    int *ir = calloc(L, sizeof(ir));
    for(int i=0; i<L; i++)
	ir[i] = i+1;
    ir[L-1] = 0;

    sys->num_updated_bonds = 0;
    for(int i=0; i<L*L; i++){
    	old_updated[i] = calloc(2, sizeof(old_updated[i]));
    	old_updated[i][0] = sys->updated[i][0];
    	old_updated[i][1] = sys->updated[i][1];
    }
    
    /*  Checks if the current flowing through the bonds exceed the threshold.  */
    /*  Absolute value of difference in voltage makes it independent of current direction.  */
    /*  Make sure the upper row with conductance bonds is checked for the voltage difference  */
    /*  between the maximum voltage and the node voltage!  */
    /*  For negative slopes after threshold (conductance < t/dV), the conductance is >= 0.  */
    /*  This is ensured by setting the conductance to the maximum of 0 and (1-b)t/dV + b  */
    /*  Plastic case: If t/dV gets larger (and it has been updated), the old conductance  */
    /*	is kept. If it gets smaller (dV increases), the conductance is updated.  */
    /*  Elastic case: Conductance is updated no matter what t/dV is. If t/dV > 1,  */
    /*  the conductance is set to 1. Otherwise, it is set to (1-b)t/dV + b.  */

    for(int j=0; j<(L-1); j++)
	for(int i=0; i<L; i++){

	    dV = fabs(sys->voltage[i+(j+1)*L] - sys->voltage[i+j*L]);
	    t = sys->threshold[i+j*L][0];

	    if(sys->perm_updated[i+j*L][0]){  
		// Plastic case. The old conductance is kept if the new
		// dV is smaller than the old.
		// For increasing dV, the conductance increase/decrease
		// for b larger/smaller than 1.
		// Thus, the conductance is compared to conclude whether
		// dV has increased or not.

		tempc = ((1-sys->b)*t/dV + sys->b) > 0 ? (1-sys->b)*t/dV + sys->b : 0;
		if(sys->b < 1 && tempc < sys->conductance[i+j*L][0]){
		    sys->conductance[i+j*L][0] = tempc;
		    sys->updated[i+j*L][0] = 1;
		    sys->num_updated_bonds++;
		}
		else if(sys->b > 1 && tempc > sys->conductance[i+j*L][0]){
		    sys->conductance[i+j*L][0] = tempc;
		    sys->updated[i+j*L][0] = 1;
		    sys->num_updated_bonds++;
		}		
	    }
	    else if( t < dV ){ 
		// No permanent update, but the bond has reached the threshold.
		// This could be the case for both plastic and elastic scenarios

		sys->conductance[i+j*L][0] = ((1-sys->b)*t/dV + sys->b) > 0 ? (1-sys->b)*t/dV + sys->b : 0;
		sys->updated[i+j*L][0] = 1;
		sys->num_updated_bonds++;
	    }
	    else if(!sys->perm_updated[i+j*L][0]){
		// Elastic case and the case that the bond is not
		// permanently broken
		sys->conductance[i+j*L][0] = 1;
		sys->updated[i+j*L][0] = 0;
	    }

	    dV = fabs(sys->voltage[ir[i] + (j+1)*L] - sys->voltage[i+j*L]);
	    t = sys->threshold[i+j*L][1];

	    if(sys->perm_updated[i+j*L][1]){  

		tempc = ((1-sys->b)*t/dV + sys->b) > 0 ? (1-sys->b)*t/dV + sys->b : 0;
		if(sys->b < 1 && tempc < sys->conductance[i+j*L][1]){
		    sys->conductance[i+j*L][1] = tempc;
		    sys->updated[i+j*L][1] = 1;
		    sys->num_updated_bonds++;
		}
		else if(sys->b > 1 && tempc > sys->conductance[i+j*L][1]){
		    sys->conductance[i+j*L][1] = tempc;
		    sys->updated[i+j*L][1] = 1;
		    sys->num_updated_bonds++;
		}		
	    }
	    else if( t < dV){

		sys->conductance[i+j*L][1] = ((1-sys->b)*t/dV + sys->b) > 0 ? (1-sys->b)*t/dV + sys->b : 0;
		sys->updated[i+j*L][1] = 1;
		sys->num_updated_bonds++;
	    } 
	    else if(!sys->perm_updated[i+j*L][1]) { 
		sys->updated[i+j*L][1] = 0;
		sys->conductance[i+j*L][1] = 1;
	    }
	}


    /*  Checks the upper electrode  */
    for(int i=0; i<L; i++){

	t = sys->threshold[i+(L-1)*L][0];
	dV = fabs(sys->max_voltage - sys->voltage[i+(L-1)*L]);
	if(sys->perm_updated[i+(L-1)*L][0]){  

	    tempc = ((1-sys->b)*t/dV + sys->b) > 0 ? (1-sys->b)*t/dV + sys->b : 0;
	    if(sys->b < 1 && tempc < sys->conductance[i+(L-1)*L][0]){
		sys->conductance[i+(L-1)*L][0] = tempc;
		sys->updated[i+(L-1)*L][0] = 1;
		sys->num_updated_bonds++;
	    }
	    else if(sys->b > 1 && tempc > sys->conductance[i+(L-1)*L][0]){
		sys->conductance[i+(L-1)*L][0] = tempc;
		sys->updated[i+(L-1)*L][0] = 1;
		sys->num_updated_bonds++;
	    }		
	}
	
	else if( t < dV){

	    sys->conductance[i+(L-1)*L][0]  = ((1-sys->b)*t/dV + sys->b) > 0 ? (1-sys->b)*t/dV + sys->b : 0;
	    sys->updated[i+(L-1)*L][0] = 1;
	    sys->num_updated_bonds++;
	} 
	else if (!sys->perm_updated[i+(L-1)*L][0]){
	    sys->conductance[i+(L-1)*L][0] = 1;
	    sys->updated[i+(L-1)*L][0] = 0;
	}

	t = sys->threshold[i+(L-1)*L][1];
	if(sys->perm_updated[i+(L-1)*L][1]){  
	    
	    tempc = ((1-sys->b)*t/dV + sys->b) > 0 ? (1-sys->b)*t/dV + sys->b : 0;
	    if(sys->b < 1 && tempc < sys->conductance[i+(L-1)*L][1]){
		sys->conductance[i+(L-1)*L][1] = tempc;
		sys->updated[i+(L-1)*L][1] = 1;
		sys->num_updated_bonds++;
	    }
	    else if(sys->b > 1 && tempc > sys->conductance[i+(L-1)*L][1]){
		sys->conductance[i+(L-1)*L][1] = tempc;
		sys->updated[i+(L-1)*L][1] = 1;
		sys->num_updated_bonds++;
	    }		
	}
	else if( t < dV ){

	    sys->conductance[i+(L-1)*L][1] = ((1-sys->b)*t/dV + sys->b) > 0 ? (1-sys->b)*t/dV + sys->b : 0;
	    sys->updated[i+(L-1)*L][1] = 1;
	    sys->num_updated_bonds++;
	} 
	else if (!sys->perm_updated[i+(L-1)*L][1]){
	    sys->conductance[i+(L-1)*L][1] = 1;
	    sys->updated[i+(L-1)*L][1] = 0;
	}
    }

    for(int i=0; i<L*L; i++)
    	if( (sys->updated[i][0] != old_updated[i][0]) ||
    	    (sys->updated[i][1] != old_updated[i][1]) ){
    	    updated = 1;
	    //printf("updated %d\n", i);
	    break;
	}

    for(int i=0; i<L*L; i++)
    	free(old_updated[i]);
    free(old_updated);
    free(ir);

    return updated;
    
}



int updateConductanceElastic(struct System *sys){
 
    int L = sys->L;
    int updated = 0;
    double dV, t;
    int **old_updated = calloc(L*L, sizeof(old_updated));
    int *ir = calloc(L, sizeof(ir));
    for(int i=0; i<L; i++)
	ir[i] = i+1;
    ir[L-1] = 0;
    //int old_updated_num = sys->num_updated_bonds;
    sys->num_updated_bonds = 0;
    for(int i=0; i<L*L; i++){
    	old_updated[i] = calloc(2, sizeof(old_updated[i]));
    	old_updated[i][0] = sys->updated[i][0];
    	old_updated[i][1] = sys->updated[i][1];
    }
    

    int count = 0;
    
    double ltv = 1;
    double tdV = 0;
    int ltv_coord = 0;
    int ltv_bond = 0;
    int new_ltv = 0;
    double ldV = 0;


    for(int j=0; j<L; j++){
	for(int i=0; i<L; i++){
	    for(int k=0; k<2; k++){
		if(j < L-1 && k == 0)
		    dV = fabs(sys->voltage[i+(j+1)*L] - sys->voltage[i+j*L]);
		else if(j < L-1 && k == 1)
		    dV = fabs(sys->voltage[ir[i] + (j+1)*L] - sys->voltage[i+j*L]);
		else
		    dV = fabs(sys->max_voltage - sys->voltage[i+j*L]);

		t = sys->threshold[i+j*L][k];
		tdV = t/dV;
	    
		if(tdV < 1){
		    new_ltv = (tdV < ltv);
		    ltv_coord = (new_ltv) ? i+j*L : ltv_coord;
		    ltv_bond = (new_ltv) ? k : ltv_bond;
		    ltv = (new_ltv) ? tdV : ltv;
		    ldV = (new_ltv) ? dV : ldV;
		    count++;
		}
	    }
	}
    }
    

    if(count > 0){
	for(int j=0; j<L; j++){
	    for(int i=0; i<L; i++){
		for(int k=0; k<2; k++){

		    if(j < L-1 && k == 0)
			dV = fabs(sys->voltage[i+(j+1)*L] - sys->voltage[i+j*L]);
		    else if(j < L-1 && k == 1)
			dV = fabs(sys->voltage[ir[i] + (j+1)*L] - sys->voltage[i+j*L]);
		    else
			dV = fabs(sys->max_voltage - sys->voltage[i+j*L]);

		    t = sys->threshold[i+j*L][k];

		    
		    // Set conductance for bonds which have broken, that is, they
		    // have passed the threshold already. Thos that have not, get
		    // unity conductance

		    // Set for bonds which have passed the maximum threshold 
		    // (and dV > max_dV)
		    if(t < dV){
			sys->conductance[i+j*L][k] = ((1-sys->b)*t/dV + sys->b > 0) ? 
			                             (1-sys->b)*t/dV + sys->b : 0;
			sys->max_dV[i+j*L][k] = dV;
			sys->updated[i+j*L][k] = 1;
			sys->num_updated_bonds++;
		    }
		    // All bonds that have not passed the threshold before this
		    // update, gets unit conductance.
		    else{
			sys->conductance[i+j*L][k] = 1;
			//sys->max_dV[i+j*L][k] = dV;
			sys->updated[i+j*L][k] = 0;
		    }	       		    
		}
	    }
	}
    
	// Set conductance of the one bond to be broken
    	sys->conductance[ltv_coord][ltv_bond] = (((1-sys->b)*ltv + sys->b) > 0) ? ((1-sys->b)*ltv + sys->b) : 0;
	sys->max_dV[ltv_coord][ltv_bond] = ldV;
	assert(sys->threshold[ltv_coord][ltv_bond] < ldV);
    	sys->updated[ltv_coord][ltv_bond] = 1;
	sys->num_updated_bonds++;
	updated = 1;
    }

    else
	for(int j=0; j<L; j++){
	    for(int i=0; i<L; i++){
		for(int k=0; k<2; k++){
		    if(sys->threshold[i+j*L][k] < sys->max_dV[i+j*L][k])
			sys->num_updated_bonds++;
		}
	    }
	}

	
    count = 0;

    for(int i=0; i<L*L; i++)
    	free(old_updated[i]);
    free(old_updated);
    free(ir);

    return updated;
    
}




void updateDV(struct System *sys){

    double dV;
    int L = sys->L;
    int *ir = calloc(L, sizeof(ir));
    for(int i=0; i<L; i++)
	ir[i] = i+1;
    ir[L-1] = 0;

    for(int j=0; j<L; j++){
	for(int i=0; i<L; i++){
	    for(int k=0; k<2; k++){
		if(j < L-1 && k == 0)
		    dV = fabs(sys->voltage[i+(j+1)*L] - sys->voltage[i+j*L]);
		else if(j < L-1 && k == 1)
		    dV = fabs(sys->voltage[ir[i] + (j+1)*L] - sys->voltage[i+j*L]);
		else
		    dV = fabs(sys->max_voltage - sys->voltage[i+j*L]);


		sys->max_dV[i+j*L][k] = dV;
	    }
	}
    }

    free(ir);
}
	    
int updateConductanceSingleElastic(struct System *sys){
 

    //#pragma omp parallel default(none) shared(2) 
    
    int L = sys->L;
    int updated = 0;
    double dV, t;
    /* int **old_updated = calloc(L*L, sizeof(old_updated)); */
    int *ir = calloc(L, sizeof(ir));
    for(int i=0; i<L; i++)
	ir[i] = i+1;
    ir[L-1] = 0;
    //int old_updated_num = sys->num_updated_bonds;
    sys->num_updated_bonds = 0;
    /* for(int i=0; i<L*L; i++){ */
    /* 	old_updated[i] = calloc(2, sizeof(old_updated[i])); */
    /* 	old_updated[i][0] = sys->updated[i][0]; */
    /* 	old_updated[i][1] = sys->updated[i][1]; */
    /* } */
    

    int count = 0;
    double tempc;
    double tdV = 0;
    int lci_coord = 0;
    int lci_bond = 0;
    double old_i = 0;
    double new_i = 0;
    double lci = 0;  //largest change in i
    int new_lci = 0; 
    double new_c = 0;
    double new_dV = 0;


    for(int j=0; j<L; j++){
	for(int i=0; i<L; i++){
	    for(int k=0; k<2; k++){
		if(j < L-1 && k == 0)
		    dV = fabs(sys->voltage[i+(j+1)*L] - sys->voltage[i+j*L]);
		else if(j < L-1 && k == 1)
		    dV = fabs(sys->voltage[ir[i] + (j+1)*L] - sys->voltage[i+j*L]);
		else
		    dV = fabs(sys->max_voltage - sys->voltage[i+j*L]);

		t = sys->threshold[i+j*L][k];
		tdV = t/dV;
	    
		old_i = sys->conductance[i+j*L][k]*sys->max_dV[i+j*L][k];
		tempc = (((1-sys->b)*tdV + sys->b) > 0) ? ((1-sys->b)*tdV + sys->b) : 0;
		tempc *= !sys->perm_updated[i+j*L][k];  //sets tempc to zero if bond is broken
		tempc *= (dV > 1e-5);
		new_i = tempc*dV;

		// (fabs(sys->b-1) < 1e-3) in case theta = 45 => conductance and tempc are both 1
		new_lci = ( (fabs(new_i-old_i) > lci) 
			    && ( (fabs(sys->conductance[i+j*L][k]-tempc) > 1e-5)
				 || (fabs(sys->b-1) < 1e-3)) );
		new_lci *= ( dV > t || sys->updated[i+j*L][k] );
		new_lci *= !sys->perm_updated[i+j*L][k];
		
		if(new_lci){
		    //assert( !sys->perm_updated[i+j*L][k] );
		    //printf("   new conductance = %.4e\n", tempc);
		    lci = fabs(new_i-old_i);
		    lci_coord = i+j*L;
		    lci_bond = k;
		    new_c = tempc;
		    new_dV = dV;
		    count++;
		}

		if(sys->updated[i+j*L][k] && dV > t)
		    sys->conductance[i+j*L][k] = tempc;
		sys->num_updated_bonds += sys->updated[i+j*L][k];

	    }
	}
    }


    if(count > 0){
    
	// Set conductance of the one bond to be changed

	//if(new_dV < sys->threshold[lci_coord][lci_bond] && sys->updated[lci_coord][lci_bond])
	//    printf("Heavy reversing!\n");

    	sys->conductance[lci_coord][lci_bond] = new_c;
	sys->max_dV[lci_coord][lci_bond] = new_dV;
	sys->num_updated_bonds -= sys->updated[lci_coord][lci_bond];
    	sys->updated[lci_coord][lci_bond] = (new_dV > sys->threshold[lci_coord][lci_bond]);
	sys->num_updated_bonds += sys->updated[lci_coord][lci_bond];
	updated = 1;
	//printf("   new conductance = %.4e\n", new_c);
    }

    else{
	int nub = 0;
#pragma omp parallel for num_threads(sys->nthreads) reduction(+:nub)
	for(int j=0; j<L; j++){
	    for(int i=0; i<L; i++){
		for(int k=0; k<2; k++){
		    //sys->num_updated_bonds += sys->updated[i+j*L][k];
		    nub += sys->updated[i+j*L][k];
		}
	    }
	}
	sys->num_updated_bonds = nub;
    }


    /* for(int i=0; i<L*L; i++) */
    /* 	free(old_updated[i]); */
    /* free(old_updated); */
    free(ir);
    
    return updated;

}



int updateConductanceSingleElasticPhi(struct System *sys, double phi){
 
    
    //sys->a = tan(phi*M_PI/180);
    int L = sys->L;
    int updated = 0;
    double dV, t;
    int *ir = calloc(L, sizeof(ir));
    for(int i=0; i<L; i++)
	ir[i] = i+1;
    ir[L-1] = 0;
    //int old_updated_num = sys->num_updated_bonds;
    sys->num_updated_bonds = 0;
    /* for(int i=0; i<L*L; i++){ */
    /* 	old_updated[i] = calloc(2, sizeof(old_updated[i])); */
    /* 	old_updated[i][0] = sys->updated[i][0]; */
    /* 	old_updated[i][1] = sys->updated[i][1]; */
    /* } */
    

    int count = 0;
    double tempc;
    double tdV = 0;
    int lci_coord = 0;
    int lci_bond = 0;
    double old_i = 0;
    double new_i = 0;
    double lci = 0;  //largest change in i
    int new_lci = 0; 
    double new_c = 0;
    double new_dV = 0;


    for(int j=0; j<L; j++){
	for(int i=0; i<L; i++){
	    for(int k=0; k<2; k++){
		if(j < L-1 && k == 0)
		    dV = fabs(sys->voltage[i+(j+1)*L] - sys->voltage[i+j*L]);
		else if(j < L-1 && k == 1)
		    dV = fabs(sys->voltage[ir[i] + (j+1)*L] - sys->voltage[i+j*L]);
		else
		    dV = fabs(sys->max_voltage - sys->voltage[i+j*L]);

		t = sys->threshold[i+j*L][k];
		//t = 1;
		tdV = t/dV;
	    
		old_i = sys->conductance[i+j*L][k]*sys->max_dV[i+j*L][k];
		tempc = (((sys->a-sys->b)*tdV + sys->b) > 0) ? ((sys->a-sys->b)*tdV + sys->b) : 0;
		//tempc = (((sys->threshold[i+j*L][k]-sys->b)*tdV + sys->b) > 0) ? ((sys->threshold[i+j*L][k]-sys->b)*tdV + sys->b) : 0;
		tempc *= !sys->perm_updated[i+j*L][k];  //sets tempc to zero if bond is broken
		tempc *= (dV > 1e-5);
		new_i = tempc*dV;

		// (fabs(sys->b-1) < 1e-3) in case theta = 45 => conductance and tempc are both 1
		new_lci = ( (fabs(new_i-old_i) > lci) 
			    && ( (fabs(sys->conductance[i+j*L][k]-tempc) > 1e-5)
				 || (fabs(sys->b-1) < 1e-3)) );
		new_lci *= ( dV > t || sys->updated[i+j*L][k] );
		new_lci *= !sys->perm_updated[i+j*L][k];
		
		if(new_lci){
		    //assert( !sys->perm_updated[i+j*L][k] );
		    //printf("   new conductance = %.4e\n", tempc);
		    lci = fabs(new_i-old_i);
		    lci_coord = i+j*L;
		    lci_bond = k;
		    new_c = tempc;
		    new_dV = dV;
		    count++;
		}

		if(sys->updated[i+j*L][k] && dV > t)
		    sys->conductance[i+j*L][k] = tempc;
		sys->num_updated_bonds += sys->updated[i+j*L][k];

	    }
	}
    }


    if(count > 0){
    
	// Set conductance of the one bond to be changed

	//if(new_dV < sys->threshold[lci_coord][lci_bond] && sys->updated[lci_coord][lci_bond])
	//    printf("Heavy reversing!\n");

    	sys->conductance[lci_coord][lci_bond] = new_c;
	sys->max_dV[lci_coord][lci_bond] = new_dV;
	sys->num_updated_bonds -= sys->updated[lci_coord][lci_bond];
    	sys->updated[lci_coord][lci_bond] = (new_dV > sys->threshold[lci_coord][lci_bond]);
	sys->num_updated_bonds += sys->updated[lci_coord][lci_bond];
	updated = 1;
	//printf("   new conductance = %.4e\n", new_c);
    }

    else{
	int nub = 0;
	for(int j=0; j<L; j++){
	    for(int i=0; i<L; i++){
		for(int k=0; k<2; k++){
		    nub += sys->updated[i+j*L][k];
		}
	    }
	}
	sys->num_updated_bonds = nub;
    }


    free(ir);
    
    return updated;

}



int updateConductancePlastic23(struct System *sys){
 
    int L = sys->L;
    int updated = 0;
    double dV, t, temp;

    int *ir = (int*)calloc(L, sizeof(int));
    for(int i=0; i<L; i++)
	ir[i] = i+1;
    ir[L-1] = 0;

    sys->num_updated_bonds = 0;
    

    int count = 0;
    
    //lcc = largest change in conductance
    double u = 0;
    double max_dvu = 0;  // max(dV/u)
    int dvu_coord = 0;
    int dvu_bond = 0;
    int new_dvu = 0;

    for(int j=0; j<L; j++)
	for(int i=0; i<L; i++){
	    for(int k=0; k<2; k++){
		if(j < L-1 && k == 0)
		    dV = fabs(sys->voltage[i+(j+1)*L] - sys->voltage[i+j*L]);
		else if(j < L-1 && k == 1)
		    dV = fabs(sys->voltage[ir[i] + (j+1)*L] - sys->voltage[i+j*L]);
		else
		    dV = fabs(sys->max_voltage - sys->voltage[i+j*L]);

		t = sys->threshold[i+j*L][k];
		u = t-t/sys->b;
		//assert(u >= 0 || sys->b >= 0);

		if(dV > u && u >= t && !sys->perm_updated[i+j*L][k]){
		    temp = sys->b*sys->max_dV[i+j*L][k]/(t*(sys->b-1));  // dV/u
		    new_dvu = (temp > max_dvu);
		    max_dvu = (new_dvu) ? temp : max_dvu;		    
		}
		sys->max_dV[i+j*L][k] = dV;
		/* if(dV < sys->threshold[i+j*L][k] && sys->perm_updated[i+j*L][k]) */
		/*     printf("Heavy reversing on permanent updates! Bond (%d)[%d], max voltage = %.4f, t-dV = %.4e\n", */
		/* 	   i+j*L, k, sys->max_voltage, t-dV); */

		if(new_dvu){
		    dvu_coord = (new_dvu) ? i+j*L : dvu_coord;
		    dvu_bond = (new_dvu) ? k : dvu_bond;
		    count++;
		    //printf("  Max_voltage = %.4f, count = %d, dvu = %.6f, bond (%d)[%d], temp = %.4f\n",
		    //	   sys->max_voltage, count, max_dvu, i+j*L, k, temp);
		}
		new_dvu = 0;
		sys->num_updated_bonds += sys->updated[i+j*L][k];
	    }
	}
    
    if(count > 0){
	assert(sys->b < 0);
	assert(!sys->perm_updated[dvu_coord][dvu_bond]);
	sys->conductance[dvu_coord][dvu_bond] = 0;
	sys->perm_updated[dvu_coord][dvu_bond] = 1;
	sys->num_updated_bonds -= sys->updated[dvu_coord][dvu_bond];
	sys->updated[dvu_coord][dvu_bond] = 1;
	sys->num_updated_bonds += sys->updated[dvu_coord][dvu_bond];
	updated = 1;
    }
    free(ir);

    return updated;

}


int updateConductanceSuperconducting(struct System *sys){
 
    int L = sys->L;
    int updated = 0;
    double dV, t, temp, tdV;
    /* int **old_updated = (int**)calloc(L*L, sizeof(int*)); */
    int *ir = (int*)calloc(L, sizeof(int));
    for(int i=0; i<L; i++)
	ir[i] = i+1;
    ir[L-1] = 0;

    sys->num_updated_bonds = 0;
    /* for(int i=0; i<L*L; i++){ */
    /* 	old_updated[i] = (int*)calloc(2, sizeof(int)); */
    /* 	old_updated[i][0] = sys->updated[i][0]; */
    /* 	old_updated[i][1] = sys->updated[i][1]; */
    /* } */
    

    int count = 0;
    
    //lcc = largest change in conductance
    //double u = 0;
    double max_dvu = 0;  // max(dV/u)
    int dvu_coord = 0;
    int dvu_bond = 0;
    int new_dvu = 0;
    int max_i = 0;
    int nt = 50;
    int loc_t = 0;
    double ic = 0;
    double new_i = 0;
    double loc_dV = 0;
    temp = 0;
    tdV = 0;    

    for(int j=0; j<L; j++)
	for(int i=0; i<L; i++){
	    for(int k=0; k<2; k++){
		if(j < L-1 && k == 0)
		    dV = fabs(sys->voltage[i+(j+1)*L] - sys->voltage[i+j*L]);
		else if(j < L-1 && k == 1)
		    dV = fabs(sys->voltage[ir[i] + (j+1)*L] - sys->voltage[i+j*L]);
		else
		    dV = fabs(sys->max_voltage - sys->voltage[i+j*L]);

		
		t = sys->threshold[i+j*L][k];


		// current = (1-sys->b)*t + dV*sys->b
		// conductance = (1-sys->b)*t/dV + sys->b

		// Conductance can only increase until it hits
		// a specified second threshold, where the bond saturates,
		// e.g. 10*t.

		//if(dV > u && u >= t && !sys->perm_updated[i+j*L][k]){
		//if(dV > t && !sys->perm_updated[i+j*L][k]){
		if(dV > t && !sys->perm_updated[i+j*L][k]){
		    
		    //temp = sys->b*sys->max_dV[i+j*L][k]/(t*(sys->b-1));  // dV/u
		    //temp = dV/t;
		    if(sys->perm_updated[i+j*L][k]){
			ic = sys->conductance[i+j*L][k]*dV;
			new_dvu = ic - sys->max_i[i+j*L][k] > max_dvu;
			temp = ic - sys->max_i[i+j*L][k];
		    }
		    else if(sys->updated[i+j*L][k]){
			ic = (1-sys->b)*t + dV*sys->b;
			new_dvu = ic - sys->max_i[i+j*L][k] > max_dvu;
			temp = ic - sys->max_i[i+j*L][k];
		    }
		    else{
			ic = (1-sys->b)*t + dV*sys->b;
			new_dvu = ic-t > max_dvu;
			temp = ic-t;
		    }

		    // Ensure that i never reverses below nt if perm_updated
		    
		    max_dvu = (new_dvu) ? temp : max_dvu;
		}
		sys->max_dV[i+j*L][k] = dV;
		
		/* if(dV > t && !sys->perm_updated[i+j*L][k]){ */
		/*     ic = (1-sys->b)*t +sys->b*dV; */
		/*     new_dvu = ic - sys->max_i[i+j*L][k] > new_i; */
		/* } */

		if(new_dvu){
		    tdV = t/dV;
		    dvu_coord = (new_dvu) ? i+j*L : dvu_coord;
		    dvu_bond = (new_dvu) ? k : dvu_bond;
		    
		    new_i = ic > new_i ? ic : new_i;
		    if(new_i)
			max_i = ic > nt*t;
		    //new_i = ic > nt*t ? nt*t : ic;
		    loc_t = t;
		    loc_dV = dV;

		    count++;
		}
		new_dvu = 0;
		sys->num_updated_bonds += sys->updated[i+j*L][k];
	    }
	}
    
    if(count > 0){
	printf("Woo! Update! %.6f    %d  %d    %.4f\n",
	       (1-sys->b)*tdV + sys->b, dvu_coord, dvu_bond, sys->max_voltage);

	assert(tdV > 0 && tdV < 1);
	sys->conductance[dvu_coord][dvu_bond] = (1-sys->b)*tdV +sys->b;//DBL_MAX;
	sys->perm_updated[dvu_coord][dvu_bond] = 1;
	sys->max_i[dvu_coord][dvu_bond] = (1-sys->b)*t + loc_dV*sys->b;
	if(max_i){
	    sys->conductance[dvu_coord][dvu_bond] = nt*tdV;
	    sys->perm_updated[dvu_coord][dvu_bond] = 1;
	    sys->max_i[dvu_coord][dvu_bond] = nt*loc_t;
	}

	sys->num_updated_bonds -= sys->updated[dvu_coord][dvu_bond];
	sys->updated[dvu_coord][dvu_bond] = 1;
	sys->num_updated_bonds += sys->updated[dvu_coord][dvu_bond];
	//sys->max_i[dvu_coord][dvu_bond] = ((1-sys->b)*tdV +sys->b)*
	updated = 1;
    }

    else{
    	for(int j=0; j<L; j++)
    	    for(int i=0; i<L; i++)
    		for(int k=0; k<2; k++){
    		    if(j < L-1 && k == 0)
    			dV = fabs(sys->voltage[i+(j+1)*L] - sys->voltage[i+j*L]);
    		    else if(j < L-1 && k == 1)
    			dV = fabs(sys->voltage[ir[i] + (j+1)*L]
    				  - sys->voltage[i+j*L]);
    		    else
    			dV = fabs(sys->max_voltage - sys->voltage[i+j*L]);


    		    t = sys->threshold[i+j*L][k];

    		    /* if(t > dV && !sys->perm_updated[i+j*L][k]) */
		    /* 	sys->conductance[i+j*L][k] = (1-sys->b)*tdV +sys->b; */
		    if(sys->perm_updated[i+j*L][k])
		    	sys->conductance[i+j*L][k] = sys->max_i[i+j*L][k]/dV;
		    /* else */
		    /* 	sys->conductance[i+j*L][k] = 1; */

    		}
    }
    

    /* if(updated == 1) */
    /* 	printf("Updated! Bond [%d][%d], max voltage = %.4f\n", */
    /* 	       dvu_coord, dvu_bond, sys->max_voltage); */
    free(ir);

    return updated;

}
int updateConductanceSuperconductingPure(struct System *sys){
 
    int L = sys->L;
    int updated = 0;
    double dV, t, temp;// tdV;
    /* int **old_updated = (int**)calloc(L*L, sizeof(int*)); */
    int *ir = (int*)calloc(L, sizeof(int));
    for(int i=0; i<L; i++)
	ir[i] = i+1;
    ir[L-1] = 0;

    sys->num_updated_bonds = 0;
    /* for(int i=0; i<L*L; i++){ */
    /* 	old_updated[i] = (int*)calloc(2, sizeof(int)); */
    /* 	old_updated[i][0] = sys->updated[i][0]; */
    /* 	old_updated[i][1] = sys->updated[i][1]; */
    /* } */
    

    int count = 0;
    
    //lcc = largest change in conductance
    //double u = 0;
    double max_dvu = 0;  // max(dV/u)
    int dvu_coord = 0;
    int dvu_bond = 0;
    int new_dvu = 0;

    temp = 0;
    //tdV = 0;    
    for(int j=0; j<L; j++)
	for(int i=0; i<L; i++){
	    for(int k=0; k<2; k++){
		if(j < L-1 && k == 0)
		    dV = fabs(sys->voltage[i+(j+1)*L] - sys->voltage[i+j*L]);
		else if(j < L-1 && k == 1)
		    dV = fabs(sys->voltage[ir[i] + (j+1)*L] - sys->voltage[i+j*L]);
		else
		    dV = fabs(sys->max_voltage - sys->voltage[i+j*L]);

		
		t = sys->threshold[i+j*L][k];

		if(dV > t && !sys->perm_updated[i+j*L][k]){
		    temp = dV/t;
		    new_dvu = (temp > max_dvu);
		    max_dvu = (new_dvu) ? temp : max_dvu;
		}

		if(new_dvu){
		    //tdV = t/dV;
		    dvu_coord = (new_dvu) ? i+j*L : dvu_coord;
		    dvu_bond = (new_dvu) ? k : dvu_bond;
		    count++;
		}
		new_dvu = 0;
		sys->num_updated_bonds += sys->updated[i+j*L][k];
	    }
	}
    
    if(count > 0){
	/* printf("Woo! Update! %.6f    %d  %d    %.4f\n", */
	/*        (1-sys->b)*tdV + sys->b, dvu_coord, dvu_bond, sys->max_voltage); */

	sys->conductance[dvu_coord][dvu_bond] = 1e5;//DBL_MAX;
	sys->perm_updated[dvu_coord][dvu_bond] = 1;
	sys->num_updated_bonds -= sys->updated[dvu_coord][dvu_bond];
	sys->updated[dvu_coord][dvu_bond] = 1;
	sys->num_updated_bonds += sys->updated[dvu_coord][dvu_bond];
	//sys->max_i[dvu_coord][dvu_bond] = ((1-sys->b)*tdV +sys->b)*
	updated = 1;
    }

    free(ir);

    return updated;

}
int updateConductanceFuse(struct System *sys){
 
    int L = sys->L;
    int updated = 0;
    double dV, t, temp;
    int **old_updated = (int**)calloc(L*L, sizeof(int*));
    int *ir = (int*)calloc(L, sizeof(int));
    for(int i=0; i<L; i++)
	ir[i] = i+1;
    ir[L-1] = 0;

    sys->num_updated_bonds = 0;
    for(int i=0; i<L*L; i++){
    	old_updated[i] = (int*)calloc(2, sizeof(int));
    	old_updated[i][0] = sys->updated[i][0];
    	old_updated[i][1] = sys->updated[i][1];
    }
    

    int count = 0;
    
    int it_coord = 0;
    int it_bond = 0;
    int new_it = 0;
    double max_it = 1e-3;  // max(i/t);
    //double I;

    for(int j=0; j<L; j++)
	for(int i=0; i<L; i++){
	    for(int k=0; k<2; k++){
		if(j < L-1 && k == 0)
		    dV = fabs(sys->voltage[i+(j+1)*L] - sys->voltage[i+j*L]);
		else if(j < L-1 && k == 1)
		    dV = fabs(sys->voltage[ir[i] + (j+1)*L] - sys->voltage[i+j*L]);
		else
		    dV = fabs(sys->max_voltage - sys->voltage[i+j*L]);

		t = sys->threshold[i+j*L][k];
		
		//I = (!sys->perm_updated[i+j*L][k])*sys->max_dV[i+j*L][k];
		temp = (!sys->perm_updated[i+j*L][k])*dV/t;
		new_it = (temp > max_it);
		max_it = (new_it) ? temp : max_it;
		
		sys->max_dV[i+j*L][k] = dV;

		if(new_it){
		    it_coord = (new_it) ? i+j*L : it_coord;
		    it_bond = (new_it) ? k : it_bond;
		    count++;
		    //printf("  FUSE:  Max_voltage = %.4f, count = %d, max_it = %.4e, bond (%d)[%d], temp = %.4e\n",
		    //	   sys->max_voltage, count, max_it, i+j*L, k, temp);

		}
		new_it = 0;
		sys->num_updated_bonds += sys->updated[i+j*L][k];
	    }
	}
    
    if(count > 0){
	sys->conductance[it_coord][it_bond] = 0;
	sys->perm_updated[it_coord][it_bond] = 1;
	sys->num_updated_bonds -= sys->updated[it_coord][it_bond];
	sys->updated[it_coord][it_bond] = 1;
	sys->num_updated_bonds ++;
	
	updated = 1;
    }
    
    for(int i=0; i<L*L; i++)
	free(old_updated[i]);
    free(old_updated);
    free(ir);

    return updated;

}


int updateConductanceAll(struct System *sys){
 
    int L = sys->L;
    int updated = 0;
    double dV, t, tempc;
    int **old_updated = (int**)calloc(L*L, sizeof(int*));
    int *ir = (int*)calloc(L, sizeof(int));
    for(int i=0; i<L; i++)
	ir[i] = i+1;
    ir[L-1] = 0;

    sys->num_updated_bonds = 0;
    for(int i=0; i<L*L; i++){
    	old_updated[i] = (int*)calloc(2, sizeof(int));
    	old_updated[i][0] = sys->updated[i][0];
    	old_updated[i][1] = sys->updated[i][1];
    }
    
    /*  Checks if the current flowing through the bonds exceed the threshold.  */
    /*  Absolute value of difference in voltage makes it independent of current direction.  */
    /*  Make sure the upper row with conductance bonds is checked for the voltage difference  */
    /*  between the maximum voltage and the node voltage!  */
    /*  For negative slopes after threshold (conductance < t/dV), the conductance is >= 0.  */
    /*  This is ensured by setting the conductance to the maximum of 0 and (1-b)t/dV + b  */
    /*  The bonds are considered elastic until they reach a point  dV >= t where the current is 0.  */
    /*  At this point, the bond is considered to break (like an elastic spring or rubber band),  */
    /*  and the conductance is set permanently to 0. For the fuse model, this will be when dV = t,  */
    /*	while for theta > -90, it will be later. For theta >= 0, this will never happen, and thus  */
    /*  those bonds will always be considered elastic.  */

    
    double tdV = 0;

    // This first run through can probably hold the general bond updated
    // which is done further down. Then, if count > 0, the ltv_bond is changed.
    for(int j=0; j<L; j++)
	for(int i=0; i<L; i++){
	    for(int k=0; k<2; k++){
		if(j < L-1 && k == 0)
		    dV = fabs(sys->voltage[i+(j+1)*L] - sys->voltage[i+j*L]);
		else if(j < L-1 && k == 1)
		    dV = fabs(sys->voltage[ir[i] + (j+1)*L] - sys->voltage[i+j*L]);
		else
		    dV = fabs(sys->max_voltage - sys->voltage[i+j*L]);
		t = sys->threshold[i+j*L][k];
		tdV = t/dV;
		
		tempc = (tdV < 1) ? (1-sys->b)*tdV + sys->b : 1;
		tempc = (tempc > 0) ? tempc : 0;
		assert( !(tempc == 0 && sys->perm_updated[i+j*L][k] == 0 && dV > 0) );
		sys->conductance[i+j*L][k] = (sys->perm_updated[i+j*L][k]) ? 0 : tempc;
		sys->updated[i+j*L][k] = (tdV < 1 || sys->perm_updated[i+j*L][k]);
		sys->perm_updated[i+j*L][k] = (tempc == 0);
		sys->num_updated_bonds += sys->updated[i+j*L][k];
		sys->max_dV[i+j*L][k] = dV;
	    }
	}
    

    for(int i=0; i<L*L; i++)
    	if( (sys->updated[i][0] != old_updated[i][0]) ||
    	    (sys->updated[i][1] != old_updated[i][1]) ){
    	    updated = 1;
    	    break;
    	}
    
    for(int i=0; i<L*L; i++)
    	free(old_updated[i]);
    free(old_updated);
    free(ir);
    
    return updated;

}




void breakBonds(struct System *sys){
    
    for(int i=0; i<sys->L*sys->L; i++){
	sys->perm_updated[i][0] = (sys->perm_updated[i][0] || sys->updated[i][0]);
	sys->perm_updated[i][1] = (sys->perm_updated[i][1] || sys->updated[i][1]);
    }

}



double checkCurrentConservation(struct System *sys){

    /*------------------------------------------------------------------------------------*/
    /*  Checks that the total current flowing between the electrodes is conserved.        */
    /*  This is done by checking that the sum of voltage differences between the          */
    /*  maximum voltage and the upper nodes (total upper current) is equal to the         */
    /*  sum of the voltage differences between the two lowest rows (total lower current)  */
    /*                                                                                    */
    /*  Returns the absolute difference between the total upper and lower currents        */
    /*------------------------------------------------------------------------------------*/

    int L = sys->L;
    double upper_current = 0;
    double lower_current = 0;
    int *ir = (int*)calloc(L, sizeof(int));

    for(int i=0; i<L; i++)
	ir[i] = i+1;

    ir[L-1] = 0;


    for(int i=0; i<L; i++){
	lower_current +=  sys->conductance[i][0] * (sys->voltage[i+L] - sys->voltage[i])
	                 +sys->conductance[i][1] * (sys->voltage[ir[i]+L] - sys->voltage[i]);

	upper_current +=  sys->conductance[i+(L-1)*L][0] * (sys->max_voltage - sys->voltage[i+(L-1)*L])
	                 +sys->conductance[i+(L-1)*L][1] * (sys->max_voltage - sys->voltage[i+(L-1)*L]);
    }

    free(ir);
    return fabs(upper_current-lower_current);
}



int CG(struct System *sys){

    /*  Ordinary conjugate gradient algorithm, see "An Introduction to the Conjugate Gradient  */
    /*  Method Without the Agonizing Pain" by J.R. Shewchuk for general details on the         */
    /*  conjugate gradient method.                                                             */
    /*  Returns number of iterations needed for the final answer (in case of bugs).            */
    /*  Adapted from a FORTRAN method written by Alex Hansen, based on the general approach    */
    /*  described in "Fourier Acceleration of Iterative Processes in Disordered Systems",      */
    /*  J. Stat. Phys. 52, 747-773 (1988) by G.G. Batrouni and A. Hansen.                      */


    int L = sys->L;
    int N = L*L;
    int m = 0;
    double *dir = (double*)calloc(N, sizeof(double));  /*  Step direction (p in BH)   */
    double *res = (double*)calloc(N, sizeof(double));  /*  Residual vector (r in BH)  */
    double *temp_vec = (double*)calloc(N, sizeof(double));
    int *il = (int*)calloc(L, sizeof(int));
    int *ir = (int*)calloc(L, sizeof(int));
    int ij;
    double alpha, beta, res_sq, res_sq_old, temp_scalar;  
    /*  alpha = step length, beta = Gram-Schmidt constants  */
    double error = 1e-10;  // cut off when res*res < error
    double dV = sys->max_voltage - sys->min_voltage;
    //int nthreads = sys->nthreads;

    /*  For easy implementation of periodic boundary conditions  */
    //#pragma omp parallel for num_threads(nthreads)
    for(int i=0; i<L; i++){
	il[i] = i-1;
	ir[i] = i+1;
    }
    il[0] = L-1;
    ir[L-1] = 0;


    /*  Find initial residual and direction  */
    //#pragma omp parallel for num_threads(nthreads) collapse(2)
    for(int j=0; j<L; j++){
	for(int i=0; i<L; i++){
	    ij = i + j*L;
	    dir[ij] = - sys->conductance[ij][0]*(sys->voltage[i + ir[j]*L] - sys->voltage[ij])
		      - sys->conductance[ij][1]*(sys->voltage[ir[i] + ir[j]*L] - sys->voltage[ij])
		      - sys->conductance[il[i] + il[j]*L][1]*(sys->voltage[il[i] + il[j]*L] - sys->voltage[ij])
		      - sys->conductance[i + il[j]*L][0]*(sys->voltage[i + il[j]*L] - sys->voltage[ij]);
	    res[ij] = dir[ij];
	}
    }

    // Electrodes are defined to be nodes 0-L (bottom electrode) and nodes L*L - L*L+L
    // (upper electrode). Thus, all bonds we have stored come out from the nodes 0-(L*L-1).

    /* for(int i=0; i<L; i++){ */
    /* 	dir[i] += sys->conductance[i+(L-1)*L][0]*dV + sys->conductance[il[i]+(L-1)*L][1]*dV; */
    /* 	res[i] = dir[i]; */
    /* } */
    
    //#pragma omp parallel for num_threads(nthreads)
    for(int i=L*L-L; i<L*L; i++){
    	dir[i] += -sys->conductance[i][0]*dV - sys->conductance[i][1]*dV;
    	res[i] = dir[i];
    }


    /*  First row is an electrode with V = min_voltage = 0. This , together with only checking  */
    /*	from the second row and onwards in the rest of the iterations in the CG (i.e., the ones */
    /*  that start from L instead of 0), ensures that it stays that way                         */

    for(int i=0; i<L; i++){
	dir[i] = 0;
	res[i] = 0;
    }


    /*  End of finding initial residual and direction  */

    res_sq = 0;
    //#pragma omp parallel for num_threads(nthreads) reduction(+:res_sq)
    for(int i=L; i<N; i++){
	res_sq += res[i]*res[i];
	/*printf("res[%d] = %.4f\n", i, res[i]);*/
    }

    /*  Solve system iteratively  */

    while(res_sq > error && m < 20*N){

	//#pragma omp parallel for num_threads(nthreads) collapse(2)
	for(int j=0; j<L; j++){
	    for(int i=0; i<L; i++){
		ij = i + j*L;
		temp_vec[ij] = sys->conductance[ij][0] * (dir[i + ir[j]*L] - dir[ij])
		             + sys->conductance[ij][1] * (dir[ir[i] + ir[j]*L] - dir[ij])
		             + sys->conductance[il[i] + il[j]*L][1] * (dir[il[i] + il[j]*L] - dir[ij])
		             + sys->conductance[i + il[j]*L][0] * (dir[i + il[j]*L] - dir[ij]);
	    }
	}

	temp_vec[0] = 0;
	temp_scalar = 0;
	//#pragma omp parallel for num_threads(nthreads) reduction(+:temp_scalar)
	for(int i=L; i<N; i++)
	    temp_scalar += dir[i]*temp_vec[i];

	
	alpha = res_sq/temp_scalar;
	
	res_sq_old = res_sq;
	res_sq = 0;


	//#pragma omp parallel for num_threads(nthreads) reduction(+:res_sq)
	for(int i=L; i<N; i++){
		sys->voltage[i] += alpha*dir[i];
		res[i] -= alpha*temp_vec[i];
		res_sq += res[i]*res[i];
	}
    


	beta = res_sq/res_sq_old;
	//#pragma omp parallel for num_threads(nthreads)
	for(int i=L; i<N; i++)
	    dir[i] = res[i] + beta*dir[i];

	m++;
    }

    // Calculate the total current flowing through the system.
    // Is currently not needed anywhere, and is thus commented out.
    /* sys->total_current = 0; */
    /* for(int i=0; i<L; i++){ */
    /* 	sys->total_current +=  sys->conductance[i][0] * (sys->voltage[i+L] - sys->voltage[i]) */
    /* 	                      +sys->conductance[i][1] * (sys->voltage[ir[i]+L] - sys->voltage[i]); */
    /* } */
    /* sys->total_current = 0; */
    /* for(int i=L*L-L; i<L*L; i++){ */
    /* 	sys->total_current +=  sys->conductance[i][0] * (sys->max_voltage - sys->voltage[i]); */
    /* 	    //+sys->conductance[i][1] * (sys->max_voltage - sys->voltage[i]); */
    /* } */

    if(m >=10*N){
	printf("Something might be wrong with the conjugate gradient method,");
	printf("it has used more iterations than what should be necessesary... %d\n", m);
    }

    free(res);
    free(dir);
    free(temp_vec);
    free(il);
    free(ir);

    return m;

}



double HKBonds(struct System *sys){

    /*  Runs the Hoshen-Kopelman algorithm on the bonds to find the percolating cluster  */
    /*  Currently only works for fuses (i.e. no plastic-elastic systems) and             */
    /*  is a fucking mess. Should be fixed if it's to be used a lot (probably not)       */

    /*  TODO: Fix so it checks the updated array instead of conductance (done). 
	Check if we have a spanning cluster that breaks the network (use shortestPath?)  */
    /*  UPDATE: This function is stupid. Check out HK instead, which goes through the 
	nodes in the same manner as shortestPath()                                       */

    int L = sys->L;
    int *cluster_array = (int*)calloc(L*(1+L), sizeof(int));
    int *ir = (int*)calloc(L, sizeof(int));
    int **bond_cluster;
    bond_cluster = (int**)calloc(L*L, sizeof(int*));
    int ij, min_index, max_index, k, temp;
    double width;

    for(int i=0; i<L*L; i++)
    	bond_cluster[i] = (int*)calloc(2, sizeof(int));

    for(int i=0; i<L; i++)
	ir[i] = i+1;
    ir[L-1] = 0;

    cluster_array[0] = 0;

    removeArms(sys->conductance, L);

    k = 2;
    cluster_array[k] = k;

    /*  Check bonds from the first node  */
    if(sys->updated[0][0]){
	bond_cluster[0][0] = k;
	k++;
	cluster_array[k] = k;
    }
    if(sys->updated[0][1]){
	if(bond_cluster[0][0] > 1)
	    bond_cluster[0][1] = bond_cluster[0][0];
	else{
	    bond_cluster[0][1] = k;
	    k++;
	    cluster_array[k] = k;
	}
    }
    
    /*  Run through first row  */
    for(int i=1; i<L; i++){
	if(sys->updated[i][0]){
	    if(bond_cluster[i-1][1] > 1)
		bond_cluster[i][0] = bond_cluster[i-1][1];
	    else{
		bond_cluster[i][0] = k;
		k++;
		cluster_array[k] = k;
	    }
	}
	if(sys->updated[i][1]){
	    if(bond_cluster[i][0] > 1)
		bond_cluster[i][1] = bond_cluster[i][0];
	    else{
		bond_cluster[i][1] = k;
		k++;
		cluster_array[k] = k;
	    }
	}
    }
    
    /*  Run through the rest  */
    for(int j=1; j<L; j++){
	
	/*  First, check the bonds from the first node in the row  */
	ij = j*L;

	min_index = 1;
	max_index = 1;
	
	/*  Check the vertical bond  */
	if(sys->updated[ij][0]){
	    if(bond_cluster[ij-1][1] > 1)
		min_index = bond_cluster[ij-1][1];
	    else if(bond_cluster[ij-1][0] > 1)
		min_index = bond_cluster[ij-1][0];
	    else{
		min_index = k;
		k++;
		cluster_array[k] = k;
	    }
	    while(cluster_array[min_index] < 0)
		min_index = -cluster_array[min_index];
	    bond_cluster[ij][0] = min_index;
	}
	
	/*  Check the diagonal bond  */
	if(sys->updated[ij][1]){
	    if(bond_cluster[ij][0] > 1 && bond_cluster[ij-L][0] > 1){
		min_index = min(bond_cluster[ij][0] , bond_cluster[ij-L][0]);
		max_index = max(bond_cluster[ij][0] , bond_cluster[ij-L][0]);
	    }
	    else if(bond_cluster[ij][0] > 1 && bond_cluster[ij-L][1] > 1){
		min_index = min(bond_cluster[ij][0] , bond_cluster[ij-L][1]);
		max_index = max(bond_cluster[ij][0] , bond_cluster[ij-L][1]);
	    }
	    else if(bond_cluster[ij][0] > 1){
		min_index = bond_cluster[ij][0];
		max_index = min_index;
	    }
	    else if(bond_cluster[ij-L][0] > 1){
		min_index = bond_cluster[ij-L][0];
		max_index = min_index;
	    }
	    else if(bond_cluster[ij-L][1] > 1){
		min_index = bond_cluster[ij-L][1];
		max_index = min_index;
	    }
	    else{
		min_index = k;
		max_index = min_index;
		k++;
		cluster_array[k] = k;
	    }
	    while(cluster_array[min_index] < 0)
		min_index = -cluster_array[min_index];
	    while(cluster_array[max_index] < 0)
		max_index = -cluster_array[max_index];
	    if(min_index != max_index){
		temp = min(min_index,max_index);
		cluster_array[max(min_index,max_index)] = -temp;
	    }

	    bond_cluster[ij][1] = min(min_index, max_index);
	}

	/*  Run through the other columns  */

	for(int i=1; i<L; i++){
	    ij = i + j*L;

	    min_index = 1;
	    max_index = 1;

	    if(sys->updated[ij][0]){

		if(bond_cluster[ij-1][1] > 1)
		    min_index = bond_cluster[ij-1][1];
		else if(bond_cluster[ij-1-L][0] > 1)
		    min_index = bond_cluster[ij-1-L][0];
		else if(bond_cluster[ij-1-L][1] > 1)
		    min_index = bond_cluster[ij-1-L][1];
		else{
		    min_index = k;
		    k++;
		    cluster_array[k] = k;
		}
		while(cluster_array[min_index] < 0)
		    min_index = -cluster_array[min_index];
		bond_cluster[ij][0] = min_index;
	    }
	    if(sys->updated[ij][1]){

		if((bond_cluster[ij][0] > 1) && (bond_cluster[ij-L][0] > 1)){
		    min_index = min(bond_cluster[ij][0] , bond_cluster[ij-L][0]);
		    max_index = max(bond_cluster[ij][0] , bond_cluster[ij-L][0]);
		}
		else if(bond_cluster[ij][0] > 1 && bond_cluster[ij-L][1] > 1){
		    min_index = min(bond_cluster[ij][0] , bond_cluster[ij-L][1]);
		    max_index = max(bond_cluster[ij][0] , bond_cluster[ij-L][1]);
		}
		else if(bond_cluster[ij][0] > 1){
		    min_index = bond_cluster[ij][0];
		    max_index = min_index;
		}
		else if(bond_cluster[ij-L][0] > 1){
		    min_index = bond_cluster[ij-L][0];
		    max_index = min_index;
		}
		else if(bond_cluster[ij-L][1] > 1){
		    min_index = bond_cluster[ij-L][1];
		    max_index = min_index;
		}
		else{
		    min_index = k;
		    max_index = min_index;
		    k++;
		    cluster_array[k] = k;
		}
		while(cluster_array[min_index] < 0)
		    min_index = -cluster_array[min_index];
		while(cluster_array[max_index] < 0)
		    max_index = -cluster_array[max_index];
		if(min_index != max_index){
		    temp = min(min_index,max_index);
		    cluster_array[max(min_index,max_index)] = -temp;
		}
		    
		bond_cluster[ij][1] = min(min_index, max_index);

	    }
	}	    
    }
    

    /* printf("Cluster values before boundary is checked\n"); */
    /* for(int i=L-1; i>=0; i--){ */
    /* 	for(int j=0; j<L; j++){ */
    /* 	    ij = i*L+j; */
    /* 		printf("%d|%d  ", bond_cluster[ij][0], bond_cluster[ij][1]); */
    /* 	} */

    /* 	printf("\n"); */
    /* } */
    /* printf("\n"); */


     /* Ensure that different clusters spanning the boundary is the same one */
    for(int j=1; j<L; j++){
    	ij = j*L;
    	min_index = 1;
    	max_index = 1;
	temp = 0;

	/*  The commented part worked for several situations, not quite what is needed though  */
	/*  Currently kept as backup, in case things go haywire again.  */

    	/* if((bond_cluster[ij][0] > 1) && (bond_cluster[ij-L][0] > 1)){ */
    	/*     min_index = min(bond_cluster[ij][0] , bond_cluster[ij-L][0]); */
    	/*     max_index = max(bond_cluster[ij][0] , bond_cluster[ij-L][0]); */
	/*     temp = 1; */
    	/* } */
    	/* else if(bond_cluster[ij][0] > 1 && bond_cluster[ij-L][1] > 1){ */
    	/*     min_index = min(bond_cluster[ij][0] , bond_cluster[ij-L][1]); */
    	/*     max_index = max(bond_cluster[ij][0] , bond_cluster[ij-L][1]); */
	/*     temp = 1; */
    	/* } */

    	if((bond_cluster[ij][0] > 1) && (bond_cluster[ij-1][0] > 1)){
    	    min_index = min(bond_cluster[ij][0] , bond_cluster[ij-1][0]);
    	    max_index = max(bond_cluster[ij][0] , bond_cluster[ij-1][0]);
	    temp = 1;
    	}
    	else if(bond_cluster[ij][0] > 1 && bond_cluster[ij-1][1] > 1){
    	    min_index = min(bond_cluster[ij][0] , bond_cluster[ij-1][1]);
    	    max_index = max(bond_cluster[ij][0] , bond_cluster[ij-1][1]);
	    temp = 1;
    	}
    	else if(bond_cluster[ij][0] > 1 && bond_cluster[ij+L-1][1] > 1){
    	    min_index = min(bond_cluster[ij][0] , bond_cluster[ij+L-1][1]);
    	    max_index = max(bond_cluster[ij][0] , bond_cluster[ij+L-1][1]);
	    temp = 1;
    	}

	if(temp){
	    while(cluster_array[min_index] < 0)
		min_index = -cluster_array[min_index];
	    while(cluster_array[max_index] < 0)
		max_index = -cluster_array[max_index];
	
	    if(min_index != max_index){
		temp = min(min_index,max_index);
		cluster_array[max(min_index,max_index)] = -temp;
	    }
	    bond_cluster[ij][1] = min(min_index, max_index);
	}
    }
    
    /*  Set the value of all components of bond_cluster to the cluster they're part of  */
    for(int i=0; i<L*L; i++)
    	for(int j=0; j<2; j++){
    	    min_index = bond_cluster[i][j];
    	    if(min_index >= 2){
    		while(cluster_array[min_index] < 0)
    		    min_index = -cluster_array[min_index];
    		bond_cluster[i][j] = min_index;
    	    }
    	}


    /* printConductance(sys); */

    /* printf("Cluster values\n"); */
    /* for(int i=L-1; i>=0; i--){ */
    /* 	for(int j=0; j<L; j++){ */
    /* 	    ij = i*L+j; */
    /* 		printf("%d  %d\t", bond_cluster[ij][0], bond_cluster[ij][1]); */
    /* 	} */

    /* 	printf("\n"); */
    /* } */
    /* printf("\n"); */
    /* for(int i=0; i<k+2; i++) */
    /* 	printf("%d , ", cluster_array[i]); */
    /* printf("\n"); */

    width = getWidth(bond_cluster, L);



    for(int i=0; i<L*L; i++)
	free(bond_cluster[i]);
    free(bond_cluster);
    free(cluster_array);
    free(ir);

    return width;

}






int HKnpb(int **updated, int *nodes, int L){
    
    /*  Runs Hoshen-Kopelman to find if we have a spanning cluster (broken network)  */
    /*  nodes is assumed to be of size N+1 = (L+1)*(L+1) + 1 or a NULL pointer       */

    int N = (L+1)*(L+1);
    int broken_network = 0;
    int nullpt = 0;
    int i, j, n, k, min_index, max_index, temp, an, bn, ln, rn;
    int *bc = calloc(L+1, sizeof(int));
    int *cluster_number = calloc(N, sizeof(int));


    if(nodes == NULL){
	nodes = calloc(N+1, sizeof(int));
	nullpt = 1;
    }

    for(int i=0; i<L+1; i++){
	bc[i] = i;
    }
    bc[L] = 0;
    nodes[N] = INT_MAX;
    
    k = 2;
    cluster_number[1] = 1;
    cluster_number[2] = 2;



    for(i=0; i<L+1; i++){	    
	for(j=0; j<L+1; j++){

	    min_index = 1;
	    max_index = 1;

	    n = i*(L+1) + j;


	    ln = (i > 0 && j > 0 && updated[i*L + bc[j] - L][0]) ? n-L-2 : N;
	    bn = (i > 0 && updated[i*L + bc[j] - L][1]) ? n-L-1 : N;
	    rn = (i < L && j < L && updated[i*L + bc[j+1]][0]) ? 1 : N;
	    an = (i < L && updated[i*L + bc[j]][1]) ? 1 : N;


	    if( bn < N || ln < N ){ 
		//Connected to either lower and left node
		min_index = min( nodes[ln], nodes[bn] );
		max_index = max( nodes[ln], nodes[bn] );
	    }
	    else if( an < N || rn < N ){
		//Connected to either upper or right node
		min_index = k;
		max_index = k;
		k++;
		cluster_number[k] = k;		
	    }
	    
	    while(cluster_number[min_index] < 0)
		min_index = -cluster_number[min_index];
	    
	    while(max_index < INT_MAX && cluster_number[max_index] < 0)
		max_index = -cluster_number[max_index];
	    

	    if(max_index < INT_MAX && min_index != max_index){
		temp = min(min_index, max_index);
		cluster_number[max(min_index, max_index)] = -temp;
	    }

	    nodes[n] = min(min_index, max_index);

	}
    }    


    /*  Check if the network is broken, i.e. we have a spanning cluster.   */
    /*  This is the case if the first and last node in a row belong to     */
    /*  the same cluster, as these nodes are the same due to periodic bc.  */
    for(i=0; i<L+1; i++){
	
	min_index = nodes[i*(L+1)];
	max_index = nodes[i*(L+1) + L];

	while(cluster_number[min_index] < 0){
	    min_index = -cluster_number[min_index];
	}
	while(cluster_number[max_index] < 0){
	    max_index = -cluster_number[max_index];
	}

	if(max_index == min_index && max_index > 1)
	    broken_network = 1;
    }


    free(bc);
    free(cluster_number);
    if(nullpt)
	free(nodes);

    return broken_network;
}


int HKV(int **updated, int L){

    /*  Runs Hoshen-Kopelman to see if there is a spanning cluster of voltage  */
    /*  nodes between the bus-bars. If there is none, the network is broken.   */
    /*  The bus bars are considered to be voltage nodes, as there are bonds    */
    /*  connecting them with the voltage nodes.                                */
    /*  Nodes on the lower bus bar (i=0) will not have any neighbours below    */
    /*  or to the left. Similarly, nodes on the upper bus bar (i=L) will not   */
    /*  have any neighbours above or to the right.                             */
    /*  This is essentially bond percolation where sites are connected if the  */
    /*  bond between them is not updated. As bonds are updated, sites are      */
    /*  being disconnected, and we essentially have reverse bond percolation.  */


    if(!updated)
	return 0;

    int N = (L+1)*L;
    int broken_network = 0;
    int spanning_cluster = 0;
    int c, i, j, k, n, min_index, max_index, temp_min, temp_max, an, bn, rn, ln;
    int *cl = calloc(L, sizeof(cl));    // Left neighbour (column)
    int *cr = calloc(L, sizeof(cr));    // Right neighbour (column)
    int *cluster_number = calloc(N, sizeof(cluster_number));
    int *vnodes = calloc(N+1, sizeof(vnodes));

    for(i=0; i<L; i++){
	cl[i] = i-1;
	cr[i] = i+1;
    }
    cl[0] = L-1;
    cr[L-1] = 0;

    vnodes[N] = INT_MAX;
    cluster_number[1] = 1;
    cluster_number[2] = 2;
    k = 2;


    // Run through the system left-right, bottom-up
    // Note that the sites (voltage nodes) are connected if
    // the bonds have NOT reached the threshold, that is, 
    // they have not been updated and updated[n][x] = 0.
    // Also connects cluster at the boundary.
    // min_index and max_index contains the cluster numbers 
    // of the neighbouring sites that the current site is
    // connected to. 
    // The first cluster gets cluster number 2. There is no
    // specific reason for this, it's just a habit.

    for(i=0; i<L+1; i++){
	for(j=0; j<L; j++){

	    n = i*L+j;
	    
	    // Fracture path
	    an = (i < L && !updated[n][0])            ? n+L : N;
	    bn = (i > 0 && !updated[n-L][0])          ? n-L : N;
	    rn = (i < L && !updated[n][1])            ? i*L+cr[j] : N;
	    ln = (i > 0 && !updated[(i-1)*L+cl[j]][1])    ? (i-1)*L+cl[j] : N;

	    // Flow path
	    /* an = (i < L && updated[n][0])            ? n+L : N; */
	    /* bn = (i > 0 && updated[n-L][0])          ? n-L : N; */
	    /* rn = (i < L && updated[n][1])            ? i*L+cr[j] : N; */
	    /* ln = (i > 0 && updated[(i-1)*L+cl[j]][1])    ? (i-1)*L+cl[j] : N; */

	    min_index = 1;
	    max_index = 1;

	    if(bn < N || ln < N){  
		// Connected to a site already belonging to a cluster
		// Note that min_index will be the minimum of the cluster
		// number of the two sites which the current node is connected to.
		// If one site is not connected, the "cluster number" returned is
		// INT_MAX, as ln/bn = N and vnodes[N] = INT_MAX.
		min_index = min(vnodes[bn], vnodes[ln]);
		max_index = max(vnodes[bn], vnodes[ln]);
		assert(min_index);
	    }
	    else if(an < N || rn < N){
		min_index = k;
		max_index = k;
		k++;
		cluster_number[k] = k;
	    }
	    
	    while(cluster_number[min_index] < 0)
		min_index = -cluster_number[min_index];

	    while(max_index < INT_MAX && cluster_number[max_index] < 0)
		max_index = -cluster_number[max_index];

	    // Setting the cluster with the largest cluster number to be
	    // a part of the cluster with the smallest cluster number.
	    if(max_index < INT_MAX && min_index != max_index){
		temp_min = min(min_index, max_index);
		temp_max = max(min_index, max_index);
		cluster_number[temp_max] = -temp_min;
	    }

	    vnodes[n] = min(min_index, max_index);
	    
	}
    }

    // Run through first and last row, see if we have a spanning cluster
    spanning_cluster = 0;
    for(i=0; i<L; i++){
	
	c = vnodes[i];
	while(cluster_number[c] < 0)
	    c = -cluster_number[c];

	if(c > 1){
	    for(j=0; j<L; j++){
		k = vnodes[L*L+j];
		while(cluster_number[k] < 0)
		    k = -cluster_number[k];
		if(c == k){
		    spanning_cluster = 1;
		    break;
		}
	    }
	}
	if(spanning_cluster)
	    break;
    }
    
    if(!spanning_cluster)
	broken_network = 1;
    
    free(cl);
    free(cr);
    free(cluster_number);
    free(vnodes);

    return broken_network;

}


int HK(int **updated, int *in_nodes, int L){
    
    /*  Runs Hoshen-Kopelman to find if we have a spanning cluster (broken network)  */
    /*  Returns broken_network, which is 1 if the network is broken, 0 otherwise.    */
    /*  in_nodes is assumed to be of size N+1 = (L+1)*(L+1) + 1 or a NULL pointer,   */
    /*  and is updated to contain only the spanning cluster if not a NULL pointer.   */
    /*  Check for "flow" clusters in the y-direction, i.e. clusters of nodes which   */
    /*  are connected if the bonds have NOT been updated, that is, they have not     */
    /*  reached the threshold. If there is no spanning cluster of these, then the    */
    /*  network is broken. Also check for "break" clusters, that is nodes which are  */
    /*  connected if the bonds between them have been updated. The largest of these  */
    /*  is almost certainly the breaking cluster, and thus the largest fracture.     */
    /*  The nodes in this fracture is written to in_nodes, if given.                 */
    
    if(!updated)
	return 0;
    

    int N = (L+1)*L;
    int broken_network = 0;
    int spanning_cluster = 0;
    int i, j, n, k, c, min_index, max_index, temp, an, bn, ln, rn, max_cluster_number;
    int *bc = calloc(L+1, sizeof(int));
    int *cluster_number_break = calloc(N, sizeof(cluster_number_break));
    int *break_nodes = calloc(N+1,sizeof(break_nodes));
    int *cluster_number_flow = calloc(N, sizeof(cluster_number_flow));
    int *flow_nodes = calloc(N+1,sizeof(flow_nodes));
    int *spanning_array = calloc(L, sizeof(spanning_array));

    for(int i=0; i<L; i++){
	bc[i] = i+1;
    }
    bc[L-1] = 0;

    break_nodes[N] = INT_MAX;

    k = 2;
    cluster_number_break[1] = 1;
    cluster_number_break[2] = 2;
    max_cluster_number = k;



    //UPDATE 31.10: Run through columnwise instead of rowwise
    //  (i.e. "rotate" network 90 degrees counterclockwise and run rowwise HK)
    //   Starts with upper left corner and runs top-down, left-right
    //   Why? This makes no sense any longer...

    //   Periodic boundary conditions are ignored at first

    for(j=0; j<L; j++)
	for(i=L; i>=0; i--){

	    min_index = 1;
	    max_index = 1;

	    n = i*L + j;

	    // xn are indexvalues for use in break_nodes[] for left, right
	    // top and bottom neighbours. If the bond between the 
	    // current node n and the given neighbour is broken/saturated,
	    // the neighbour coordinate is set, Otherwise N=(L+1)L is set.
	    ln = (i > 0 && j > 0    && updated[n-L][0])   ? n-L-1 : N;
	    bn = (i > 0 &&             updated[n-L][1])   ? 1     : N;
	    rn = (i < L && j < L-1  && updated[n+1][0])   ? 1     : N;
	    an = (i < L &&             updated[n][1])     ? n+L   : N;
	    
	    if( an < N || ln < N ){ 
		// Connected to either upper (above) or left node,
		// i.e. the current node belongs to the same cluster,
		// so take the cluster indices for the top and left node.
		min_index = min( break_nodes[an], break_nodes[ln] );
		max_index = max( break_nodes[an], break_nodes[ln] );
		assert(min_index);
	    }
	    else if( bn < N || rn < N ){
		// Connected to either lower (bottom) or right node,
		// neither of which have been checked. Starts a new
		// cluster.
		min_index = k;
		max_index = k;
		k++;
		cluster_number_break[k] = k;
		max_cluster_number = k;	
	    }


	    // min_index < INT_MAX, run through cluster_number array
	    // to find the correct cluster that min_index belongs to
	    while(cluster_number_break[min_index] < 0)
		min_index = -cluster_number_break[min_index];
	    
	    // find the correct cluster number that max_index belongs to
	    while(max_index < INT_MAX && cluster_number_break[max_index] < 0)
		max_index = -cluster_number_break[max_index];
	    
	    // In case of the current node being connected to two different
	    // clusters, the cluster with the largest cluster number is 
	    // considered a subset of the cluster with the lowest cluster number.
	    // Updates cluster_number for the largest cluster number, sets it
	    // equal to the negative value of the lowest cluster number.
	    if(max_index < INT_MAX && min_index != max_index){
		temp = min(min_index, max_index);
		cluster_number_break[max(min_index, max_index)] = -temp;
		//		printf("temp = %d, max(mini, maxi) = %d, i = %d, j = %d\n", temp, max(min_index, max_index), i, j);
	    }
	    
	    break_nodes[n] = min(min_index, max_index);
	    
	}


    for(i=0; i<N; i++)
	while(cluster_number_break[break_nodes[i]] < 0)
	    break_nodes[i] = -cluster_number_break[break_nodes[i]];


    // Check if the network is broken, i.e. we have a spanning cluster, in the
    // case of no periodic boundary conditions. This is the case if two sites
    // in both the first and last columns belong to the same cluster.

    spanning_cluster = 0;
    int q = 0;    
    for(i=0; i<L+1; i++){
	if(break_nodes[i*L] > 1){
	    c = break_nodes[i*L];
	    for(j=1; j<L+2; j++){
		k = break_nodes[j*L-1];
		if(c == k){  //c > 1
		    spanning_cluster = c;
		    if(q == 0 || c != spanning_array[q-1]){
			spanning_array[q] = c;
			q++;
		    }
		}		    
	    }
	}
    }


    // Connect clusters crossing the boundary, just like earlier.
    // See comments further up for details.
    for(i=1; i<L; i++){
	
    	min_index = 1;
    	max_index = 1;
	
    	n = i*L;
	
    	ln = updated[n-L][0] ? n-1 : N;

	
    	if( ln < N ){
    	    min_index = min( break_nodes[n], break_nodes[ln] );
    	    max_index = max( break_nodes[n], break_nodes[ln] );
	
	    while(cluster_number_break[min_index] < 0)
		min_index = -cluster_number_break[min_index];
	
	    while(max_index < INT_MAX && cluster_number_break[max_index] < 0)
		max_index = -cluster_number_break[max_index];
	
	    assert(max_index > 0);
	    assert(min_index > 0);
	    if(min_index == 1 && max_index == 1){  //new cluster
		min_index = max_cluster_number;
		max_index = max_cluster_number;
		max_cluster_number++;
		cluster_number_break[max_cluster_number] = max_cluster_number;
	    }
	    else if( max(min_index, max_index) > 1 && min(min_index, max_index) < 2){
		min_index = max(min_index, max_index);
		max_index = max(min_index, max_index);
	    }
	    else if(max_index < INT_MAX && min_index != max_index){
		temp = min(min_index, max_index);
		cluster_number_break[max(min_index, max_index)] = -temp;
	    }


	    break_nodes[n] = min(min_index, max_index);
	    break_nodes[ln] = min(min_index, max_index);
	}	
    }


    for(i=0; i<N; i++)
    	while(cluster_number_break[break_nodes[i]] < 0)
    	    break_nodes[i] = -cluster_number_break[break_nodes[i]];

    while(cluster_number_break[spanning_cluster] < 0)
	spanning_cluster = -cluster_number_break[spanning_cluster];

    for(int q=0; q<L; q++)
	while(cluster_number_break[spanning_array[q]] < 0)
	    spanning_array[q] = -cluster_number_break[spanning_array[q]];    


    // If there is a spanning cluster (in the non-periodic boundary case),
    // check if the cluster spans around the boundary.
    if(spanning_cluster){
    	for(i=1; i<L+1; i++){
    	    c = break_nodes[i*L];
    	    k = break_nodes[i*L-1];
	    
    	    while(cluster_number_break[c] < 0)
    		c = -cluster_number_break[c];
    	    while(cluster_number_break[k] < 0)
    		k = -cluster_number_break[k];
	    for(int q=0; q<L; q++){
		if(c == spanning_array[q] && c == k && updated[(i-1)*L][0]){
		    spanning_array[q] = -c;
		}
	    }
	    
	    
	}
    }
    



    /* if(in_nodes){ */
    /* 	printf("after boundary check\n"); */
    /* 	for(i=L; i>=0; i--){ */
    /* 	    if(i<10) */
    /* 		printf(" %d:  ", i); */
    /* 	    else */
    /* 		printf("%d:  ", i); */

    /* 	    for(j=0; j<L; j++){ */

    /* 		n = i*L + j; */
    /* 		printf("%d  ", break_nodes[n]); */
    /* 	    } */
    /* 	    if(i<10) */
    /* 		printf(": %d", i); */
    /* 	    else */
    /* 		printf(":%d", i); */

    /* 	    printf("\n"); */
    /* 	} */
    /* 	printf("\n"); */
    /* 	printf("Broken network: %s\n", broken_network ? "yes" : "no"); */
    /* } */

    /*  in_nodes contains the spanning cluster if given as input array.               */
    /*  UPDATE 29.10: in_break_nodes[in_n] = break_nodes[n] is changed to                         */
    /*                in_break_nodes[in_n] = (break_nodes[n] == spanning_cluster) ? break_nodes[n] : 0  */
    /*                to ensure that ONLY the spanning fracture is in in_nodes        */

    int count = 0;
    int *spanning_cluster_array = calloc(L, sizeof(spanning_cluster_array));
    int r = 0;
    spanning_cluster = 0;
    for(int q=0; q < L; q++)
	if(spanning_array[q] < 0 && spanning_array[q] != -spanning_cluster){
	    spanning_cluster = -spanning_array[q];
	    spanning_cluster_array[r] = spanning_cluster;
	    broken_network = 1;
	    r++;
	    count++;
	}
    
    if(count > 1){
	// More than one cluster satisfy the condition for a spanning cluster as set
	// Check
	int **bonds = calloc(L*L, sizeof(bonds));
	for(i=0; i<L*L; i++)
	    bonds[i] = calloc(2, sizeof(bonds[i]));

	printf("  %d clusters satisfy the condition for spanning clusters!\n", count);
	for(r=0; r<L; r++){
	    int sc = spanning_cluster_array[r];
	    for(i=0; i<L; i++){
		for(j=0; j<L-1; j++){
		    
		    n = i*L + j;
		    bonds[n]  [1] = 
			(break_nodes[n] == sc && break_nodes[n+L] == sc && updated[n][1]); //above
		    bonds[n+1][0] =
			(break_nodes[n] == sc && break_nodes[n+L+1] == sc && updated[n+1][0]);  //right
		}
		//Boundary
		n = i*L + L-1;
		bonds[n]  [1] = 
		    (break_nodes[n] == sc && break_nodes[n+L] == sc && updated[n][1]); //above
		bonds[i*L][0] =
		    (break_nodes[n] == sc && break_nodes[n+1] == sc && updated[i*L][0]);  //right
	    }
	    spanning_cluster = (HKV(bonds, L))*sc;
	    if(spanning_cluster)
		break;	    
	}
	for(i=0; i<L*L; i++)
	    free(bonds[i]);
	free(bonds);
    }

    
    if(in_nodes){

	for(i=0; i<L+1; i++){
	    for(j=0; j<L; j++){
		n = i*L+j;
		in_nodes[n] = (break_nodes[n] == spanning_cluster) ? break_nodes[n] : 0;
	    }
	    //in_nodes[i*(L+1)+L] = in_nodes[i*(L+1)];
	}
	in_nodes[N] = INT_MAX;
    }

    free(spanning_cluster_array);
    free(spanning_array);
    free(bc);
    free(cluster_number_flow);
    free(flow_nodes);
    free(cluster_number_break);
    free(break_nodes);
    return broken_network;
}


int **removeDanglingArms(int **updated, int L){


    int i, j, n, an, bn, rn, ln, neighbours, bonds_removed;

    int *cl = calloc(L, sizeof(cl));    // Left neighbour (column)
    int *cr = calloc(L, sizeof(cr));    // Right neighbour (column)
    int **bonds = calloc(L*L, sizeof(bonds));
    for(i=0; i<L*L; i++)
	bonds[i] = calloc(2, sizeof(bonds[i]));

    for(i=0; i<L; i++){
	cl[i] = i-1;
	cr[i] = i+1;
    }
    cl[0] = L-1;
    cr[L-1] = 0;

    for(i=0; i<L*L; i++){
	bonds[i][0] = updated[i][0];
	bonds[i][1] = updated[i][1];
    }

    bonds_removed = 1;
    while(bonds_removed){
	bonds_removed = 0;
	for(i=0; i<L+1; i++){
	    for(j=0; j<L; j++){

		n = i*L+j;

		an = (i < L && bonds[n][1])            ? 1 : 0;
		bn = (i > 0 && bonds[n-L][1])          ? 1 : 0;
		rn = (i < L && bonds[i*L+cr[j]][0])    ? 1 : 0;
		ln = (i > 0 && bonds[n-L][0])          ? 1 : 0;

		neighbours = an+bn+rn+ln;
	    
	    
		if(neighbours == 1){  
		
		    if(i > 0){
			bonds[n-L][1] = 0;
			bonds[n-L][0] = 0;
		    }
		    if(i < L){
			bonds[n][1] = 0;
			bonds[i*L+cr[j]][0] = 0;
		    }

		    bonds_removed = 1;
		} // end if neighbours == 1
	    


	    
	    }
	}
    }

    free(cr);
    free(cl);

    return bonds;
}


int **getFracture(int **updated, int L){

    /*  Runs HK to get the largest fracture, i.e. the spanning cluster.            */
    /*  Returns bonds, which have values 1 (belongs to fracture) or 0 (does not    */
    /*  belong to fracture).                                                       */
    /*  Bonds is of size [L*L][2]                                                  */

    if(!updated)
	return NULL;

    int N = (L+1)*(L+1);
    int i, j, n;
    int *nodes = calloc(N+1, sizeof(int));  /*  is initialized in HK  */
    int *bc = calloc(L+2, sizeof(int));
    int **bonds = calloc(L*L, sizeof(int*));
    for(i=0; i<L*L; i++)
	bonds[i] = calloc(2, sizeof(int));

    for(i=0; i<L+1; i++)
	bc[i] = i;
    bc[L] = 0;
    bc[L+1] = 1;

    if(!HKV(updated, L)){
	/*  There is no spanning cluster  */

        for(i=0; i<L*L; i++)
	    free(bonds[i]);
	free(bonds);
	free(nodes);
	free(bc);
	return NULL;
    }

    HK(updated, nodes, L);

    /*  Run through all nodes and check neighbours above and to the right.  */
    /*  If neighbours belong to a cluster (i.e. the fracture) and the bond between them  */
    /*  is saturated (updated), the bond between them in bonds is set to 1. Otherwise 0. */

    for(i=0; i<L; i++){
	for(j=0; j<L-1; j++){

	    n = i*L + j;
	    bonds[n]  [1] = 
		(nodes[n] > 1 && nodes[n+L] > 1 && updated[n][1]); //above
	    bonds[n+1][0] =
	    	(nodes[n] > 1 && nodes[n+L+1] > 1 && updated[n+1][0]);  //right
	}
	//Boundary
	n = i*L + L-1;
	bonds[n]  [1] = 
	    (nodes[n] > 1 && nodes[n+L] > 1 && updated[n][1]); //above
	bonds[i*L][0] =
	   (nodes[n] > 1 && nodes[n+1] > 1 && updated[i*L][0]);  //right

    }


    free(nodes);
    free(bc);
    
    return bonds;

}



int **getBackbone(int **updated, int L){

    /*  Runs HK and removes all arms to get the backbone of the spanning cluster.  */
    /*  Returns bonds, which have values 1 (belongs to backbone) or 0 (does not    */
    /*  belong to backbone).                                                       */
    /*	Arms are removed by checking that each node has at least two neighbours,   */
    /*  except for the edges (first and last column) which only need one           */
    /*  neighbour, belonging to a cluster (i.e. nodes[n] >= 2).                    */
    /*  Bonds is of size [L*L][2]                                                  */

    if(!updated)
	return NULL;

    int N = (L+1)*(L+1);
    int i, j, k, l, m, n, arm_removed, bond_removed, neighbours, an, bn, rn, ln,
	broken_network, nodes_check, spanning_cluster;
    int *nodes = calloc(N+1, sizeof(nodes));       /*  is initialized in HK  */
    int *temp_nodes = calloc(N+1, sizeof(temp_nodes));  /*  is initialized in HK  */
    int *bc = calloc(L+1, sizeof(bc));
    int **bonds;

    for(i=0; i<L+1; i++)
	bc[i] = i+1;
    bc[L-1] = 0;
    bc[L] = 1;

    if(!HKV(updated, L)){
	/*  There is no spanning cluster  */

	free(nodes);
	free(temp_nodes);
	free(bc);
	return NULL;
    }

    
    // Set bonds to contain all the bonds in the fracture, and
    // get nodes from running HK on the fracture.
    bonds = getFracture(updated, L);
    int **temp = bonds;
    bonds = removeDanglingArms(bonds, L);
    HK(bonds, nodes, L);

    for(i=0; i<L*L; i++)
	free(temp[i]);
    free(temp);

    bond_removed = 1;
    arm_removed = 1;


    while(arm_removed || bond_removed){

    	bond_removed = 0;
    	arm_removed = 0;

    	//Run through all nodes
    	for(i=0; i<L+1; i++){
    	    for(j=0; j<L; j++){

    		n = i*L+j;
		an = (i < L && bonds[n][1])           ? 1 : 0;
		rn = (i < L && bonds[ i*L+bc[j] ][0]) ? 1 : 0;
		bn = (i > 0 && bonds[n-L][1])         ? 1 : 0;
		ln = (i > 0 && bonds[n-L][0])         ? 1 : 0;

		neighbours = an+bn+rn+ln;
		spanning_cluster = nodes[n];


		if(neighbours == 1){  

		    //Only one neighbour => it's an arm.
		    //Cut the bond, remove the node
 		    if(i < L){
			bonds[n][1] = 0;
			bonds[ i*L+bc[j] ][0] = 0;
		    }
		    if(i > 0){
			bonds[n-L][1] = 0;
			bonds[n-L][0] = 0;
		    }
		    nodes[n] = 0;
		    bond_removed = 1;
		    //printf("Removed node %d\n", n);

		} // end if neighbours == 1
		
		else if(neighbours > 1){
		    
		    // At least two neighbours. Remove bonds+node, run HK
		    // to remove dangling arms from the nodes
		    //printf("Removing bonds around node %d\n", n);
		    
		    if(i < L){
			bonds[n][1] = 0;
			bonds[ i*L+bc[j] ][0] = 0;
		    }
		    if(i > 0){
			bonds[n-L][1] = 0;
			bonds[n-L][0] = 0;
		    }
		    nodes[n] = 0;
		    

		    HK(bonds, temp_nodes, L);
		    broken_network = HKV(bonds, L);

		    nodes_check = 1;
		    for(k=0; k<N; k++)
		    	if(temp_nodes[k] != nodes[k] && k != n)
		    	    nodes_check = 0;

		    if(!broken_network || nodes_check){
			// If !broken_network, we removed a single link node,
			// and there's no fracture in the network anymore. 
			// If nodes_check == 1, the node removed belonged to the backbone.
			// Put things back the way they were
			
			//printf("Putting back bonds around node %d\n", n);
			if(i < L){
			    bonds[n][1] = an;
			    bonds[ i*L+bc[j] ][0] = rn;
			}
			if(i > 0){
			    bonds[n-L][1] = bn;
			    bonds[n-L][0] = ln;
			}
			nodes[n] = spanning_cluster;
			
		    } //end if !broken_network || nodes_check
		    
		    else{  // Removed an arm, update bonds and nodes

			/* printf("Removed an arm when removing node %d. Printing nodes+temp_nodes\n", n); */

			/* printf("Nodes:\n"); */
			/* for(k=L; k>=0; k--){ */
			/*     for(l=0; l<L; l++) */
			/* 	printf("%d ", nodes[k*L+l]); */
			/*     printf("\n"); */
			/* } */
			/* printf("\n"); */
			/* printf("Temp nodes:\n"); */
			/* for(k=L; k>=0; k--){ */
			/*     for(l=0; l<L; l++) */
			/* 	printf("%d ", temp_nodes[k*L+l]); */
			/*     printf("\n"); */
			/* } */
			/* printf("\n"); */

			// Loop through and remove all bonds around nodes not
			// belonging to the spanning cluster

			for(k=0; k<L+1; k++)
			    for(l=0; l<L; l++){
				m = k*L + l;
				nodes[m] = temp_nodes[m];
				if(nodes[m] == 0){
				    if(k < L){
					bonds[m][1] = 0;
					bonds[ k*L+bc[l] ][0] = 0;
				    }
				    if(k > 0){
					bonds[m-L][1] = 0;
					bonds[m-L][0] = 0;
				    }
				}
			    }


			// Have removed bonds to arms. However, the node might still be connected
			// to the backbone, e.g. if it has three neighbouring nodes belonging to
			// the cluster, of which only one is a dangling arm. Check if the
			// neighbouring nodes belong to the spanning cluster (nodes[x] > 0)
			// and if the bond between the nodes was originally updated (xn == 1).
			// If that's the case, set the node to belong to the cluster, and
			// update the bond.

			int lcl, rcl, bcl, acl;  
			lcl = (i > 0 && ((j > 0 && nodes[n-L-1]) || (j==0 && nodes[n-1])) && ln);
			rcl = (i < L && nodes[i*L+L+bc[j]] && rn);
			bcl = (i < L && nodes[n-L] && bn);
			acl = (i > 0 && nodes[n+L] && an);


			if(acl){
			    bonds[n][1] = 1;
			    nodes[n] = spanning_cluster;
			}
			if(rcl){
			    bonds[i*L+bc[j]][0] = 1;
			    nodes[n] = spanning_cluster;
			}
			if(bcl){
			    bonds[n-L][1] = 1;
			    nodes[n] = spanning_cluster;
			}
			if(lcl){
			    bonds[n-L][0] = 1;
			    nodes[n] = spanning_cluster;
			}

			arm_removed = 1;
		    }  // end else (removing arm)
		}  // end else if neighbours == 2

    	    }  //end for(j)
	}  //end for(i)
    }


    free(nodes);
    free(temp_nodes);
    free(bc);
    
    return bonds;

}



int burntBonds(struct System *sys){
    int burnt_bonds = 0;
    for(int i=0; i<sys->L*sys->L; i++)
	for(int j=0; j<2; j++)
	    if(sys->conductance[i][j] < 1e-5)
		burnt_bonds++;
    return burnt_bonds;
}



void removeArms(double **conductance, int L){

    /*  Removes arms that are not belonging to the spanning cluster  */
    /*  Also remove non-spanning clusters                            */
    /*  Uses the fmin function from GCC                              */
    /*  Not completely sure if this works properly with boundaries,  */
    /*  more testing required                                        */

    int removed=0, upper=0, lower=0, left=0, right=0;
    int *il = (int*)calloc(L, sizeof(int));
    int *ir = (int*)calloc(L, sizeof(int));
    int ij;

    for(int i=0; i<L; i++){
	il[i] = i-1;
	ir[i] = i+1;
    }
    il[0] = L-1;
    ir[L-1] = 0;


    /*  For first row, j=0  */
    for(int i=0; i<L; i++){
	int j=0;
	ij = i+j*L;
	if(conductance[ij][0] < 1e-5){
	    left = (int) conductance[il[i]+j*L][1];
	    right = (int) fmin( conductance[ij][1] ,
				fmin( conductance[i+ir[j]*L][1] ,
				      conductance[ir[i]+ir[j]*L][0]));
	    if(left+right){
		conductance[ij][0] = 1;
		removed = 1;
		//	    printf("Removing c(%d,%d)[0]\n", i+1, j+1);
	    }
	}

	if(conductance[ij][1] < 1e-5){
	    lower = (int) conductance[ir[i]+j*L][0];
	    upper = (int) fmin( conductance[ij][0] ,
				fmin( conductance[i+ir[j]*L][1] ,
				      conductance[ir[i]+ir[j]*L][0]));
	    if(lower+upper){
		conductance[ij][1] = 1;
		removed = 1;
		//	    printf("Removing c(%d,%d)[1]\n", i+1, j+1);
	    }
	}
    }
    
    for(int j=1; j<L-1; j++)
	for(int i=0; i<L; i++){
	    ij = i+j*L;
	    if(conductance[ij][0] < 1e-5){
		left = (int) fmin( conductance[il[i]+il[j]*L][0] ,
				   fmin( conductance[il[i]+il[j]*L][1] ,
					 conductance[il[i]+j*L][1]));
		right = (int) fmin( conductance[ij][1] ,
				    fmin( conductance[i+ir[j]*L][1] ,
					  conductance[ir[i]+ir[j]*L][0]));
		if(left+right){
		    conductance[ij][0] = 1;
		    removed = 1;
		    //	    printf("Removing c(%d,%d)[0]\n", i+1, j+1);
		}
	    }

	    if(conductance[ij][1] < 1e-5){
		lower = (int) fmin( conductance[i+il[j]*L][0] ,
				    fmin( conductance[i+il[j]*L][1] ,
					  conductance[ir[i]+j*L][0]));
		upper = (int) fmin( conductance[ij][0] ,
				    fmin( conductance[i+ir[j]*L][1] ,
					  conductance[ir[i]+ir[j]*L][0]));
		if(lower+upper){
		    conductance[ij][1] = 1;
		    removed = 1;
		    //	    printf("Removing c(%d,%d)[1]\n", i+1, j+1);
		}
	    }
	}
    

    /*  Last row, j = L-1  */
    for(int i=0; i<L; i++){
	int j = L-1;
	ij = i+j*L;
	if(conductance[ij][0] < 1e-5){
	    left = (int) fmin( conductance[il[i]+il[j]*L][0] ,
			       fmin( conductance[il[i]+il[j]*L][1] ,
				     conductance[il[i]+j*L][1]));
	    right = (int) conductance[ij][1];

	    if(left+right){
		conductance[ij][0] = 1;
		removed = 1;
		//	    printf("Removing c(%d,%d)[0]\n", i+1, j+1);
	    }
	}

	if(conductance[ij][1] < 1e-5){
	    lower = (int) fmin( conductance[i+il[j]*L][0] ,
				fmin( conductance[i+il[j]*L][1] ,
				      conductance[ir[i]+j*L][0]));
	    upper = (int) conductance[ij][0];

	    if(lower+upper){
		conductance[ij][1] = 1;
		removed = 1;
		//	    printf("Removing c(%d,%d)[1]\n", i+1, j+1);
	    }
	}
    }

    free(ir);
    free(il);

    if(removed){
	//	printf("Running removeArms() again\n");
	removeArms(conductance, L);
    }


}



double getWidth(int **bonds, int L){
    int spanning_cluster, num_bonds, temp, in_column;
    double width, y, y_sq;

    /*  19.09 - TODO: Fix so that spanning_cluster actually is the spanning cluster!  */
    /*  20.09 - Should work now                                                       */

    y = 0;
    y_sq = 0;
    num_bonds = 0;
    spanning_cluster = 0;

    for(int i=0; i<L; i++){
	if(bonds[i*L][0] >= 2){
	    spanning_cluster = bonds[i*L][0];
	    /* printf("Set spanning cluster to %d, at %d\n", spanning_cluster, i*L); */

	    for(int k=0; k<L; k++){  /* check column-wise */
		in_column = 0;
		for(int l=0; l<L; l++)
		    if(bonds[k+L*l][0] == spanning_cluster){
			in_column = 1;
			break;
		    }
		if(!in_column){
		    spanning_cluster = 0;
		    break;
		}  /* end if */
	    }  /* end for(k) */	    
	}  /* end if */
	if(spanning_cluster)  /* break out of loop, no need to check the rest of the column */
	    break;
    }
    


    for(int i=0; i<L*L; i++){
	if(bonds[i][0] == spanning_cluster){
	    temp = i/L+1;
	    y += temp;
	    y_sq += temp*temp;
	    num_bonds++;
	}
	if(bonds[i][1] == spanning_cluster){
	    temp = i/L+1;
	    y += temp;
	    y_sq += temp*temp;
	    num_bonds++;
	}
    }


    width = sqrt(y_sq/num_bonds - y*y/(num_bonds*num_bonds));
    /* printf("num_bonds = %d, spanning cluster = %d\n", num_bonds, spanning_cluster); */

    assert(num_bonds >= 2*L);
    
    return width;
}



int **shortestPath(double **threshold, int L){

    
    /*  Implements the Floyd-Warshall algorithm. The matrices involved are sparse, so     */
    /*  using something like Johnson's algorithm for sparse graphs is better, but         */
    /*  Floyd-Warshall is used as it is easier to implement.                              */
    /*  The matrix D is the adjacency matrix, giving the shortest path between 'nodes' m. */
    /*  It should be symmetric when finished, as the graph is esentially undirected.      */
    /*  The predecessor matrix, however, is not symmetric.                                */
    /*  The 'nodes' used here are the spaces surrounded by bonducting bonds, and not      */
    /*  the nodes connecting the bonds, which are used to find the voltage differece,     */
    /*  i.e. one 'node' m as used here has four voltage nodes n around it, with           */
    /*  conducting bonds cond[n][0] between n and n+L and cond[n][1] between n and n+L+1. */
	
    /*               	                /|n=m+L+1                                         */
    /* 	                               / |                                                */
    /* 	                              /  |                                                */
    /*                            n=m/ m |n=m+1                                           */
    /* 	                             |   /                                                */
    /*                               |  /                                                 */
    /*                    	     | /                                                  */
    /*                  	     |/n=m-L                                              */
 
      
    int cols = L+L/2;       // Number of columns to be used in the modified system to find the shortest path
    int N = (L+1)*cols;  /*  Include 'nodes' below the upper electrode. These were ignored in the CG  */
    int start, stop, min_i, min_j;
    double min;
    double *temp_D;
    int *temp_p;
    double *D = calloc(N*N, sizeof(D));
    double *D_old = calloc(N*N, sizeof(D_old));
    int *predecessor = calloc(N*N, sizeof(predecessor));
    int *predecessor_old = calloc(N*N, sizeof(predecessor_old));
    int *bc = calloc(cols+1, sizeof(bc));
    int **path;

    
    // Initialize bc, is used to find the correct threshold index
    for(int i=0; i<L; i++)
	bc[i] = i;
    for(int i=L; i<cols+1; i++)
	bc[i] = i-L;
    

    //TODO: Fix so it works for several columns across the boundary
    //      Have to look into how neighbours are selected 
    //      Change N and all the shit that depends on it
    //      Add/copy columns on the right side


    // Initialize D and predecessor matrices
    // Faster to initialize D and pre to DBL_MAX/0 first,
    // then just initialize weights for neighbours? Avoids branching.
    for(int i=0; i<L+1; i++)         // row
	for(int j=0; j<cols; j++){   // column
	    int m = i*cols + j;      // m will have values ranging from 0 to N
	    int ti = i*L + bc[j];    // Threshold index. Will have values ranging from 0 to L*L
	    int tir = i*L + bc[j+1];
	    for(int n=0; n<N; n++){
		
		
		if(n == m + cols + 1 && j < cols-1){  // upper right neighbour
		    D[m*N + n] = threshold[tir][0];   // column furthest to the right
		    predecessor[m*N + n] = m;         // (j == cols-1) has no right neighbour
		}
		else if(n == m + cols){  // upper neighbour
		    D[m*N + n] = threshold[ti][1];
		    predecessor[m*N + n] = m;
		}
		else if(n == m - cols){  // lower neighbour
		    D[m*N + n] = threshold[ti-L][1];
		    predecessor[m*N + n] = m;
		}
		else if(n == m - cols - 1 && j > 0){  // lower left neighbour
		    D[m*N + n] = threshold[ti-L][0];  // column furthest to the left
		    predecessor[m*N + n] = m;         // (j == 0) has no left neighbour
		}
		else if(n == m){
		    D[m*N + n] = 0;
		    predecessor[m*N + n] = 0;
		}
		else{
		    D[m*N + n] = DBL_MAX;
		    predecessor[m*N + n] = 0;
		}
	    }
	}


    // Calculate shortest path weights and predecessors according
    // to Floyd-Warshall.
    for(int k=0; k<N; k++){
	
	temp_D = D;
	D = D_old;
	D_old = temp_D;
	temp_p = predecessor;
	predecessor = predecessor_old;
	predecessor_old = temp_p;

	for(int m=0; m<N; m++)
	    for(int n=0; n<N; n++){
		
		if(D_old[m*N + n] <= D_old[k*N + n] + D_old[m*N + k]){
		    D[m*N + n] = D_old[m*N + n];
		    predecessor[m*N + n] = predecessor_old[m*N + n];
		}
		else{
		    D[m*N + n] = D_old[k*N + n] + D_old[m*N + k];
		    predecessor[m*N + n] = predecessor_old[k*N + n];
		}
	    }
    }


    // Find the node in the first column that have the shortest path spanning
    // around to itself. start is the node in column 0, while stop is the
    // corresponding node in column L. min_i and min_j are then stored to be used
    // as starting indices when finding the path from the predecessor matrix in
    // getShortestPath().
    min_i = 0;
    min_j = 0;
    min = DBL_MAX;
    for(int i=0; i<L+1; i++){
	start = i*cols;
	stop = i*cols + L;

	if(D[start*N + stop] < min){
	    min = D[start*N + stop];
	    min_i = start;
	    min_j = stop;
	}
    }


    path = getShortestPath(predecessor, min_i, min_j, L);

    free(bc);
    free(D);
    free(D_old);
    
    free(predecessor);
    free(predecessor_old);

    return path;
}



int **getShortestPath(int *predecessor, int min_i, int min_j, int L){
    
    /*  Given the predecessor array(matrix), system size L and the row and column  */
    /*  i and j (i.e. the coordinate) indicating where to start in the predecessor */
    /*  matrix. Returns the coordinates of all bonds included in the shortest path */
    /*  in the matrix bonds. bonds[i][0] contain the array index of the bond       */
    /*  (voltage node) coordinate, while bonds[i][1] contain which bond going     */
    /*  out from the node is in the shortest path (i.e. 0 or 1).  */

    int cols = L+L/2;
    int **bonds;
    int *bc = calloc(cols, sizeof(bc));
    int N = (L+1)*cols;
    int curr_node, next_node, final_node, row, a, b, counter, size, sign, diff;

    /*  Set size to the first power of two larger or equal to 2*(L+1), which is  */
    /*  the minimum number of bonds broken. Algorithm from Bit Twiddling Hacks.  */
    size = 2*L+1;
    size |= size >> 1;
    size |= size >> 2;
    size |= size >> 4;
    size |= size >> 8;
    size |= size >> 16;
    size++;

    bonds = calloc(size, sizeof(bonds));
    for(int i=0; i<size; i++)
	bonds[i] = calloc(2, sizeof(bonds[i]));
    
    for(int i=0; i<L; i++)
	bc[i] = i;
    for(int i=L; i<cols; i++)
	bc[i] = i-L;


    a = !(cols & 1);    // a=0 if L+L/2 is odd, 1 if L+L/2 is even
    final_node = min_j;
    curr_node = min_i;
    next_node = predecessor[curr_node + N*final_node];
    counter = 0;


    while(next_node != final_node){
	diff = next_node-curr_node;
	b = diff & 1;    // b=1 if next_node-curr_node is odd, 0 if even

	sign = -(diff < 0);    // sign = -1 if next_node lies on the row 
	                       // below the current node, 0 otherwise

	// The row of the voltage node where the bond connecting
	// the nodes originates. Same as the row of the current node
	// if the next node is above or to the right. If the next
	// node is below or to the left, the bond originates in the row below       
	row = curr_node/cols + sign;  

	// bond (voltage node) coordinate is given by the row (row*L)
	// + the column. This is the same as the column of the current
	// node (curr_node%cols) for next_node to the left/above/below.
	// In the case of next_node being to the right, the column
	// is given by (curr_node%cols) + 1. This is the case if a^b == sign.
	// sign = -1 if next_node is below/to the left, 0 otherwise.
	// a^b = 0 if next_node is to the right/left, 1 otherwise.
	// => (a^b)==sign = 1 if next_node is to the right, 0 otherwise.
	bonds[counter][0] = row*L + bc[ curr_node%cols + ((a^b)==sign)];
	bonds[counter][1] = a^b;

	curr_node = next_node;
	next_node = predecessor[curr_node + N*final_node];
	counter++;
	if(counter+2 == size){
	    size <<= 1;
	    bonds = realloc(bonds, size*sizeof(bonds));

	    if(!bonds){
		printf("Something wrong happened in getShortestPath()! bonds array is pointing to NULL!\n");
		exit(EXIT_FAILURE);
	    }
	    for(int i=size>>1; i<size; i++)
		bonds[i] = calloc(2, sizeof(bonds[i]));
	}
    }

    // next_node == final_node
    diff = next_node-curr_node;
    b = diff & 1;  // b=1 if curr_node-next_node is odd, 0 if even
    
    sign = -(diff < 0);
    row = curr_node/cols + sign;
    bonds[counter][0] = row*L + bc[ curr_node%cols + ((a^b)==sign)];
    bonds[counter][1] = a^b;
    

    bonds[counter+1][0] = size;         /*  Gives the size of bonds for proper freeing of memory  */
    bonds[counter+1][1] = -1;           /*  Denotes that this is the last element of bonds        */

    free(bc);
    return bonds;
}



inline double minNeighbour3(double s1, double s2, double s3, int *k){

    double m;
    
    *k = (s1 < s2) ? 1 : 2;
    m = (s1 < s2) ? s1 : s2;
    *k = (s3 < m) ? 3 : *k;
    m = (s3 < m) ? s3 : m;

    return m;
}



inline double minNeighbour4(double s1, double s2, double s3, double s4, int *k){

    double m;

    *k = (s1 < s2) ? 1 : 2;    
    m = (s1 < s2) ? s1 : s2;
    *k = (s3 < m) ? 3 : *k;
    m = (s3 < m) ? s3 : m;
    *k = (s4 < m) ? 4 : *k;
    m = (s4 < m) ? s4 : m;

    return m;
}
void primeOptimalPathscapeNoSum(double **restrict nodes, 
				double **restrict thresholds, 
				int **bonds, int L, double tol, int col){

    /*------------------------------------------------------------------------ */
    /*  Primes the optimal pathscape to speed up the calculation of the final  */
    /*  pathscape. Takes in nodes, thresholds, bonds system size L and         */
    /*  tolerance for when to cut of the while loop.                           */
    /*  Nodes contain all the nodes with their lowest total threshold for      */
    /*  connecting with a given edge node.                                     */
    /*  Thresholds contains all the bond thresholds.                           */
    /*  Bonds contain the bonds that must connects a given node to the path    */
    /*  with the lowest thresholds.                                            */
    /*------------------------------------------------------------------------ */

    
    int N = 2*L;
    int l = L/2;
    int n, t;
    double updated_node, old_val, s1, s2, s3, s4;

    updated_node = 1;
    
    // sx are the sum of node x and the thresholds of the bond between
    // node n and x. x = 1,2,3,4 denotes the neighbour nodes to node n,
    // with 1 = (lower) left node, 2 = below, 3 = above, 4 = (upper) right
    enum{ LEFT = 1, BELOW, ABOVE, RIGHT};
    
    //Assume col == 1 or 2!

    int base_node = col == 1 ? 0 : L+1;
    int zero_node = col == 1 ? l : L+l;
    //int zero_node = (2*(L+1)-L-2)*N+L+l;
    

    if(col < 0){
    int runs = 0;
    while(updated_node){
    	updated_node = 0;
    	runs++;
    	nodes[base_node][zero_node] = 0;
    	// j == 0, bottom row
    	for(int k=0; k<N-1; k++){

    	    old_val = nodes[base_node][ k ];
    	    s3 = nodes[base_node][ k+N   ] + thresholds[ k   ][1];
    	    s4 = nodes[base_node][ k+N+1 ] + thresholds[ k+1 ][0];
    	    nodes[base_node][ k ] = (s3 < s4) ? s3 : s4;
    	    bonds[base_node][ k ] = (nodes[base_node][ k ] > 0) * (s3 < s4 ? ABOVE : RIGHT);
		
    	    nodes[base_node][zero_node] = 0;
    	    //nodes[base_node][l] = 0;

    	    updated_node += (fabs(old_val - nodes[base_node][ k ]) > tol);

    	}

    	// j == 0, k == N-1

    	old_val = nodes[base_node][ N-1 ];
    	nodes[base_node][ N-1 ] = (nodes[base_node][ 2*N-1 ] + thresholds[ N-1 ][1]);// *
    	bonds[base_node][ N-1 ] = (nodes[base_node][ N-1 ] > 0) * ABOVE;

    	nodes[base_node][zero_node] = 0;
    	//nodes[base_node][l] = 0;

    	updated_node += (fabs(old_val - nodes[base_node][ N-1 ]) > tol);

    	for(int j=1; j<L; j++){

    	    // k == 0
    	    n = j*N;

    	    old_val = nodes[base_node][ n ];
    	    s2 = nodes[base_node][ j*N-N   ] + thresholds[ j*N-N ][1];
    	    s3 = nodes[base_node][ j*N+N   ] + thresholds[ j*N   ][1];
    	    s4 = nodes[base_node][ j*N+N+1 ] + thresholds[ j*N+1 ][0];
    	    nodes[base_node][ n ] = minNeighbour3(s2, s3, s4, &t);

    	    nodes[base_node][zero_node] = 0;
    	    //nodes[base_node][l] = 0;

    	    bonds[base_node][ n ] = (nodes[base_node][ n ] > 0) * (t+1);
    	    updated_node += (fabs(old_val - nodes[base_node][ n ]) > tol);
		

    	    for(int k=1; k<N-1; k++){

    		n = j*N+k;
    		old_val = nodes[base_node][ n ];
    		s1 = nodes[base_node][ j*N+k-N-1 ] + thresholds[ j*N+k-N ][0];
    		s2 = nodes[base_node][ j*N+k-N   ] + thresholds[ j*N+k-N ][1];
    		s3 = nodes[base_node][ j*N+k+N   ] + thresholds[ j*N+k   ][1];
    		s4 = nodes[base_node][ j*N+k+N+1 ] + thresholds[ j*N+k+1 ][0];
		    
    		nodes[base_node][ n ] = minNeighbour4(s1, s2, s3, s4, &t);

    		nodes[base_node][zero_node] = 0;
    		//nodes[base_node][l] = 0;

    		bonds[base_node][ n ] = (nodes[base_node][ n ] > 0) * t;
    		updated_node += (fabs(old_val - nodes[base_node][ n ]) > tol);


    	    }

    	    // k == N-1
    	    n = j*N+N-1;
    	    old_val = nodes[base_node][ n ];
    	    s1 = nodes[base_node][ j*N-2     ] + thresholds[ j*N-1   ][0];
    	    s2 = nodes[base_node][ j*N+-1    ] + thresholds[ j*N-1   ][1];
    	    s3 = nodes[base_node][ j*N+2*N-1 ] + thresholds[ j*N+N-1 ][1];
    	    nodes[base_node][ n ] = minNeighbour3(s1, s2, s3, &t);

    	    nodes[base_node][zero_node] = 0;
    	    //nodes[base_node][l] = 0;

    	    bonds[base_node][ n ] = (nodes[base_node][ n ] > 0) * t;
    	    updated_node += (fabs(old_val - nodes[base_node][ n ]) > tol);


    	}

    	// j == L
    	// k == 0
    	old_val = nodes[base_node][ L*N ];
    	nodes[base_node][ L*N ] = (nodes[base_node][ L*N-N   ] + thresholds[ L*N-N ][1]);

    	nodes[base_node][zero_node] = 0;
    	//nodes[base_node][l] = 0;

    	bonds[base_node][ L*N ] = (nodes[base_node][ L*N ] > 0) * BELOW;
    	updated_node += (fabs(old_val - nodes[base_node][ L*N ]) > tol);


    	for(int k=1; k<N; k++){

    	    n = L*N+k;
    	    old_val = nodes[base_node][ n ];
    	    s1 = nodes[base_node][ L*N+k-N-1 ] + thresholds[ L*N+k-N ][0];
    	    s2 = nodes[base_node][ L*N+k-N   ] + thresholds[ L*N+k-N ][1];
    	    nodes[base_node][ n ] = (s1 < s2) ? s1 : s2;

    	    nodes[base_node][zero_node] = 0;
    	    //nodes[base_node][l] = 0;

    	    bonds[base_node][ n ] = (nodes[base_node][ n ] > 0) * (s1 < s2 ? LEFT : BELOW);
    	    updated_node += (fabs(old_val - nodes[base_node][ n ]) > tol);


    	    //}
    	}
    }
    
    if(col == 1)
    	for(int i=1; i<L+1; i++)
    	    for(int j=0; j<N*(L+1); j++){
    		nodes[i][j] = nodes[base_node][j];
    		bonds[i][j] = bonds[base_node][j];
    	    }
    else if(col == 2)
    	for(int i=L+2; i<2*(L+1); i++)
    	    for(int j=0; j<N*(L+1); j++){
    		nodes[i][j] = nodes[base_node][j];
    		bonds[i][j] = bonds[base_node][j];
    	    }
    else{
    	printf("primePathscape: col != 1 or 2! Exiting\n");
    	assert(col == 1);
    }

    }

    else
	for(int j=0; j<N*(L+1); j++){
	    nodes[col+1][j] = nodes[col][j];
	    bonds[col+1][j] = bonds[col][j];
	}


}



void primeOptimalPathscape(double **restrict nodes, 
			   double **restrict thresholds, 
			   int **bonds, int L, double tol){

    /*------------------------------------------------------------------------ */
    /*  Primes the optimal pathscape to speed up the calculation of the final  */
    /*  pathscape. Takes in nodes, thresholds, bonds system size L and         */
    /*  tolerance for when to cut of the while loop.                           */
    /*  Nodes contain all the nodes with their lowest total threshold for      */
    /*  connecting with a given edge node.                                     */
    /*  Thresholds contains all the bond thresholds.                           */
    /*  Bonds contain the bonds that must connects a given node to the path    */
    /*  with the lowest thresholds.                                            */
    /*------------------------------------------------------------------------ */

    
    int N = 2*L;
    int l = L/2;
    int zero_cols[2] = {l, l+L};
    int n, t;
    double sum, old_sum, s1, s2, s3, s4;
    sum = 0;
    old_sum = 1;
    
    // sx are the sum of node x and the thresholds of the bond between
    // node n and x. x = 1,2,3,4 denotes the neighbour nodes to node n,
    // with 1 = (lower) left node, 2 = below, 3 = above, 4 = (upper) right
    enum{ LEFT = 1, BELOW, ABOVE, RIGHT};

    while(fabs(sum-old_sum) > tol){
	old_sum = sum;
	sum = 0;


	// j == 0, bottom row
	for(int k=0; k<N-1; k++){
	    s3 = nodes[0][ k+N   ] + thresholds[ k   ][1];
	    s4 = nodes[0][ k+N+1 ] + thresholds[ k+1 ][0];
	    nodes[0][ k ] = (s3 < s4) ? s3 : s4;
	    nodes[0][ k ] *= ((k != zero_cols[0]) && (k != zero_cols[1]));
	    bonds[0][ k ] = (nodes[0][ k ] > 0) * (s3 < s4 ? ABOVE : RIGHT);

	    sum += nodes[0][ k ];
	}

	// j == 0, k == N-1
	nodes[0][ N-1 ] = (nodes[0][ 2*N-1 ] + thresholds[ N-1 ][1]) * 
	    (((N-1) != zero_cols[0]) && ((N-1) != zero_cols[1])); 
	bonds[0][ N-1 ] = (nodes[0][ N-1 ] > 0) * ABOVE;
	sum += nodes[0][ N-1 ];

	for(int j=1; j<L; j++){

	    // k == 0
	    n = j*N;
	    s2 = nodes[0][ j*N-N   ] + thresholds[ j*N-N ][1];
	    s3 = nodes[0][ j*N+N   ] + thresholds[ j*N   ][1];
	    s4 = nodes[0][ j*N+N+1 ] + thresholds[ j*N+1 ][0];
	    nodes[0][ n ] = minNeighbour3(s2, s3, s4, &t) * ((0 != zero_cols[0]) && (0 != zero_cols[1]));
	    bonds[0][ n ] = (nodes[0][ n ] > 0) * (t+1);
	    sum += nodes[0][ n ];
		
	    for(int k=1; k<N-1; k++){

		n = j*N+k;
		s1 = nodes[0][ j*N+k-N-1 ] + thresholds[ j*N+k-N ][0];
		s2 = nodes[0][ j*N+k-N   ] + thresholds[ j*N+k-N ][1];
		s3 = nodes[0][ j*N+k+N   ] + thresholds[ j*N+k   ][1];
		s4 = nodes[0][ j*N+k+N+1 ] + thresholds[ j*N+k+1 ][0];
		    
		nodes[0][ n ] = minNeighbour4(s1, s2, s3, s4, &t) * ((k != zero_cols[0]) && (k != zero_cols[1]));
		bonds[0][ n ] = (nodes[0][ n ] > 0) * t;
		sum += nodes[0][ n ];
	    }

	    // k == N-1
	    n = j*N+N-1;
	    s1 = nodes[0][ j*N-2     ] + thresholds[ j*N-1   ][0];
	    s2 = nodes[0][ j*N+-1    ] + thresholds[ j*N-1   ][1];
	    s3 = nodes[0][ j*N+2*N-1 ] + thresholds[ j*N+N-1 ][1];
	    nodes[0][ n ] = minNeighbour3(s1, s2, s3, &t) * (((N-1) != zero_cols[0]) && ((N-1) != zero_cols[1]));
	    bonds[0][ n ] = (nodes[0][ n ] > 0) * t;
	    sum += nodes[0][ n ];

	}

	// j == L
	// k == 0
	nodes[0][ L*N ] = (nodes[0][ L*N-N   ] + thresholds[ L*N-N ][1]) * 
	    ((0 != zero_cols[0]) && (0 != zero_cols[1]));
	bonds[0][ L*N ] = (nodes[0][ L*N ] > 0) * BELOW;
	sum += nodes[0][ L*N ];

	for(int k=1; k<N; k++){
	    n = L*N+k;
	    s1 = nodes[0][ L*N+k-N-1 ] + thresholds[ L*N+k-N ][0];
	    s2 = nodes[0][ L*N+k-N   ] + thresholds[ L*N+k-N ][1];
	    nodes[0][ n ] = (s1 < s2) ? s1 : s2;
	    nodes[0][ n ] *= ((k != zero_cols[0]) && (k != zero_cols[1]));
	    bonds[0][ n ] = (nodes[0][ n ] > 0) * (s1 < s2 ? LEFT : BELOW);
	    sum += nodes[0][ n ];
	}
    }
    
    for(int i=1; i<2*(L+1); i++)
	for(int j=0; j<N*(L+1); j++){
	    nodes[i][j] = nodes[0][j];
	    bonds[i][j] = bonds[0][j];
	}

}


void calculateOptimalPathscapeNEWTHATDIDNTREALLYHELP(double **restrict nodes, 
			       double **restrict temp_thresholds, 
			       int **bonds, int L, double tol, int nthreads, int row){
    /*********************************/
    /*    FUNCTION IS DEPRECATED     */
    /*********************************/

    int N = 2*L;
    int l = L/2;


    enum{ LEFT = 1, BELOW, ABOVE, RIGHT};

    //if(nthreads > 1)
    //nthreads = 2;
    //for(row = 0; row < L+1; row++)
    //#pragma omp parallel for num_threads(nthreads)
    for(int i=0; i<2; i++){

	double sum, old_sum, s1, s2, s3, s4;
	int i_node, n, t;

	sum = 0;
	old_sum = 1;
	
	// i_node should not be updated (locked to zero)
	i_node = (i == 0) ? row*N+l : row*N+L+l;
	nodes[i][i_node] = 0;
	while(fabs(sum-old_sum) > tol){

	    old_sum = sum;
	    sum = 0;
	    

	    // j == 0
	    for(int k=0; k<N-1; k++){
		s3 = nodes[i][ k+N   ] + temp_thresholds[ k   ][1];
		s4 = nodes[i][ k+N+1 ] + temp_thresholds[ k+1 ][0];
		nodes[i][ k ] = (s3 < s4) ? s3 : s4;
		nodes[i][ k ] *= (k != i_node);
		bonds[i][ k ] = (nodes[i][ k ] > 0) * (s3 < s4 ? ABOVE : RIGHT);
		sum += nodes[i][ k ];
	    }

	    // j == 0, k == N-1
	    nodes[i][ N-1 ] = (nodes[i][ 2*N-1 ] + temp_thresholds[ N-1 ][1]) * (N-1 != i_node); 
	    bonds[i][ N-1 ] = (nodes[i][ N-1 ] > 0) * ABOVE;
	    sum += nodes[i][ N-1 ];

	    for(int j=1; j<L; j++){

		// k == 0
		n = j*N;
		s2 = nodes[i][ j*N-N   ] + temp_thresholds[ j*N-N ][1];
		s3 = nodes[i][ j*N+N   ] + temp_thresholds[ j*N   ][1];
		s4 = nodes[i][ j*N+N+1 ] + temp_thresholds[ j*N+1 ][0];
		nodes[i][ n ] = minNeighbour3(s2, s3, s4, &t) * (n != i_node);
		bonds[i][ n ] = (nodes[i][ n ] > 0) * (t+1);
		sum += nodes[i][ n ];
		
		for(int k=1; k<N-1; k++){

		    n = j*N+k;
		    s1 = nodes[i][ j*N+k-N-1 ] + temp_thresholds[ j*N+k-N ][0];
		    s2 = nodes[i][ j*N+k-N   ] + temp_thresholds[ j*N+k-N ][1];
		    s3 = nodes[i][ j*N+k+N   ] + temp_thresholds[ j*N+k   ][1];
		    s4 = nodes[i][ j*N+k+N+1 ] + temp_thresholds[ j*N+k+1 ][0];
		    
		    nodes[i][ n ] = minNeighbour4(s1, s2, s3, s4, &t) * (n != i_node);
		    bonds[i][ n ] = (nodes[i][ n ] > 0) * t;
		    sum += nodes[i][ n ];
		}

		// k == N-1
		n = j*N+N-1;
		s1 = nodes[i][ j*N-2     ] + temp_thresholds[ j*N-1   ][0];
		s2 = nodes[i][ j*N+-1    ] + temp_thresholds[ j*N-1   ][1];
		s3 = nodes[i][ j*N+2*N-1 ] + temp_thresholds[ j*N+N-1 ][1];
		nodes[i][ n ] = minNeighbour3(s1, s2, s3, &t) * (n != i_node);
		bonds[i][ n ] = (nodes[i][ n ] > 0) * t;
		sum += nodes[i][ n ];

	    }

	    // j == L
	    // k == 0
	    nodes[i][ L*N ] = (nodes[i][ L*N-N   ] + temp_thresholds[ L*N-N ][1]) * (L*N != i_node);
	    bonds[i][ L*N ] = (nodes[i][ L*N ] > 0) * BELOW;
	    sum += nodes[i][ L*N ];

	    for(int k=1; k<N; k++){
		n = L*N+k;
		s1 = nodes[i][ L*N+k-N-1 ] + temp_thresholds[ L*N+k-N ][0];
		s2 = nodes[i][ L*N+k-N   ] + temp_thresholds[ L*N+k-N ][1];
		nodes[i][ n ] = (s1 < s2) ? s1 : s2;
		nodes[i][ n ] *= (n != i_node);
		bonds[i][ n ] = (nodes[i][ n ] > 0) * (s1 < s2 ? LEFT : BELOW);
		sum += nodes[i][ n ];
	    }
	}

	// As source nodes close to each other have similar distances to
	// the other nodes, the next run through starts from the results
	// of this one. No need for priming pathscape, as this is much
	// faster (factor of ~15 for 16x16). 
	// Need to adjust this for 4 threads
	// ASSUMPTION: The first call to calculateOptimalPathscape
	// is for row 0, followed by rows 1, 2, ..., L+1.
	if(i == 0 && nthreads == 1 && row == 0)
	    for(int j=0; j<N*(L+1); j++){
		nodes[1][j] = nodes[0][j];
		bonds[1][j] = bonds[0][j];
	    }
    }
}



void optimalPathscapeNEWTHATDIDNTREALLYHELP(struct System *sys){
    
    /*********************************/
    /*    FUNCTION IS DEPRECATED     */
    /*********************************/

    /*-------------------------------------------------------------------------*/
    /*  Calculates the optimal pathscape for the system using the approach in  */
    /*  A. Hansen and J. Kertesz, Phys. Rev. Lett 93, 040601 (2004).           */
    /*                                                                         */
    /*  Assumes a system of size Lx(L+1) with L even!                          */
    /*-------------------------------------------------------------------------*/

    int L = sys->L, N = 2*L, l = L/2;
    int k1, k2,
	*conv_col,
	**bonds, **temp_bonds0, **temp_bonds1;
    //***temp_bonds;
    double *sp_nodes, 
	   **nodes, **temp_nodes, **temp_thresholds;

    conv_col = calloc(N, sizeof(conv_col));
    for(int i=0; i<l; i++)
	conv_col[i] = i+l;
    for(int i=l; i<l+L; i++)
	conv_col[i] = i-l;
    for(int i=l+L; i<N; i++)
	conv_col[i] = i-l-L;

    nodes = calloc(2, sizeof(nodes));
    temp_nodes = calloc(L+1, sizeof(temp_nodes));
    //sp_nodes = calloc(L+1, sizeof(sp_nodes)); //N*(L+1)
    sp_nodes = calloc(N*(L+1), sizeof(sp_nodes));
    bonds = calloc(2, sizeof(bonds));
    temp_bonds0 = calloc(L+1, sizeof(temp_bonds0));
    temp_bonds1 = calloc(L+1, sizeof(temp_bonds1));
    temp_thresholds = calloc(N*L, sizeof(temp_thresholds));
    for(int i=0; i<2; i++){
    	nodes[i] = calloc(N*(L+1), sizeof(nodes[i]));
    	bonds[i] = calloc(N*(L+1), sizeof(bonds[i]));
    }
    for(int i=0; i<L+1; i++){
	//sp_nodes[i] = calloc(N*(L+1), sizeof(sp_nodes[i]));
    	temp_nodes[i] = calloc(L*(L+1), sizeof(temp_nodes[i]));
    	temp_bonds0[i] = calloc(L*(L+1), sizeof(temp_bonds0[i]));
    	temp_bonds1[i] = calloc(L*(L+1), sizeof(temp_bonds1[i]));
    }


    for(int i=0; i<L; i++){
    	for(int j=0; j<N; j++){
    	    temp_thresholds[ i*N+j ] = calloc(2, sizeof(temp_thresholds[i*N+j]));
    	    temp_thresholds[ i*N+j ][0] = sys->threshold[ i*L+conv_col[j] ][0];
    	    temp_thresholds[ i*N+j ][1] = sys->threshold[ i*L+conv_col[j] ][1];
    	}
    }

    /* printf("Size of:\n"); */
    /* printf("\t Nodes           \t %lu\n"  */
    /* 	   "\t Temp Nodes      \t %lu\n" */
    /* 	   "\t SP Nodes        \t %lu\n" */
    /* 	   "\t Bonds           \t %lu\n" */
    /* 	   "\t Temp bonds      \t %lu\n" */
    /* 	   "\t Temp thresholds \t %lu\n", */
    /* 	   2*(L+1)*N*(L+1)*sizeof(nodes)/1000000, (L+1)*L*(L+1)*sizeof(temp_nodes)/1000000, N*(L+1)*sizeof(sp_nodes)/1000000, */
    /* 	   2*N*(L+1)*sizeof(bonds)/1000000, (L+1)*L*(L+1)*2*sizeof(temp_bonds0)/1000000, N*L*2*sizeof(temp_thresholds)/1000000); */
    // Extract optimal pathscape to sys

    for(int i=0; i<L+1; i++){

    	// i is the current row that the spanning path is calculated for
    	calculateOptimalPathscapeNEWTHATDIDNTREALLYHELP(nodes, temp_thresholds, bonds, L, 1e-10, sys->nthreads, i);


    	// sp_nodes[i][:] are set to the sum of nodes[0][:] and new_nodes[1][:],
    	// i.e. total threshold for spanning paths going through the different nodes
    	for(int j=0; j<N*(L+1); j++){
    	    sp_nodes[j] = (nodes[0][j] + nodes[1][j]);
    	    // temp_bonds[i][j][0] = bonds[i][j][0]
    	    // temp_bonds[i][j][1] = bonds[i+L+1][j][1]
    	}

    	// temp_nodes[] is of same size as pathscape.
    	// Set node values to the minimum of paths going through columns k and k'
    	for(int j=0; j<L+1; j++)
    	    for(int k=0; k<L; k++){
    		k1 = k+l;
    		k2 = (k < l) ? k+L+l : k-l;

    		temp_bonds0[i][j*L+k] = sp_nodes[j*N+k1] < sp_nodes[j*N+k2] ?
    		    bonds[0][j*N+k1] : bonds[0][j*N+k2];

    		temp_bonds1[i][j*L+k] = sp_nodes[j*N+k1] < sp_nodes[j*N+k2] ?
    		    bonds[1][j*N+k1] : bonds[1][j*N+k2];

    		temp_nodes[i][j*L+k] =
    		    sp_nodes[j*N+k1] < sp_nodes[j*N+k2] ?
    		    sp_nodes[j*N+k1] : sp_nodes[j*N+k2];
    	    }
    }

    // temp_nodes and temp_bonds take a lot of memory.
    // Write directly to pathscape after finding sp_nodes
    // instead of going via temp_nodes and temp_bonds?
    for(int i=0; i<L+1; i++)
    	for(int j=0; j<L*(L+1); j++){
    	    /* if(i == 0) */
    	    /* 	printf("%d\n", j); */
    	    if(fabs(temp_nodes[i][j]) > 1e-6){
		
    		sys->pathscape_bonds[j][0] =
    		    sys->pathscape[j] < temp_nodes[i][j] ?
    		    sys->pathscape_bonds[j][0] : temp_bonds0[i][j];

    		sys->pathscape_bonds[j][1] =
    		    sys->pathscape[j] < temp_nodes[i][j] ?
    		    sys->pathscape_bonds[j][1] : temp_bonds1[i][j];

    		sys->pathscape[j] =
    		    sys->pathscape[j] < temp_nodes[i][j] ?
    		    sys->pathscape[j] : temp_nodes[i][j];
    	    }
    	    else
    	    	printf("Low val for i = %d, j = %d\n", i, j);
    	}

    
    free(conv_col);
    for(int i=0; i<N*L; i++)
	free(temp_thresholds[i]);
    for(int i=0; i<2; i++){
	free(bonds[i]);
	free(nodes[i]);
    }
    for(int i=0; i<L+1; i++){
	/* for(int j=0; j<L*(L+1); j++) */
	/*     free(temp_bonds[i][j]); */
	free(temp_bonds0[i]);
	free(temp_bonds1[i]);
	free(temp_nodes[i]);
	//free(sp_nodes[i]);
    }
    free(sp_nodes);
    free(temp_thresholds);
    free(bonds);
    free(temp_bonds0);
    free(temp_bonds1);
    free(nodes);
    free(temp_nodes);

}



void calculateOptimalPathscapeBlocks(double **restrict nodes, 
			       double **restrict temp_thresholds, 
			       int **bonds, int L, double tol, int nthreads){

    
    int N = 2*L;
    int l = L/2;
    int bsize = 64;

    double *temp_nodes = calloc((N+2)*(L+3), sizeof(temp_nodes));

    for(int i=0; i<L+3; i++){
	temp_nodes[i*(N+2)] = DBL_MAX;
	temp_nodes[i*(N+2)+N+1] = DBL_MAX;
    }
    for(int i=0; i<N+2; i++){
	temp_nodes[i] = DBL_MAX;
	temp_nodes[(N+2)*(L+2)+i] = DBL_MAX;
    }

    enum{ LEFT = 1, BELOW, ABOVE, RIGHT};


    for(int i=0; i<2*(L+1); i++){

	double sum, old_sum, s1, s2, s3, s4;
	int i_node, n, n_bonds, t, row, col, brow, trow, lcol, rcol;


	sum = 0;
	old_sum = 1;

	// i_node should not be updated (locked to zero)
	//i_node = (i < L+1) ? i*N+l : (i-L-1)*N+L+l;
	i_node = (i < L+1) ? (i+1)*(N+2)+l+1 : (i-L)*(N+2)+L+l+1;
	temp_nodes[i_node] = 0;
	col = (i < L+1) ? l+1 : l+L+1;
	row = (i < L+1) ? i : i-L-1;
	//printf("i_node = %d, row = %d, col = %d\n", i_node, row, col);

	for(int block_size = bsize; block_size < bsize+N+2; block_size += bsize){
	    /* printf("Block size = %d\n", block_size); */
	    brow = (row-block_size) > 0 ? row-block_size : 1;
	    trow = (row+block_size) < L+2 ? row+block_size : L+1;
	    lcol = (col-block_size) > 0 ? col-block_size : 1;
	    rcol = (col+block_size) < N+1 ? col+block_size : N;
	    
	    // Set nodes outside of block to DBL_MAX
	    for(int j=lcol-1; j<rcol+1; j++){
	    	temp_nodes[(brow-1)*(N+2)+j] = DBL_MAX;
	    	temp_nodes[(trow+1)*(N+2)+j] = DBL_MAX;
	    }
	    for(int j=brow-1; j<trow+1; j++){
	    	temp_nodes[j*(N+2)+lcol-1] = DBL_MAX;
	    	temp_nodes[j*(N+2)+rcol+1] = DBL_MAX;
	    }

	    sum = 0;
	    old_sum = 1;
	    while(fabs(sum-old_sum) > tol){
		old_sum = sum;
		sum = 0;
	    
		// brow and trow in range (inclusive) [1, L+1]
		// lcol and rcol in range (inclusive) [1, N]
		for(int j=brow; j<=trow; j++){		
		    for(int k=lcol; k<=rcol; k++){

			n = j*(N+2)+k;
			n_bonds = (j-1)*N+k-1;
			s1 = temp_nodes[ j*(N+2)+k-(N+2)-1 ] + temp_thresholds[ j*(N+2)+k-(N+2) ][0];
			s2 = temp_nodes[ j*(N+2)+k-(N+2)   ] + temp_thresholds[ j*(N+2)+k-(N+2) ][1];
			s3 = temp_nodes[ j*(N+2)+k+(N+2)   ] + temp_thresholds[ j*(N+2)+k       ][1];
			s4 = temp_nodes[ j*(N+2)+k+(N+2)+1 ] + temp_thresholds[ j*(N+2)+k+1     ][0];

			temp_nodes[ n ] = minNeighbour4(s1, s2, s3, s4, &t) * (n != i_node);
			bonds[i][ n_bonds ] = (temp_nodes[ n ] > 0) * t;
			sum += temp_nodes[ n ];
		    }
		}
	    }

	}

	// As source nodes close to each other have similar distances to
	// the other nodes, run i+1 through starts from the results
	// of run i. No need for priming pathscape, as this is much
	// faster (factor of ~15 for 16x16). For single-threaded
	// run-through, the run L+1 starts from the results of run 0,
	// as they are more similar than L and L+1 (top and bottom of column
	// 1 and 2 respectively).

	for(int j=1; j<L+2; j++){
	    for(int k=1; k<N+1; k++){
		nodes[i][(j-1)*N+k-1] = temp_nodes[j*(N+2)+k];
	    }
	}
    }


    free(temp_nodes);

}



void calculateOptimalPathscape(double **restrict nodes, 
			       double **restrict temp_thresholds, 
			       int **bonds, int L, double tol, int nthreads){

    
    int N = 2*L;
    int l = L/2;

    enum{ LEFT = 1, BELOW, ABOVE, RIGHT};

    /* double t1 = walltime(); */
    //if(nthreads < 4){
    //nthreads = nthreads <= 1 ? 1 : 2;

#pragma omp parallel for num_threads(nthreads) schedule(static,L+1)
    // pragma commented out for now, as the overhead seems to be quite large

    //  	private(sum, old_sum, s1, s2, s3, s4, i_node, n, t)
    //nthreads = 1;
    for(int i=0; i<2*(L+1); i++){

	double sum, old_sum, s1, s2, s3, s4;
	int i_node, n, t;

	sum = 0;
	old_sum = 1;
	
	// i_node should not be updated (locked to zero)
	i_node = (i < L+1) ? i*N+l : (i-L-1)*N+L+l;
	nodes[i][i_node] = 0;

	while(fabs(sum-old_sum) > tol){
	    old_sum = sum;
	    sum = 0;


	    // j == 0
	    for(int k=0; k<N-1; k++){
		s3 = nodes[i][ k+N   ] + temp_thresholds[ k   ][1];
		s4 = nodes[i][ k+N+1 ] + temp_thresholds[ k+1 ][0];
		nodes[i][ k ] = (s3 < s4) ? s3 : s4;
		nodes[i][ k ] *= (k != i_node);
		bonds[i][ k ] = (nodes[i][ k ] > 0) * (s3 < s4 ? ABOVE : RIGHT);
		sum += nodes[i][ k ];
	    }

	    // j == 0, k == N-1
	    nodes[i][ N-1 ] = (nodes[i][ 2*N-1 ] + temp_thresholds[ N-1 ][1]) * (N-1 != i_node); 
	    bonds[i][ N-1 ] = (nodes[i][ N-1 ] > 0) * ABOVE;
	    sum += nodes[i][ N-1 ];

	    for(int j=1; j<L; j++){

		// k == 0
		n = j*N;
		s2 = nodes[i][ j*N-N   ] + temp_thresholds[ j*N-N ][1];
		s3 = nodes[i][ j*N+N   ] + temp_thresholds[ j*N   ][1];
		s4 = nodes[i][ j*N+N+1 ] + temp_thresholds[ j*N+1 ][0];
		nodes[i][ n ] = minNeighbour3(s2, s3, s4, &t) * (n != i_node);
		bonds[i][ n ] = (nodes[i][ n ] > 0) * (t+1);
		sum += nodes[i][ n ];
		
		for(int k=1; k<N-1; k++){

		    n = j*N+k;
		    s1 = nodes[i][ j*N+k-N-1 ] + temp_thresholds[ j*N+k-N ][0];
		    s2 = nodes[i][ j*N+k-N   ] + temp_thresholds[ j*N+k-N ][1];
		    s3 = nodes[i][ j*N+k+N   ] + temp_thresholds[ j*N+k   ][1];
		    s4 = nodes[i][ j*N+k+N+1 ] + temp_thresholds[ j*N+k+1 ][0];
		    
		    nodes[i][ n ] = minNeighbour4(s1, s2, s3, s4, &t) * (n != i_node);
		    bonds[i][ n ] = (nodes[i][ n ] > 0) * t;
		    sum += nodes[i][ n ];
		}

		// k == N-1
		n = j*N+N-1;
		s1 = nodes[i][ j*N-2     ] + temp_thresholds[ j*N-1   ][0];
		s2 = nodes[i][ j*N+-1    ] + temp_thresholds[ j*N-1   ][1];
		s3 = nodes[i][ j*N+2*N-1 ] + temp_thresholds[ j*N+N-1 ][1];
		nodes[i][ n ] = minNeighbour3(s1, s2, s3, &t) * (n != i_node);
		bonds[i][ n ] = (nodes[i][ n ] > 0) * t;
		sum += nodes[i][ n ];

	    }
	    
	    // j == L
	    // k == 0
	    nodes[i][ L*N ] = (nodes[i][ L*N-N   ] + temp_thresholds[ L*N-N ][1]) * (L*N != i_node);
	    bonds[i][ L*N ] = (nodes[i][ L*N ] > 0) * BELOW;
	    sum += nodes[i][ L*N ];

	    for(int k=1; k<N; k++){
		n = L*N+k;
		s1 = nodes[i][ L*N+k-N-1 ] + temp_thresholds[ L*N+k-N ][0];
		s2 = nodes[i][ L*N+k-N   ] + temp_thresholds[ L*N+k-N ][1];
		nodes[i][ n ] = (s1 < s2) ? s1 : s2;
		nodes[i][ n ] *= (n != i_node);
		bonds[i][ n ] = (nodes[i][ n ] > 0) * (s1 < s2 ? LEFT : BELOW);
		sum += nodes[i][ n ];
	    }
	}

	// As source nodes close to each other have similar distances to
	// the other nodes, run i+1 through starts from the results
	// of run i. No need for priming pathscape, as this is much
	// faster (factor of ~15 for 16x16). For single-threaded
	// run-through, the run L+1 starts from the results of run 0,
	// as they are more similar than L and L+1 (top and bottom of column
	// 1 and 2 respectively).

	if(i < 2*(L+1)-1 && i != L){
	    if(i == 0 && nthreads == 1)
		for(int j=0; j<N*(L+1); j++){
		    nodes[i+1][j] = nodes[i][j];
		    bonds[i+1][j] = bonds[i][j];
		    nodes[L+1][j] = nodes[i][j];
		    bonds[L+1][j] = bonds[i][j];
		}
	    else
		for(int j=0; j<N*(L+1); j++){
		    nodes[i+1][j] = nodes[i][j];
		    bonds[i+1][j] = bonds[i][j];
		}
	}
	

	/* if(i == 0 || i == L+1) */
	/*     printf("2 threads, finished the first run in %.4fs, i = %d\n", */
	/* 	   walltime()-t1, i); */
    }
    /* printf("2 threads, finished all the runs in %.4fs\n", walltime()-t1); */
}



void calculateOptimalPathscape4T(double **restrict nodes, 
			       double **restrict temp_thresholds, 
			       int **bonds, int L, double tol, int nthreads){

    
    int N = 2*L;
    int l = L/2;
    

    enum{ LEFT = 1, BELOW, ABOVE, RIGHT};

    /* double t1 = walltime(); */
#pragma omp parallel for num_threads(2)
    //  	private(sum, old_sum, s1, s2, s3, s4, i_node, n, t)
    for(int i=L/2; i<L/2+L+2; i+=L+1){

	double sum, old_sum, s1, s2, s3, s4;
	int i_node, n, t;

	sum = 0;
	old_sum = 1;
	
	// i_node should not be updated (locked to zero)
	i_node = (i < L+1) ? i*N+l : (i-L-1)*N+L+l;
	nodes[i][i_node] = 0;

	while(fabs(sum-old_sum) > tol){
	    old_sum = sum;
	    sum = 0;
	    

	    // j == 0
	    for(int k=0; k<N-1; k++){
		s3 = nodes[i][ k+N   ] + temp_thresholds[ k   ][1];
		s4 = nodes[i][ k+N+1 ] + temp_thresholds[ k+1 ][0];
		nodes[i][ k ] = (s3 < s4) ? s3 : s4;
		nodes[i][ k ] *= (k != i_node);
		bonds[i][ k ] = (nodes[i][ k ] > 0) * (s3 < s4 ? ABOVE : RIGHT);
		sum += nodes[i][ k ];
	    }

	    // j == 0, k == N-1
	    nodes[i][ N-1 ] = (nodes[i][ 2*N-1 ] + temp_thresholds[ N-1 ][1]) * (N-1 != i_node); 
	    bonds[i][ N-1 ] = (nodes[i][ N-1 ] > 0) * ABOVE;
	    sum += nodes[i][ N-1 ];

	    for(int j=1; j<L; j++){

		// k == 0
		n = j*N;
		s2 = nodes[i][ j*N-N   ] + temp_thresholds[ j*N-N ][1];
		s3 = nodes[i][ j*N+N   ] + temp_thresholds[ j*N   ][1];
		s4 = nodes[i][ j*N+N+1 ] + temp_thresholds[ j*N+1 ][0];
		nodes[i][ n ] = minNeighbour3(s2, s3, s4, &t) * (n != i_node);
		bonds[i][ n ] = (nodes[i][ n ] > 0) * (t+1);
		sum += nodes[i][ n ];
		
		for(int k=1; k<N-1; k++){

		    n = j*N+k;
		    s1 = nodes[i][ j*N+k-N-1 ] + temp_thresholds[ j*N+k-N ][0];
		    s2 = nodes[i][ j*N+k-N   ] + temp_thresholds[ j*N+k-N ][1];
		    s3 = nodes[i][ j*N+k+N   ] + temp_thresholds[ j*N+k   ][1];
		    s4 = nodes[i][ j*N+k+N+1 ] + temp_thresholds[ j*N+k+1 ][0];
		    
		    nodes[i][ n ] = minNeighbour4(s1, s2, s3, s4, &t) * (n != i_node);
		    bonds[i][ n ] = (nodes[i][ n ] > 0) * t;
		    sum += nodes[i][ n ];
		}

		// k == N-1
		n = j*N+N-1;
		s1 = nodes[i][ j*N-2     ] + temp_thresholds[ j*N-1   ][0];
		s2 = nodes[i][ j*N+-1    ] + temp_thresholds[ j*N-1   ][1];
		s3 = nodes[i][ j*N+2*N-1 ] + temp_thresholds[ j*N+N-1 ][1];
		nodes[i][ n ] = minNeighbour3(s1, s2, s3, &t) * (n != i_node);
		bonds[i][ n ] = (nodes[i][ n ] > 0) * t;
		sum += nodes[i][ n ];

	    }
	    
	    // j == L
	    // k == 0
	    nodes[i][ L*N ] = (nodes[i][ L*N-N   ] + temp_thresholds[ L*N-N ][1]) * (L*N != i_node);
	    bonds[i][ L*N ] = (nodes[i][ L*N ] > 0) * BELOW;
	    sum += nodes[i][ L*N ];

	    for(int k=1; k<N; k++){
		n = L*N+k;
		s1 = nodes[i][ L*N+k-N-1 ] + temp_thresholds[ L*N+k-N ][0];
		s2 = nodes[i][ L*N+k-N   ] + temp_thresholds[ L*N+k-N ][1];
		nodes[i][ n ] = (s1 < s2) ? s1 : s2;
		nodes[i][ n ] *= (n != i_node);
		bonds[i][ n ] = (nodes[i][ n ] > 0) * (s1 < s2 ? LEFT : BELOW);
		sum += nodes[i][ n ];
	    }
	}

	// As source nodes close to each other have similar distances to
	// the other nodes, the next run through starts from the results
	// of this one. No need for priming pathscape, as this is much
	// faster (factor of ~15 for 16x16). 
	// Need to adjust this for 4 threads
	if(i == L/2)
	    for(int k=0; k<L+1; k++)
		for(int j=0; j<N*(L+1); j++){
		    nodes[k][j] = nodes[i][j];
		    bonds[k][j] = bonds[i][j];
		}
	if(i == L/2+L+1)
	    for(int k=L+1; k<2*L+2; k++)
		for(int j=0; j<N*(L+1); j++){
		    nodes[k][j] = nodes[i][j];
		    bonds[k][j] = bonds[i][j];
		}
	
    }
    /* double t2 = walltime(); */
    /* printf("4 Threads, finished the two first runs in %.4fs\n", t2-t1); */

    //#pragma omp parallel for num_threads(4) schedule(static,L/2) 
    //  	private(sum, old_sum, s1, s2, s3, s4, i_node, n, t)
    //for(int i=1; i<2*L+1; i++){

#pragma omp parallel num_threads(4)
    {
	int thread_num = omp_get_thread_num();
	for(int r=1; r<L/2+1; r++){
	    int i = 0;
	    double sum, old_sum, s1, s2, s3, s4;
	    int i_node, n, t;
	
	    switch(thread_num){
	    case 0:
		//i = L/2+r;
		i = L/2-r;
		break;
	    case 1:
		//i = L+1+L/2+r;
		i = L/2+r;
		break;
	    case 2:
		//i = L/2-r;
		i = L+1+L/2-r;
		break;
	    case 3:
		//i = L+1+L/2-r;
		i = L+1+L/2+r;
		break;
	    }

	    sum = 0;
	    old_sum = 1;
	
	    // i_node should not be updated (locked to zero)
	    i_node = (i < L+1) ? i*N+l : (i-L-1)*N+L+l;
	    nodes[i][i_node] = 0;

	    while(fabs(sum-old_sum) > tol){
		old_sum = sum;
		sum = 0;
	    

		// j == 0
		for(int k=0; k<N-1; k++){
		    s3 = nodes[i][ k+N   ] + temp_thresholds[ k   ][1];
		    s4 = nodes[i][ k+N+1 ] + temp_thresholds[ k+1 ][0];
		    nodes[i][ k ] = (s3 < s4) ? s3 : s4;
		    nodes[i][ k ] *= (k != i_node);
		    bonds[i][ k ] = (nodes[i][ k ] > 0) * (s3 < s4 ? ABOVE : RIGHT);
		    sum += nodes[i][ k ];
		}

		// j == 0, k == N-1
		nodes[i][ N-1 ] = (nodes[i][ 2*N-1 ] + temp_thresholds[ N-1 ][1]) * (N-1 != i_node); 
		bonds[i][ N-1 ] = (nodes[i][ N-1 ] > 0) * ABOVE;
		sum += nodes[i][ N-1 ];

		for(int j=1; j<L; j++){

		    // k == 0
		    n = j*N;
		    s2 = nodes[i][ j*N-N   ] + temp_thresholds[ j*N-N ][1];
		    s3 = nodes[i][ j*N+N   ] + temp_thresholds[ j*N   ][1];
		    s4 = nodes[i][ j*N+N+1 ] + temp_thresholds[ j*N+1 ][0];
		    nodes[i][ n ] = minNeighbour3(s2, s3, s4, &t) * (n != i_node);
		    bonds[i][ n ] = (nodes[i][ n ] > 0) * (t+1);
		    sum += nodes[i][ n ];
		
		    for(int k=1; k<N-1; k++){

			n = j*N+k;
			s1 = nodes[i][ j*N+k-N-1 ] + temp_thresholds[ j*N+k-N ][0];
			s2 = nodes[i][ j*N+k-N   ] + temp_thresholds[ j*N+k-N ][1];
			s3 = nodes[i][ j*N+k+N   ] + temp_thresholds[ j*N+k   ][1];
			s4 = nodes[i][ j*N+k+N+1 ] + temp_thresholds[ j*N+k+1 ][0];
		    
			nodes[i][ n ] = minNeighbour4(s1, s2, s3, s4, &t) * (n != i_node);
			bonds[i][ n ] = (nodes[i][ n ] > 0) * t;
			sum += nodes[i][ n ];
		    }

		    // k == N-1
		    n = j*N+N-1;
		    s1 = nodes[i][ j*N-2     ] + temp_thresholds[ j*N-1   ][0];
		    s2 = nodes[i][ j*N+-1    ] + temp_thresholds[ j*N-1   ][1];
		    s3 = nodes[i][ j*N+2*N-1 ] + temp_thresholds[ j*N+N-1 ][1];
		    nodes[i][ n ] = minNeighbour3(s1, s2, s3, &t) * (n != i_node);
		    bonds[i][ n ] = (nodes[i][ n ] > 0) * t;
		    sum += nodes[i][ n ];

		}
	    
		// j == L
		// k == 0
		nodes[i][ L*N ] = (nodes[i][ L*N-N   ] + temp_thresholds[ L*N-N ][1]) * (L*N != i_node);
		bonds[i][ L*N ] = (nodes[i][ L*N ] > 0) * BELOW;
		sum += nodes[i][ L*N ];

		for(int k=1; k<N; k++){
		    n = L*N+k;
		    s1 = nodes[i][ L*N+k-N-1 ] + temp_thresholds[ L*N+k-N ][0];
		    s2 = nodes[i][ L*N+k-N   ] + temp_thresholds[ L*N+k-N ][1];
		    nodes[i][ n ] = (s1 < s2) ? s1 : s2;
		    nodes[i][ n ] *= (n != i_node);
		    bonds[i][ n ] = (nodes[i][ n ] > 0) * (s1 < s2 ? LEFT : BELOW);
		    sum += nodes[i][ n ];
		}
	    }

	    // As source nodes close to each other have similar distances to
	    // the other nodes, the next run through starts from the results
	    // of this one. No need for priming pathscape, as this is much
	    // faster (factor of ~15 for 16x16). 
	    if(i != 0 && i != L && i != L+1 && i!= 2*L+1){
		//int next_node = thread_num < 2 ? i+1 : i-1;
		int next_node = thread_num & 1 ? i+1 : i-1; 
		for(int j=0; j<N*(L+1); j++){
		    nodes[next_node][j] = nodes[i][j];
		    bonds[next_node][j] = bonds[i][j];
		}
	    }
	}
    }
    /* t1 = walltime(); */
    /* printf("4 Threads, finished the rest of the runs in %.4fs\n", t1-t2); */

}


void optimalPathscape(struct System *sys){

    /*-------------------------------------------------------------------------*/
    /*  Calculates the optimal pathscape for the system using the approach in  */
    /*  A. Hansen and J. Kertesz, Phys. Rev. Lett 93, 040601 (2004).           */
    /*                                                                         */
    /*  Assumes a system of size Lx(L+1) with L even!                          */
    /*-------------------------------------------------------------------------*/

    int L = sys->L, N = 2*L, l = L/2;
    int *conv_col;
    int k1, k2;
    double **nodes, **temp_nodes, **temp_thresholds;//, **btemp_thresholds;
    int **bonds, **temp_bonds0, **temp_bonds1;
    //int ***temp_bonds;

    conv_col = calloc(N, sizeof(conv_col));
    for(int i=0; i<l; i++)
	conv_col[i] = i+l;
    for(int i=l; i<l+L; i++)
	conv_col[i] = i-l;
    for(int i=l+L; i<N; i++)
	conv_col[i] = i-l-L;

    nodes = calloc(2*(L+1), sizeof(nodes));
    temp_nodes = calloc(L+1, sizeof(temp_nodes));
    bonds = calloc(2*(L+1), sizeof(bonds));
    temp_bonds0 = calloc(L+1, sizeof(temp_bonds0));
    temp_bonds1 = calloc(L+1, sizeof(temp_bonds1));
    temp_thresholds = calloc(N*L, sizeof(temp_thresholds));
    //btemp_thresholds = calloc((N+2)*(L+2), sizeof(btemp_thresholds));
    for(int i=0; i<2*(L+1); i++){
	nodes[i] = calloc(N*(L+1), sizeof(nodes[i]));
	bonds[i] = calloc(N*(L+1), sizeof(bonds[i]));
	if(i < L+1){
	    temp_nodes[i] = calloc(L*(L+1), sizeof(temp_nodes[i]));
	    temp_bonds0[i] = calloc(L*(L+1), sizeof(temp_bonds0[i]));
	    temp_bonds1[i] = calloc(L*(L+1), sizeof(temp_bonds1[i]));
	    /* temp_bonds[i] = calloc(L*(L+1), sizeof(temp_bonds[i])); */
	    /* for(int j=0; j<L*(L+1); j++) */
	    /* 	temp_bonds[i][j] = calloc(2, sizeof(temp_bonds[i][j])); */
	}
    }
    for(int i=0; i<L; i++){
	for(int j=0; j<N; j++){
	    temp_thresholds[ i*N+j ] = calloc(2, sizeof(temp_thresholds[i*N+j]));
	    temp_thresholds[ i*N+j ][0] = sys->threshold[ i*L+conv_col[j] ][0];
	    temp_thresholds[ i*N+j ][1] = sys->threshold[ i*L+conv_col[j] ][1];
	    /* btemp_thresholds[ (i+1)*(N+2)+j+1 ] = calloc(2, sizeof(btemp_thresholds[(i+1)*(N+2)+j+1])); */
	    /* btemp_thresholds[ (i+1)*(N+2)+j+1 ][0] = sys->threshold[ i*L+conv_col[j] ][0]; */
	    /* btemp_thresholds[ (i+1)*(N+2)+j+1 ][1] = sys->threshold[ i*L+conv_col[j] ][1]; */
	}
    }
    /* for(int i=0; i<L+2; i++){ */
    /* 	btemp_thresholds[ i*(N+2) ] = calloc(2, sizeof(btemp_thresholds[i*(N+2)])); */
    /* 	btemp_thresholds[ i*(N+2)+N+1 ] = calloc(2, sizeof(btemp_thresholds[i*(N+2)+N+1])); */
    /* 	btemp_thresholds[ i*(N+2) ][0] = DBL_MAX; */
    /* 	btemp_thresholds[ i*(N+2) ][1] = DBL_MAX; */
    /* 	btemp_thresholds[ i*(N+2)+N+1 ][0] = DBL_MAX; */
    /* 	btemp_thresholds[ i*(N+2)+N+1 ][1] = DBL_MAX; */
    /* } */
    /* for(int i=0; i<N+2; i++){ */
    /* 	btemp_thresholds[ i ] = calloc(2, sizeof(btemp_thresholds[i])); */
    /* 	btemp_thresholds[ i+(L+1)*(N+2) ] = calloc(2, sizeof(btemp_thresholds[i+(L+1)*(N+2)])); */
    /* 	btemp_thresholds[ i ][0] = DBL_MAX; */
    /* 	btemp_thresholds[ i ][1] = DBL_MAX; */
    /* 	btemp_thresholds[ i+(L+1)*(N+2) ][0] = DBL_MAX; */
    /* 	btemp_thresholds[ i+(L+1)*(N+2) ][1] = DBL_MAX; */
    /* } */


    calculateOptimalPathscape(nodes, temp_thresholds, bonds, L, 1e-10, sys->nthreads);
    /* if(sys->nthreads == 2) */
    /* 	calculateOptimalPathscape(nodes, temp_thresholds, bonds, L, 1e-10, sys->nthreads); */
    /* else */
    /* 	calculateOptimalPathscapeBlocks(nodes, btemp_thresholds, bonds, L, 1e-10, sys->nthreads); */
    

    // Extract optimal pathscape to sys

    for(int i=0; i<L+1; i++){

    	// nodes[i][:] are set to the sum of new_nodes[i][:] and new_nodes[i+L][:],
    	// i.e. total threshold for spanning paths going through the different nodes
    	for(int j=0; j<N*(L+1); j++){
    	    nodes[i][j] = (nodes[i][j] + nodes[i+L+1][j]);
    	    // temp_bonds[i][j][0] = bonds[i][j][0]
    	    // temp_bonds[i][j][1] = bonds[i+L+1][j][1]
    	}

    	// temp_nodes[] is of same size as pathscape.
    	// Set node values to the minimum of paths going through columns k and k'
    	for(int j=0; j<L+1; j++)
    	    for(int k=0; k<L; k++){
    		k1 = k+l;
    		k2 = (k < l) ? k+L+l : k-l;

    		temp_bonds0[i][j*L+k] = nodes[i][j*N+k1] < nodes[i][j*N+k2] ?
    		    bonds[i][j*N+k1] : bonds[i][j*N+k2];

    		temp_bonds1[i][j*L+k] = nodes[i][j*N+k1] < nodes[i][j*N+k2] ?
    		    bonds[i+L+1][j*N+k1] : bonds[i+L+1][j*N+k2];

    		temp_nodes[i][j*L+k] =
    		    nodes[i][j*N+k1] < nodes[i][j*N+k2] ?
    		    nodes[i][j*N+k1] : nodes[i][j*N+k2];
    	    }
    }

    for(int i=0; i<L+1; i++)
    	for(int j=0; j<L*(L+1); j++){

    	    if(fabs(temp_nodes[i][j]) > 1e-6){

    		sys->pathscape_bonds[j][0] =
    		    sys->pathscape[j] < temp_nodes[i][j] ?
    		    sys->pathscape_bonds[j][0] : temp_bonds0[i][j];

    		sys->pathscape_bonds[j][1] =
    		    sys->pathscape[j] < temp_nodes[i][j] ?
    		    sys->pathscape_bonds[j][1] : temp_bonds1[i][j];

    		sys->pathscape[j] =
    		    sys->pathscape[j] < temp_nodes[i][j] ?
    		    sys->pathscape[j] : temp_nodes[i][j];
    		}
    	    else
    	    printf("Low val for i = %d, j = %d\n", i, j);
    	}
    
    free(conv_col);
    for(int i=0; i<N*L; i++)
    	free(temp_thresholds[i]);
    /* for(int i=0; i<(N+2)*(L+2); i++) */
    /* 	free(btemp_thresholds[i]); */

    for(int i=0; i<2*(L+1); i++){
	free(bonds[i]);
	free(nodes[i]);
    }
    for(int i=0; i<L+1; i++){
	/* for(int j=0; j<L*(L+1); j++) */
	/*     free(temp_bonds[i][j]); */
	/* free(temp_bonds[i]); */
	free(temp_bonds0[i]);
	free(temp_bonds1[i]);
	free(temp_nodes[i]);
    }
    free(temp_thresholds);
    //free(btemp_thresholds);
    free(bonds);
    free(temp_bonds0);
    free(temp_bonds1);
    free(nodes);
    free(temp_nodes);

}



void calculateOptimalPathscapeFlow(double **restrict nodes, 
				   double **restrict thresholds, 
				   int **bonds, int L, double tol, int nthreads){

    /*  Similar to calculateOptimalPathscape(), but using voltage nodes as nodes,  */
    /*  i.e. pathscape for spanning paths between the electrodes.                  */
    /*  Have L*L*2 bonds, thus L+1 nodes in each column (between electrodes), and  */
    /*  L nodes in each row, with periodic boundaries. As the edge nodes are       */
    /*  initialized to infinity (DBL_MAX), except for one, only the internal       */
    /*  nodes are run through, that is, row 1 to row L-1 (rows 0 and L are edges). */

    //int N = 2*L;
    int N = L;//sys->N;
    int t;
    double sum, old_sum, s1, s2, s3, s4;    
    int *ir, *il;

    ir = calloc(N, sizeof(ir));
    il = calloc(N, sizeof(il));
    for(int i=0; i<N; i++){
	ir[i] = i+1;
	il[i] = i-1;
    }
    ir[N-1] = 0;
    il[0] = N-1;


    /* Initialize edge nodes, all internal nodes are zero */
    /* nodes[i][j] is zero if i==j, DBL_MAX otherwise  (for edge nodes) */

    for(int i=0; i<2*N; i++)
	for(int j=0; j<N; j++){
	    nodes[i][j] = DBL_MAX;
	    nodes[i][j+N*L] = DBL_MAX;
	}

    for(int i=0; i<N; i++){
	nodes[i][i] = 0;
	nodes[i+N][i+N*L] = 0;
    }


    enum{ LEFT = 1, BELOW, ABOVE, RIGHT};

    /* double t1 = walltime(); */
    //if(nthreads < 4){
    //nthreads = nthreads <= 1 ? 1 : 2;

    //#pragma omp parallel for num_threads(nthreads) schedule(static,L+1)
    // pragma overhead seems to be quite large, better to comment out
    // if it is known that only one thread will be used

    for(int i=0; i<2*N; i++){
	//printf("Run %d\n", i);
	sum = 0;
	old_sum = 1;

	while(fabs(sum-old_sum) > 1e-10){
	    old_sum = sum;
	    sum = 0;

	    for(int j=1; j<L; j++){
		for(int k=0; k<N; k++){

		    s1 = thresholds[ j*N+il[k]-N ][1] + nodes[ i ][ j*N+il[k]-N ];
		    s2 = thresholds[ j*N+k-N     ][0] + nodes[ i ][ j*N+k-N     ];
		    s3 = thresholds[ j*N+k       ][0] + nodes[ i ][ j*N+k+N     ];
		    s4 = thresholds[ j*N+k       ][1] + nodes[ i ][ j*N+ir[k]+N ];
		    
		    nodes[ i ][ j*N+k ] = minNeighbour4(s1, s2, s3, s4, &t);
		    bonds[ i ][ j*N+k ] = (nodes[i][ j*N+k ] > 0) * t;
		    sum += nodes[ i ][ j*N+k ];
		    
		}
	    }
	}
	// As source nodes close to each other have similar distances to
	// the other nodes, run i+1 through starts from the results
	// of run i. No need for priming pathscape, as this is much
	// faster (factor of ~15 for 16x16). For single-threaded
	// run-through, the run L starts from the results of run 0,
	// as they are more similar than L-1 and L (last of first
	// row, first of second row).


	if(i < 2*N-1 && i != N-1){
	    if(i == 0 && nthreads == 1)
		for(int j=1; j<L; j++)
		    for(int k=0; k<N; k++){
			nodes[i+1][j*N+k] = nodes[i][j*N+k];
			bonds[i+1][j*N+k] = bonds[i][j*N+k];
			nodes[N][j*N+k] = nodes[i][j*N+k];
			bonds[N][j*N+k] = bonds[i][j*N+k];
		    }
	    else
		for(int j=1; j<L; j++)
		    for(int k=0; k<N; k++){
			//printf("j = %d, k = %d, j*N+k = %d\n", j, k, j*N+k);
			nodes[i+1][j*N+k] = nodes[i][j*N+k];
			bonds[i+1][j*N+k] = bonds[i][j*N+k];
		    }
	}
    }

    free(ir);
    free(il);

}



void optimalPathscapeFlow(struct System *sys){

    /*-------------------------------------------------------------------------*/
    /*  Calculates the optimal pathscape for the system using the approach in  */
    /*  A. Hansen and J. Kertesz, Phys. Rev. Lett 93, 040601 (2004).           */
    /*                                                                         */
    /*  Assumes a system of size Lx(L+1) with L even.                          */
    /*  Calculates the pathscape for flow between electrodes ("flowscape"),    */
    /*  using voltage nodes as nodes. This is in contrast to optimalPathscape()*/
    /*  which calculates the pathscape parallel to the nodes, which is used    */
    /*  in fracture problems (theta <= 0), or the "fracturescape".             */
    /*-------------------------------------------------------------------------*/
    
    // L is typical system size. In case of L != N, L is the height of the
    // system, (distance between electrodes) and N is the width (length of electrodes)
    int L = sys->L, N = L;
    double **nodes;
    int **bonds;

    nodes = calloc(2*N, sizeof(nodes));
    bonds = calloc(2*N, sizeof(bonds));
    for(int i=0; i<2*N; i++){
	nodes[i] = calloc(N*(L+1), sizeof(nodes[i]));
	bonds[i] = calloc(N*(L+1), sizeof(bonds[i]));
    }


    calculateOptimalPathscapeFlow(nodes, sys->threshold, bonds, L, 1e-10, sys->nthreads);
    

    // Extract optimal pathscape to sys

    /* for(int i=0; i<L; i++){ */

    /* 	// nodes[i][:] are set to the sum of new_nodes[i][:] and new_nodes[i+L][:], */
    /* 	// i.e. total threshold for spanning paths going through the different nodes */
    /* 	for(int j=1; j<L-1; j++) */
    /* 	    for(int k=0; k<N; k++) */
    /* 		nodes[i][j*N+k] = (nodes[i][j*N+k] + nodes[i+L+1][j*N+k]);	 */
    /* } */

    for(int i=0; i<N; i++)
	for(int j=N; j<2*N; j++)
	    for(int k=N; k<N*L; k++){

    	    if(fabs(nodes[i][k]+nodes[j][k]) > 1e-6){

    		sys->pathscape_bonds[k][0] =
    		    sys->pathscape[k] < nodes[i][k] + nodes[j][k] ?
    		    sys->pathscape_bonds[k][0] : bonds[i][k];

    		sys->pathscape_bonds[k][1] =
    		    sys->pathscape[k] < nodes[i][k] + nodes[j][k] ?
    		    sys->pathscape_bonds[k][1] : bonds[j][k];

    		sys->pathscape[k] =
    		    sys->pathscape[k] < nodes[i][k] + nodes[j][k] ?
    		    sys->pathscape[k] : nodes[i][k] + nodes[j][k];
	    }
    	    else
		printf("Low val for i = %d, j = %d\n", i, j);
    	}


    for(int i=0; i<2*N; i++){
	free(bonds[i]);
	free(nodes[i]);
    }
    free(bonds);
    free(nodes);

}



void compare(int max_runs, double b_min, double b_max, double b_step, double dV, double beta, char *fname){

    struct System A;
    struct Statistics stats;
    double start, stop, total_time;
    int runs = max_runs;
    int min_size = 16;
    int max_size = 16;
    int size_step = 16;
    int **shortest_path, **backbone, **fracture;
    int faultcount[3];
    int end, index;
    int L_range, b_range, ft_range, b_index, L_index/*,  r_size */;
    double b, temp;


    L_range = (max_size - min_size)/size_step + 1;
    b_range = (b_max - b_min)/b_step + 1;
    ft_range = 3;

    temp = b_min;
    for(int i=0; i<b_range; i++)
	temp += b_step;
    stats.theta_incr = b_step;
    stats.theta_min = b_min;
    stats.theta_max = temp;

    allocStats(&stats, L_range, b_range, ft_range, fname);
    stats.L_incr = size_step;

    b_index = 0;
    for(b=b_min; b<=b_max; b+=b_step, b_index++){
    
	total_time = 0;

	printf("\nRunning through set with theta = %.3f, dV = %.4f\n", b, dV);


	total_time = 0;
	start = walltime();
	runs = max_runs;

	L_index = 0;
	stats.b[b_index] = b;

	
	for(int size = min_size; size <= max_size; size += size_step, L_index++){
	    printf("Running through size %d, total runs = %d\n", size, runs);
	    // updated_mass: contains the total mass of the updated bonds
	    // fracture_mass: contains the total mass of the largest fracture
	    // backbone_mass: contains the total mass of the backbone
	    // sp_updated:  contains the number of bonds belonging to the
	    //              shortest path which are updated
	    // sp_fracture: contains the number of bonds belonging to both the
	    //              largest fracture and the shortest path
	    // sp_backbone: contains the number of bonds belonging to both the
	    //              backbone and the shortest paht
	    double *updated_mass = calloc(runs, sizeof(updated_mass));
	    double *fracture_mass = calloc(runs, sizeof(fracture_mass));
	    double *backbone_mass = calloc(runs, sizeof(backbone_mass));
	    double *sp_updated = calloc(runs, sizeof(sp_updated));
	    double *sp_fracture = calloc(runs, sizeof(sp_fracture));
	    double *sp_backbone = calloc(runs, sizeof(sp_backbone));
	    stats.L[L_index] = size;
	    stats.runs[L_index] = runs;

	    for(int run=0; run<runs; run++){
		printf("\t run %d\n", run);
		allocSys(&A, size);
		initSys(&A, b, size+run, beta, 1);

		shortest_path = shortestPath(A.threshold, size);
	    
		A.max_voltage = dV;

		fracture = NULL;
		backbone = NULL;


		int breaklimit = 16*size*size;
		int updated_count = 0;
		int bn = 0;  // broken network
		while( !bn ){
		    updated_count = 0;
		    CG(&A);
		    while(!bn && (updateConductancePlastic23(&A) || updateConductanceSingleElastic(&A))){
			CG(&A);		
			updateDV(&A);
			bn = HKV(A.updated, A.L);
			updated_count++;
			if(updated_count > breaklimit){
			    printf("Breaklimit! Max voltage = %.4f\n", A.max_voltage);
			    break;
			}
		    }

		    A.max_voltage += dV;
		}

		
		fracture = getFracture(A.updated, A.L);
		backbone = getBackbone(A.updated, A.L);
		
		assert(HKV(fracture, A.L));  /*  make sure fracture spans network  */
		assert(HKV(backbone, A.L));  /*  make sure backbone spans network  */
		

		
		// Check if shortest path is in the largest fracture or the
		// backbone. faultcount[] couns the number of bonds in the
		// shortest path which are NOT in the fracture (array index 0)
		// or the backbone (array index 1).
		
		index=0;
		faultcount[0] = 0;
		faultcount[1] = 0;
		faultcount[2] = 0;
		while(shortest_path[index][1] >= 0){

		    if(  fracture[ shortest_path[index][0] ][ shortest_path[index][1] ] == 0){
			faultcount[0]++;
		    }
		    if(  backbone[ shortest_path[index][0] ][ shortest_path[index][1] ] == 0){
			faultcount[1]++;
		    }
		    if(  A.updated[ shortest_path[index][0] ][ shortest_path[index][1] ] == 0){
			faultcount[2]++;
		    }
		    index++;
		}
		sp_fracture[run] = 1.0*(index-faultcount[0])/index;
		stats.avg_sp_fracture[L_index][b_index] += sp_fracture[run];
		sp_backbone[run] = 1.0*(index-faultcount[1])/index;
		stats.avg_sp_backbone[L_index][b_index] += sp_backbone[run];
		sp_updated[run] = 1.0*(index-faultcount[2])/index;
		stats.avg_sp_updated[L_index][b_index] += sp_updated[run];
			      		
		
		for(int i=0; i<size*size; i++)
		    for(int j=0; j<2; j++){
			updated_mass[run] += A.updated[i][j];
			fracture_mass[run] += fracture[i][j];
			backbone_mass[run] += backbone[i][j];			
		    }
		
		stats.avg_updated_mass[L_index][b_index] += updated_mass[run];
		stats.avg_fracture_mass[L_index][b_index] += fracture_mass[run];
		stats.avg_backbone_mass[L_index][b_index] += backbone_mass[run];
		
		
		
		/*  Free memory  */
		end = INT_MAX;
		for(int i=0; i<end; i++){
		    if(shortest_path[i][1] < 0)
			end = shortest_path[i][0];
		    free(shortest_path[i]);
		}
		
		for(int i=0; i<size*size; i++){
		    free(fracture[i]);
		    free(backbone[i]);
		}
		free(fracture);
		free(backbone);
		free(shortest_path);
		freeSys(&A);
		
		
	    }  /*  end for runs  */  
	    

	    stats.avg_updated_mass[L_index][b_index] /= runs;
	    stats.avg_fracture_mass[L_index][b_index] /= runs;
	    stats.avg_backbone_mass[L_index][b_index] /= runs;
	    stats.avg_sp_updated[L_index][b_index] /= runs;
	    stats.avg_sp_fracture[L_index][b_index] /= runs;
	    stats.avg_sp_backbone[L_index][b_index] /= runs;


	    /*  Calculate variance of avg_fracture_mass, avg_backbone_mass and avg_sp  */
	    stats.var_updated_mass[L_index][b_index] = 0;
	    stats.var_fracture_mass[L_index][b_index] = 0;
	    stats.var_backbone_mass[L_index][b_index] = 0;
	    stats.var_sp_updated[L_index][b_index] = 0;
	    stats.var_sp_fracture[L_index][b_index] = 0;
	    stats.var_sp_backbone[L_index][b_index] = 0;
	    for(int i=0; i<runs; i++){
		stats.var_updated_mass[L_index][b_index] += (updated_mass[i] - stats.avg_updated_mass[L_index][b_index]) *
		                                             (updated_mass[i] - stats.avg_updated_mass[L_index][b_index]);
		stats.var_fracture_mass[L_index][b_index] += (fracture_mass[i] - stats.avg_fracture_mass[L_index][b_index]) *
		                                             (fracture_mass[i] - stats.avg_fracture_mass[L_index][b_index]);
		stats.var_backbone_mass[L_index][b_index] += (backbone_mass[i] - stats.avg_backbone_mass[L_index][b_index]) *
		                                             (backbone_mass[i] - stats.avg_backbone_mass[L_index][b_index]);
		stats.var_sp_updated[L_index][b_index] += (sp_updated[i] - stats.avg_sp_updated[L_index][b_index]) *
		                                           (sp_updated[i] - stats.avg_sp_updated[L_index][b_index]);
		stats.var_sp_fracture[L_index][b_index] += (sp_fracture[i] - stats.avg_sp_fracture[L_index][b_index]) *
		                                           (sp_fracture[i] - stats.avg_sp_fracture[L_index][b_index]);
		stats.var_sp_backbone[L_index][b_index] += (sp_backbone[i] - stats.avg_sp_backbone[L_index][b_index]) *
		                                           (sp_backbone[i] - stats.avg_sp_backbone[L_index][b_index]);
	    }
	    stats.var_updated_mass[L_index][b_index] /= runs;
	    stats.var_fracture_mass[L_index][b_index] /= runs;
	    stats.var_backbone_mass[L_index][b_index] /= runs;
	    stats.var_sp_updated[L_index][b_index] /= runs;
	    stats.var_sp_fracture[L_index][b_index] /= runs;
	    stats.var_sp_backbone[L_index][b_index] /= runs;


	    free(updated_mass);
	    free(fracture_mass);
	    free(backbone_mass);
	    free(sp_updated);
	    free(sp_fracture);
	    free(sp_backbone);


	    //runs = (runs/2 > 20) ? runs/2 : 20;
	    runs = 25;
	}  /*  end for size  */
	


	stop = walltime();
	total_time += stop-start;

	printf("Finished run for theta = %.2f in %.3f\n\n", b, total_time);


    }  //end for b

    // print all stats to file
    fprintStats(&stats);

    freeStats(&stats);

}



void checkCurrent(double b_min, double b_max, double b_step, double dV){
    struct System A;
    int **backbone;
    int index, arr_size, size, zero_current;
    double *current, *voltage, *ideal_current;
    double b, c, macro_threshold;    
    FILE *outfile;
    char filename[40];

    size =15;
    

    for(b=b_min; b<b_max; b+=b_step){

	zero_current = 0;
	index = 0;
	arr_size = 1024;
	current = calloc(arr_size, sizeof(current));
	voltage = calloc(arr_size, sizeof(voltage));
	ideal_current = calloc(arr_size, sizeof(ideal_current));


	allocSys(&A, size);
	initSys(&A, b, size+16, 0, 1);

	A.max_voltage = dV;
	index = 0;

	backbone = NULL;
	/* while( A.num_updated_bonds < 2*size || !HK(backbone, NULL, A.L)){ */
	
	/*     CG(&A); */

	/*     /\*  If the conductance is updated, CG is run again as *\/ */
	/*     /\*  more bonds might break                            *\/ */

	/*     while(updateConductance(&A)) */
	/* 	CG(&A); */

	/*     if(backbone){ */
	/* 	for(int i=0; i<size*size; i++) */
	/* 	    free(backbone[i]); */
	/* 	free(backbone); */
	/*     } */
	/*     backbone = getBackbone(A.updated, A.L); */

	/*     A.max_voltage += dV; */
	/* } */

	while( A.num_updated_bonds < 2*size || !HKnpb(A.updated, NULL, A.L)){
	
	    CG(&A);

	    /*  If the conductance is updated, CG is run again as */
	    /*  more bonds might break                            */

	    while(updateConductance(&A))
		CG(&A);
		    	
	    if(index == arr_size-1){
		arr_size *= 2;
		voltage = realloc(voltage, arr_size*sizeof(voltage));
		current = realloc(current, arr_size*sizeof(current));
		ideal_current = realloc(ideal_current, arr_size*sizeof(ideal_current));
	    }	    
	    voltage[index] = A.max_voltage;
	    current[index] = A.total_current;
	    ideal_current[index] = voltage[index];
	    index++;
	    A.max_voltage += dV;
	}
	backbone = getBackbone(A.updated, A.L);
	macro_threshold = A.max_voltage - dV;

	for(double v = A.max_voltage; A.max_voltage < v+10; A.max_voltage += dV){
	    CG(&A);

	    /*  If the conductance is updated, CG is run again as */
	    /*  more bonds might break                            */

	    while(updateConductance(&A))
		CG(&A);

	    if(index == arr_size-1){
		arr_size *= 2;
		voltage = realloc(voltage, arr_size*sizeof(voltage));
		current = realloc(current, arr_size*sizeof(current));
		ideal_current = realloc(ideal_current, arr_size*sizeof(ideal_current));
	    }
	    c = (1-b)*macro_threshold/A.max_voltage + b;
	    voltage[index] = A.max_voltage;
	    current[index] = A.total_current;
	    ideal_current[index] = voltage[index]*c;
	    if(A.total_current < 1e-2)
		zero_current = 1;
	    if(zero_current)
		current[index] = 0;
	    index++;
	}

	sprintf(filename, "checkCurrent-b %.3f.dat", b);		
	outfile = fopen(filename, "w");
	fprintf(outfile, "#current for b = %.4f\n", b);
	fprintf(outfile, "#voltage\t#current\t#ideal current\t#macro_threshold\n");
	for(int i=0; i<=index; i++)
	    fprintf(outfile, "%f\t%f\t%f\t%f\n", voltage[i], current[i], ideal_current[i], macro_threshold);

	fclose(outfile);

	assert(HK(backbone, NULL, A.L));  /*  make sure backbone spans network  */		
		
	/*  Free memory  */

	assert(backbone);  //This should not be a null pointer
	for(int i=0; i<size*size; i++)
	    free(backbone[i]);
	free(backbone);
	freeSys(&A);
	free(current);
	free(voltage);
	free(ideal_current);

    }    
}


int dblcmp(const void *p1, const void *p2){
    double x = *(double *)p1;
    double y = *(double *)p2;
    return x <= y ? (x < y ? -1 : 0) : 1;
}


void printToPS(struct System *sys, char* filename){

    //Creates a postscript file which draws the system, largest fracture, backbone and shortest path

    int l = 8;   //Standard line length
    int L = sys->L;
    //char *filename;
    FILE *outfile;
    int **shortest_path = shortestPath(sys->threshold, sys->L);
    int **fracture = getFracture(sys->updated, sys->L);
    int **backbone = getBackbone(sys->updated, sys->L);
    int **sp = calloc(L*L, sizeof(sp));
    int *sp_pathscape = calloc(L*(L+1), sizeof(sp_pathscape));
    for(int i=0; i<L*L; i++)
	sp[i] = calloc(2, sizeof(sp[i]));
    int end, ij, index;

    double lt;             //lowest threshold
    int tbond[2];          //bond with the lowest threshold
    int num_fbonds = 0;    //number of broken bonds
    int num_tbonds = 0;    //number of bonds in tbonds
    int **t_updated = calloc(L*L, sizeof(t_updated));
    for(int i=0; i<L*L; i++){
	t_updated[i] = calloc(2, sizeof(t_updated[i]));
	num_fbonds += sys->updated[i][0];
	num_fbonds += sys->updated[i][1];       
    }
    
    tbond[0] = 0;
    tbond[1] = 0;
    while(num_tbonds < num_fbonds){
	lt = 2;
	for(int i=0; i<L*L; i++){
	    if(!t_updated[i][0] && sys->threshold[i][0] < lt){
		lt = sys->threshold[i][0];
		tbond[0] = i;
		tbond[1] = 0;
	    }
	    if(!t_updated[i][1] && sys->threshold[i][1] < lt){
		lt = sys->threshold[i][1];
		tbond[0] = i;
		tbond[1] = 1;
	    }
	}
	t_updated[tbond[0]][tbond[1]] = 1;
	num_tbonds++;
    }
    
    
    
    //filename = "OutputPS/drawsys.ps";

    outfile = fopen(filename, "w");

    //Start of file, define certain variables. i is the system size L, l is the standard line length (in pixels),
    //j is the size of the drawn system in pixels (image is of size j*j), and l_neg is -l, as that does not work in RPN
    fprintf(outfile, "%%!PS\n\n");
    fprintf(outfile, "/i %d def\n /j %d def\n /l %d def\n /l_neg %d def\n\n", L, L*l, l, -l);

    //Draw electrode function
    fprintf(outfile, "/drawelectrode{\n newpath\n 0 0 moveto\n j 0 lineto\n 0 setgray\n 6 setlinewidth \n stroke\n");
    fprintf(outfile, "0 j moveto\n j j lineto\n 3 setlinewidth \n stroke\n 1 setlinewidth\n } def\n\n");

    //Draw grid function
    fprintf(outfile, "/drawgrid{\n newpath\n 0 0 moveto\n /k 0 def\n");
    fprintf(outfile, "k l j 1 sub{\n k 0 moveto\n 0 1 i 1 sub{\n l l rlineto\n l_neg l_neg rmoveto\n 0 l rlineto\n} for\n");
    fprintf(outfile, "/k k l add def\n} for\n .7 setgray\n stroke\n} def\n\n");


    /* //Draw fracture function (bonds)*/
    /* fprintf(outfile, "/drawfracture{\n newpath\n"); */
    /* for(int i=0; i<L; i++) */
    /* 	for(int j=0; j<L; j++){ */
    /* 	    ij = i*L+j; */
    /* 	    if(fracture[ij][0]) */
    /* 		fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l, j*l, 0, l); */
    /* 	    if(fracture[ij][1]) */
    /* 		fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l, j*l, l, l); */
    /* 	} */

    /* fprintf(outfile, "1 0 0 setrgbcolor\n 2 setlinewidth\n stroke\n } def\n\n"); */

   
    index = 0;
    while(shortest_path[index][1] >= 0){
	sp[ shortest_path[index][0] ][ shortest_path[index][1] ] = 1;
	index++;
    }
    

    //Draw shortest path (commented lines color the bonds that the path crosses)
    fprintf(outfile, "/drawsp{\n newpath\n");
    for(int i=0; i<L; i++)
	for(int j=0; j<L; j++){
	    ij = i+j*L;
	    if(sp[ij][0]){
		//fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l, j*l, 0, l);
		fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l-l/2, j*l, l, l);
		if(i == 0)
		    //fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", l*L, j*l, 0, l);
		    fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", L*l-l/2, j*l, l/2, l/2);
	    }
	    if(sp[ij][1])
		//fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l, j*l, l, l);
		fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l+l/2, j*l+l, 0, -l);
	}

    fprintf(outfile, "1 0 0 setrgbcolor\n 1 setlinewidth\n 1 setlinecap\n stroke\n } def\n\n");



    // Draw pathscape
    double *srt_arr = calloc(L*(L+1), sizeof(srt_arr));
    memcpy(srt_arr, sys->pathscape, L*(L+1)*sizeof(srt_arr[0]));
    qsort(srt_arr, L*(L+1), sizeof(srt_arr[0]), dblcmp);

    // 0 <= cval <= 1 = (path_val-min_val) / (max_val-min_val)
    // 0 <= rval, gval, cval <= 1 are values for red, green and blue
    double den, min_val, max_val, cval, rval, gval, bval;
    min_val = srt_arr[0];
    max_val = srt_arr[L*(L+1)-1];
    den = 1/(max_val-min_val);

    fprintf(outfile, "/drawpathscape{\n newpath\n");

    for(int i=0; i<L+1; i++)
	for(int j=0; j<L; j++){
	    ij = i*L+j;
	    if( i==L && j == L){
		printf("Lollers, %d, %.4f\n", ij, sys->pathscape[0]);
		continue;
	    }
	    cval = (sys->pathscape[ij]-min_val)*den;
	    if(cval < 0.33333333){
		rval = cval*3;
		gval = 0;
		bval = 0;
	    }
	    else if(cval < 0.66666666){
		rval = 1;
		gval = (cval-0.33333333)*3;
		bval = 0;
	    }
	    else{
		rval = 1;
		gval = 1;
		bval = (cval-0.66666666)*3;
	    }

	    // Draw each point twice, due to some weird behaviour with postscript
	    // Commented lines draw smaller points, revealing the underlying grid
	    if(i == L){
	    	fprintf(outfile, "%d %d moveto\n %d %d rlineto\n %d %d rlineto\n closepath\n fill\n",
	    		j*l, i*l-l, l, l, -l, 0);
	    	fprintf(outfile, "%.4f %.4f %.4f setrgbcolor\n 1 setlinewidth\n 1 setlinecap\n stroke\n",
	    		rval, gval, bval);
	    	fprintf(outfile, "%d %d moveto\n %d %d rlineto\n %d %d rlineto\n closepath\n fill\n",
	    		j*l, i*l-l, l, l, -l, 0);
	    	fprintf(outfile, "%.4f %.4f %.4f setrgbcolor\n 1 setlinewidth\n 1 setlinecap\n stroke\n",
	    		rval, gval, bval);
	    	/* fprintf(outfile, "%d %d moveto\n %d %d rlineto\n %d %d rlineto\n closepath\n fill\n", */
	    	/* 	j*l+1, i*l-l+3, l-2, l-2, -l+2, 0); */
	    	/* fprintf(outfile, "%.4f %.4f %.4f setrgbcolor\n 1 setlinewidth\n 1 setlinecap\n stroke\n", */
	    	/* 	rval, gval, bval); */
	    	/* fprintf(outfile, "%d %d moveto\n %d %d rlineto\n %d %d rlineto\n closepath\n fill\n", */
	    	/* 	j*l+1, i*l-l+3, l-2, l-2, -l+2, 0); */
	    	/* fprintf(outfile, "%.4f %.4f %.4f setrgbcolor\n 1 setlinewidth\n 1 setlinecap\n stroke\n", */
	    	/* 	rval, gval, bval); */

	    }
	    else{
		fprintf(outfile, "%d %d moveto\n %d %d rlineto\n %d %d rlineto\n %d %d rlineto\n closepath\n fill\n",
			j*l, i*l-l, l, l, 0, l, -l, -l);
		fprintf(outfile, "%.4f %.4f %.4f setrgbcolor\n 1 setlinewidth\n 1 setlinecap\n stroke\n",
			rval, gval, bval);
		fprintf(outfile, "%d %d moveto\n %d %d rlineto\n %d %d rlineto\n %d %d rlineto\n closepath\n fill\n",
			j*l, i*l-l, l, l, 0, l, -l, -l);
		fprintf(outfile, "%.4f %.4f %.4f setrgbcolor\n 1 setlinewidth\n 1 setlinecap\n stroke\n",
			rval, gval, bval);
		/* fprintf(outfile, "%d %d moveto\n %d %d rlineto\n %d %d rlineto\n %d %d rlineto\n closepath\n fill\n", */
		/* 	j*l+1, i*l-l+3, l-2, l-2, 0, l-4, -l+2, -l+2); */
		/* fprintf(outfile, "%.4f %.4f %.4f setrgbcolor\n 1 setlinewidth\n 1 setlinecap\n stroke\n", */
		/* 	rval, gval, bval); */
		/* fprintf(outfile, "%d %d moveto\n %d %d rlineto\n %d %d rlineto\n %d %d rlineto\n closepath\n fill\n", */
		/* 	j*l+1, i*l-l+3, l-2, l-2, 0, l-4, -l+2, -l+2); */
		/* fprintf(outfile, "%.4f %.4f %.4f setrgbcolor\n 1 setlinewidth\n 1 setlinecap\n stroke\n", */
		/* 	rval, gval, bval); */
	    }
	}

    fprintf(outfile, "} def\n\n");
    free(srt_arr);

    //Draw shortest path from pathscape
    fprintf(outfile, "/drawoptpath{\n newpath\n");
    for(int i=0; i<L+1; i++)
	for(int j=0; j<L; j++){
	    ij = i*L+j;
	    if( fabs(sys->pathscape[ij] - min_val) < 1e-5)
		fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", j*l-l/2-l/8+l, i*l, l/4, -l/8);
	}

    fprintf(outfile, "0 0 0 setrgbcolor\n 1 setlinewidth\n 1 setlinecap\n stroke\n } def\n\n");


    //Draw largest fracture
    if(fracture){
	fprintf(outfile, "/drawfracture{\n newpath\n");
	for(int i=0; i<L; i++)
	    for(int j=0; j<L; j++){
		ij = i+j*L;
		if(fracture[ij][0]){
		    //fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l, j*l, 0, l);
		    fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l-l/2, j*l, l, l);
		    if(i == 0)
			//fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", l*L, j*l, 0, l);
			fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", L*l-l/2, j*l, l/2, l/2);
		}
		if(fracture[ij][1])
		    //fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l, j*l, l, l);
		    fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l+l/2, j*l+l, 0, -l);
	    }
	
	fprintf(outfile, "0 1 0 setrgbcolor\n 0.5 setlinewidth\n 1 setlinecap\n stroke\n } def\n\n");
    }

    //Draw lowest thresholds corresponding to number of fractures
    fprintf(outfile, "/drawlt{\n newpath\n");
    for(int i=0; i<L; i++)
	for(int j=0; j<L; j++){
	    ij = i+j*L;
	    if(t_updated[ij][0]){
		//fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l, j*l, 0, l);
		fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l-l/2, j*l, l, l);
		if(i == 0)
		    //fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", l*L, j*l, 0, l);
		    fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", L*l-l/2, j*l, l/2, l/2);
	    }
	    if(t_updated[ij][1])
		//fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l, j*l, l, l);
		fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l+l/2, j*l+l, 0, -l);
	}

    fprintf(outfile, "0 0 1 setrgbcolor\n 2 setlinewidth\n stroke\n } def\n\n");



    //Draw backbone
    /* if(checkForLoops(sys->perm_updated, L, -90, 1)) */
    /* 	printf("\t...on perm_updated in printToPS()\n"); */
    //backbone = removeDanglingArms(sys->perm_updated, L);
    /* if(checkForLoops(backbone, L, -90, 1)) */
    /* 	printf("\t...on backbone in printToPS()\n"); */
    if(backbone){
    fprintf(outfile, "/drawbackbone{\n newpath\n");
    for(int i=0; i<L; i++)
	for(int j=0; j<L; j++){
	    ij = i+j*L;
	    if(backbone[ij][0]){
		//fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l, j*l, 0, l);
		fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l-l/2, j*l, l, l);
		if(i == 0)
		    //fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", l*L, j*l, 0, l);
		    fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", L*l-l/2, j*l, l/2, l/2);
	    }
	    if(backbone[ij][1])
		//fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l, j*l, l, l);
		fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l+l/2, j*l+l, 0, -l);
	}

    fprintf(outfile, "0 0 1 setrgbcolor\n 1 setlinewidth\n 1 setlinecap\n stroke\n } def\n\n");
    }

    //Draw updated
    fprintf(outfile, "/drawupdated{\n newpath\n");
    for(int i=0; i<L; i++)
	for(int j=0; j<L; j++){
	    ij = i+j*L;
	    if(sys->perm_updated[ij][0]){
		//fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l, j*l, 0, l);
		fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l-l/2, j*l, l, l);
		if(i == 0)
		    //fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", l*L, j*l, 0, l);
		    fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", L*l-l/2, j*l, l/2, l/2);
	    }
	    if(sys->perm_updated[ij][1])
		//fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l, j*l, l, l);
		fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l+l/2, j*l+l, 0, -l);
	}

    fprintf(outfile, "0 0 0 setrgbcolor\n 3 setlinewidth\n 1 setlinecap\n stroke\n } def\n\n");

    //Draw updated bonds
    fprintf(outfile, "/drawubonds{\n newpath\n 0 0 moveto\n /k 0 def\n");

    for(int i=0; i<L; i++)
	for(int j=0; j<L; j++){
	    ij = i+j*L;
	    if(sys->perm_updated[ij][0]){
		//fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l, j*l, 0, l);
		fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l, j*l, 0, l);
	    }
	    if(sys->perm_updated[ij][1])
		//fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l, j*l, l, l);
		fprintf(outfile, "%d %d moveto\n %d %d rlineto\n", i*l, j*l, l, l);
	}
    fprintf(outfile, "0 0 1 setrgbcolor\n 2 setlinewidth\n stroke\n } def\n\n");
    
    fprintf(outfile, 
	    "drawgrid \ndrawpathscape \n%%drawupdated \ndrawfracture  \ndrawsp \n%%drawoptpath \n%%drawlt \ndrawbackbone \n%%drawubonds \ndrawelectrode\n");

    end = INT_MAX;
    for(int i=0; i<end; i++){
	if(shortest_path[i][1] < 0)
	    end = shortest_path[i][0];
	free(shortest_path[i]);
    }		
    free(shortest_path);
    free(sp_pathscape);

    for(int i=0; i<L*L; i++){
	if(fracture)
	    free(fracture[i]);
	if(backbone)
	    free(backbone[i]);
	free(sp[i]);
    }
    if(fracture)
	free(fracture);
    if(backbone)
	free(backbone);
    free(sp);

    for(int i=0; i<L*L; i++)
	free(t_updated[i]);
    
    free(t_updated);


    fclose(outfile);
}



void drawSys(){

    popen("gs drawsys.ps", "r");

}




void breakLT(struct System *sys){

    int L = sys->L;
    double t;
    double lt = 100000;
    int lt_coord = 0;
    int lt_bond = 0;

    for(int i=0; i<L*L; i++){
	for(int k=0; k<2; k++){
	    t = sys->threshold[i][k];
	    if(t < lt && !sys->updated[i][k]){
		lt = t;
		lt_coord = i;
		lt_bond = k;
	    }
	}
    }
    sys->updated[lt_coord][lt_bond] = 1;    
}



void dVDist(struct System *sys, double theta){
    int L = sys->L;
    double **dV = calloc(L*L, sizeof(dV));
    int *ir = calloc(L, sizeof(ir));
    for(int i=0; i<L; i++)
	ir[i] = i+1;
    ir[L-1] = 0;
    
    for(int i=0; i<L*L; i++)
    	dV[i] = calloc(2, sizeof(dV));


    // This first run through can probably hold the general bond updated
    // which is done further down. Then, if count > 0, the ltv_bond is changed.
    for(int j=0; j<L; j++)
	for(int i=0; i<L; i++){
	    for(int k=0; k<2; k++){
		if(j < L-1 && k == 0)
		    dV[i+j*L][k] = fabs(sys->voltage[i+(j+1)*L] - sys->voltage[i+j*L]);
		else if(j < L-1 && k == 1)
		    dV[i+j*L][k] = fabs(sys->voltage[ir[i] + (j+1)*L] - sys->voltage[i+j*L]);
		else
		    dV[i+j*L][k] = fabs(sys->max_voltage - sys->voltage[i+j*L]);
	    }
	}

    FILE *out;
    char *filename = calloc(40, sizeof(filename));
    sprintf(filename, "dV_dist_theta_%.2f.dat", theta);
    out = fopen(filename, "w");
    int count = 0;
    fprintf(out, "#updated+0.01*theta \t #dV/max_voltage \t #threshold\n\n");
    for(int j=0; j<L; j++)
	for(int i=0; i<L; i++)
	    for(int k=0; k<2; k++){
		count++;
		fprintf(out, "%.4f \t %.6f \t %.6f\n", 
			sys->updated[i+j*L][k]+0.01*theta, dV[i+j*L][k], sys->threshold[i+j*L][k]);
	    }

    fclose(out);
    free(filename);
    for(int i=0; i<L*L; i++)
    	free(dV[i]);
    free(dV);
    free(ir);

}
				 


void writeBonds(){

    struct System A;//, B;
    
    int size;

    size = 64;

    allocSys(&A, size);

    for(int run = 0; run < 16; run++){
	printf("\t run %d\n", run);


	FILE *outfile;
	char *txtfile;
	txtfile = calloc(50, sizeof(char));

	sprintf(txtfile, "OutputPS/L%d/bonds_run%d.txt", size, run);
	outfile = fopen(txtfile, "w");
	
	
	initSys(&A, 0, size+run+81, -1, 1);

	for(int i=size-1; i >= 0; i--){
	    for(int j=0; j < size; j++)
		fprintf(outfile, "%.8f %.8f ",
			A.threshold[i*size + j][0], A.threshold[i*size + j][1]);
	    fprintf(outfile, "\n");
	}

	fclose(outfile);
	free(txtfile);
    }

    freeSys(&A);
}

void checkPrint(){

  //struct System A;//, B;
    
    int size;
    double b, dV;

    size = 128;
    
    dV = 0.001;
    b = -90;

    char *dir_path;
    dir_path = calloc(50, sizeof(dir_path));
    sprintf(dir_path, "OutputPS/L%d", size);

    if(opendir(dir_path) == NULL){ // could not open directory
	if(errno != 2){ // errno == 2 => directory does not exist, go to else
	    printf("Something wrong when checking if %s exists! Error: %s\n\n",
		   dir_path, strerror(errno));
	    exit(1);
	}
	else{ // mkdir w. read, write, exec. permissions for owner
	    if(mkdir(dir_path, S_IRWXU) == -1){  		
		printf("Something wrong when creating directory %s! Error: %s\n\n",
		       dir_path, strerror(errno));
		exit(1);
	    }
	}
    }


    free(dir_path);

    //allocSys(&A, size);
    //allocSys(&B, size);
    double total_start = walltime();

    // #pragma omp parallel for num_threads(2) schedule(dynamic) private(b)
    for(int run = 0; run < 20; run++){
	printf("\t run %d\n", run);



	//double b_list[13] = {-90, -60, -30, 0, 5, 10, 15, 20, 25, 30, 35, 40, 45};
	double b_list[4] = {2, 5, 10, 15};

	//#pragma omp parallel for num_threads(2) schedule(dynamic) private(b)

	struct System A;
	allocSys(&A, size);

        /* omp_set_nested(1); */
	/* omp_set_dynamic(0); */
	/* omp_set_num_threads(2); */

	for(int r=0; r<4; r++){


	    if(r == 4){
	
		double start, stop;
	
		start = walltime();
		// Initialize A and write pathscape to file
		initSys(&A, 0, size+run+81, -1, 1);
		optimalPathscape(&A);

#pragma omp critical
		{
		    fwritePathscape(&A, run);
		}

		stop = walltime();
		printf("Run %d: Time to get pathscape: %.4fs\n", run, stop-start);
	    }

	    else{

		double start, stop;
		b = b_list[r]; 
		start = walltime();
		printf("theta = %.2f, no trapping\n", b);
		//if(b == 0)
		//initSys(&A, b, size+run+81, -1, 2);
		//else
		initSys(&A, b, size+run+81, -1, 1);
	
		A.max_voltage = dV;

		if((int)b == -90)
		    breakFuseNetwork(&A, run, b);
		int bnA, bnB;
		int breaklimit = 4*size*size;
		int updated_count = 0;
		bnA = (b > 50); // Ensures we're not moving into the superconducting regime
		bnB = 1;//(b > -50);
		while( !bnA || !bnB ){//|| A.max_voltage < 30){

		    CG(&A);

		    updated_count = 0;

		    if(!bnA){
			while(!bnA && (/* updateConductancePlastic23(&A) || */ updateConductanceSingleElastic(&A))){
			    CG(&A);		
			    updateDV(&A);
			    bnA = HKV(A.updated, A.L);

			    updated_count++;
		
			    if(updated_count > breaklimit){
				printf("Breaklimit! Max voltage = %.4f\n", A.max_voltage);
				break;
			    }
			}
		    }

		    if(!bnA)
			A.max_voltage += dV;
		}
       

		breakBonds(&A);
       
		printf("    A:  Max voltage: %.4f, updated bonds: %d\n", A.max_voltage, A.num_updated_bonds);

		stop = walltime();
		printf("Run %d, no trapping: Time to break network: %.4fs\n", run, stop-start);


		int *nodes = calloc((size+1)*(size+1)+1, sizeof(nodes));  /*  is initialized in HK  */

		int **sc = getFracture(A.updated, size);
		int **backbone = getBackbone(A.updated, size);
		HK(A.updated, nodes, size);

		//#pragma omp critical
		{
		    fwriteBackbone(size, run, b, backbone, sc, NULL, 45);
		}


		for(int i=0; i<size*size; i++){
		    free(sc[i]);
		    free(backbone[i]);
		}
	
		free(sc);
		free(backbone);
		free(nodes);
	    }
	}
	freeSys(&A);
    }
    
    double total_stop = walltime();

    //    freeSys(&B);
    //freeSys(&A);
    
    printf("Finished run in %.4fs\n", total_stop-total_start);
   
}



void checkPrintTrapping(){

    //struct System A, B;
    
    int size;
    double b, dV;

    size = 128;
    
    dV = 0.001;
    b = -90;

    //allocSys(&A, size);
    //allocSys(&B, size);
    double total_start = walltime();

    /* omp_set_nested(1); */
    /* omp_set_dynamic(0); */
    //pragma omp parallel for num_threads(2) private(b) schedule(dynamic)
    for(int run = 0; run < 20; run++){
	printf("\t run %d\n", run);

	double start, stop;

	struct System A, B;
	allocSys(&A, size);
	allocSys(&B, size);

	/* start = walltime(); */
	initSys(&A, 0, size+run+81, -1, 1);
	/* optimalPathscape(&A); */

	/* #pragma omp critical */
	/* { */

	/*     txtfile = calloc(50, sizeof(char)); */
	/*     sprintf(txtfile, "test_trapping/L%d/pathscape_run%d.txt", size, run); */

	/*     //sprintf(txtfile, "pathscape_test%d_L16.txt", run); */
	/*     outfile = fopen(txtfile, "w");	 */

       
	/*     // Initialize A and write pathscape to file */
	/*     for(int i=size; i >= 0; i--){ */
	/* 	for(int j=0; j < size; j++) */
	/* 	    fprintf(outfile, "%.6f ", A.pathscape[i*size + j]); */
	/* 	fprintf(outfile, "\n"); */
	/*     } */
	/*     fclose(outfile); */
	/*     free(txtfile); */

	/*     txtfile = calloc(50, sizeof(char)); */
 
	/*     sprintf(txtfile, "test_trapping/L%d/pathscape_bonds_run%d.txt", size, run); */
	/*     //sprintf(txtfile, "pathscape_bonds_test%d_3L.txt", run); */
	/*     outfile = fopen(txtfile, "w"); */
	
	/*     for(int i=size; i >= 0; i--){ */
	/* 	for(int j=0; j < size; j++) */
	/* 	    fprintf(outfile, "(%d,%d) ", */
	/* 		    A.pathscape_bonds[i*size + j][0], A.pathscape_bonds[i*size + j][1]); */
	/* 	fprintf(outfile, "\n"); */
	/*     } */

	/*     fclose(outfile); */
	/*     free(txtfile); */
	/* } */

	/* stop = walltime(); */
	/* printf("Time to get pathscape: %.4fs\n", stop-start); */

	double b_list[4] = {2, 5, 10, 15};

	//for(b = 5; b <= 45; b += 5){
	for(int r = 0; r < 4; r++){

	    b = b_list[r]; 
	    start = walltime();
	    printf("theta = %.2f, trapping\n", b);
	    initSys(&A, b, size+run+81, -1, 1);
	    initSys(&B, b, size+run+81, -1, 1);
	
	    A.max_voltage = dV;
	    B.max_voltage = dV;

	    int bnA, bnB;
	    int breaklimit = 4*size*size;
	    int updated_count = 0;
	    int num_upd_bonds = 0;

	    double **old_dV = calloc(size*size, sizeof(old_dV));
	    for(int i=0; i<size*size; i++){
		old_dV[i] = calloc(2, sizeof(old_dV[i]));
	    }


	    bnA = (b > 50); // Ensures we're not moving into the superconducting regime
	    bnB = 1;//(b > -50);
	    while( !bnA || !bnB ){//|| A.max_voltage < 30){

		CG(&A);
		
		updated_count = 0;
		if(!bnA){
		    while(!bnA && updateConductanceSingleElastic(&A)){
		    /* while(!bnA && (updateConductancePlastic23(&A) || updateConductanceSingleElastic(&A))){ */

			if(A.num_updated_bonds > num_upd_bonds){
			    num_upd_bonds = A.num_updated_bonds;
			    for(int i=0; i<size*size; i++)
				for(int j=0; j<2; j++){
				    B.updated[i][j] = A.updated[i][j];
				    B.conductance[i][j] = !B.updated[i][j];				    
				}
			    
			    B.max_voltage = 1;
			    CG(&B);
			    updateDV(&B);
			    for(int i=0; i<size*size; i++)
				for(int j=0; j<2; j++)
				    old_dV[i][j] = B.max_dV[i][j];
			    
			    B.max_voltage = 100000;
			    CG(&B);
			    updateDV(&B);
			    for(int i=0; i<size*size; i++)
				for(int j=0; j<2; j++)
				    if(fabs(B.max_dV[i][j] - old_dV[i][j]) < 1e-5
				       && A.threshold[i][j] < 100){
					B.updated[i][j] = 1;
				    }

			    if(!HKV(B.updated, B.L)){
				for(int i=0; i<size*size; i++)
				    for(int j=0; j<2; j++)
					if(fabs(B.max_dV[i][j] - old_dV[i][j]) < 1e-5
					   && A.threshold[i][j] < 100 && A.updated[i][j] == 0){
					    A.threshold[i][j] = DBL_MAX;
					    //A.threshold[i][j] = A.threshold[i][j];
					    //printf("Setting threshold %d %d to DBL_MAX, V = %.4f\n",
					    //   i, j, A.max_voltage);
					}
			    }
			    else
				for(int i=0; i<size*size; i++)
				    for(int j=0; j<2; j++)
					if(fabs(B.max_dV[i][j] - old_dV[i][j]) < 1e-5
					   && A.threshold[i][j] < 100)
					    B.updated[i][j] = 0;
			}



			CG(&A);
			updateDV(&A);
			bnA = HKV(A.updated, A.L);

			updated_count++;
		
			if(updated_count > breaklimit){
			    printf("Breaklimit! Max voltage = %.4f\n", A.max_voltage);
			    break;
			}
		    }
		}

		if(!bnA){
		    A.max_voltage += dV;

		}
		
	    }
       
	    for(int i=0; i<size*size; i++)
		free(old_dV[i]);
	    free(old_dV);

	    breakBonds(&A);
       
	    printf("    A:  Max voltage: %.4f, updated bonds: %d\n", A.max_voltage, A.num_updated_bonds);

	    stop = walltime();
	    printf("Time to break network: %.4fs\n", stop-start);


	    int *nodes = calloc((size+1)*(size+1)+1, sizeof(nodes));  /*  is initialized in HK  */

	    int **sc = getFracture(A.updated, size);
	    int **backbone = getBackbone(A.updated, size);
	    HK(A.updated, nodes, size);

	    FILE *outfile;
	    char *txtfile;

	    //#pragma omp critical
	    {
	    txtfile = calloc(50, sizeof(char));
	    sprintf(txtfile, "test_trapping/L%d/scb_run%d_theta%d.txt", size, run, (int)b);
	    outfile = fopen(txtfile, "w");
	
	
	    fprintf(outfile, "# Cols 1-3 contain coords (x, y) and bond (0 or 1) of bonds belonging to the spanning cluster\n");
	    fprintf(outfile, "# Cols 4-6 contain coords (x, y) and bond (0 or 1) of bonds belonging to the backbone\n");
	    for(int i=0; i < size; i++){
	    	for(int j=0; j < size; j++){
	    	    for(int k=0; k<2; k++){
	    		if(sc[i*size+j][k]){
	    		    fprintf(outfile, "%d %d %d ", i, j, k);
	    		    if(backbone[i*size+j][k])
	    			fprintf(outfile, "%d %d %d", i, j, k);
	    		    else
	    			fprintf(outfile, "%d %d %d", -size, -size, size);
	    		}
	    		if(sc[i*size+j][k] || backbone[i*size+j][k])
	    		    fprintf(outfile, "\n");
	    	    }
	    	}
	    }


	    fclose(outfile);
	    free(txtfile);
	    }

	    for(int i=0; i<size*size; i++){
		free(sc[i]);
		free(backbone[i]);
	    }
	
	    free(sc);
	    free(backbone);
	    free(nodes);
	    
	    stop = walltime();
	    printf("Run %d, trapping: Time to break network: %.4fs\n", run, stop-start);
	}
	freeSys(&B);
	freeSys(&A);

    }

    double total_stop = walltime();

    /* freeSys(&B); */
    /* freeSys(&A); */
    
    printf("Finished run with trapping rules in %.4fs\n", total_stop-total_start);
   
}



void testVoltageUpdate(){

    struct System A, B;
    
    int size;
    double b, dV;
    int nthreads = 4;
    size = 64;
    
    dV = 0.001;
    b = 0;

    allocSys(&A, size);
    /* initSys(&A, 0, 81+3, -1, 2); */
    /* allocSys(&B, size); */
    /* initSys(&B, 0, 81+3, -1, 4); */


    double total_start = walltime();
    printf("Optimal pathscape, size = %d, nthreads = %d\n", size, nthreads);
    optimalPathscape(&A);
    double mid_time = walltime();
    printf("    Finished A in %.4fs, starting B\n", mid_time-total_start);
    optimalPathscape(&B);
    printf("    Finished B in %.4fs\n", walltime()-mid_time);
    int errors = 0;
    for(int i=0; i<size*(size+1); i++)
    	if(fabs(A.pathscape[i]-B.pathscape[i]) > 1e-15)
    	    errors++;
    printf("Errors: %d\n", errors);

    for(int run = 3; run < 2; run++){
	printf("\t run %d\n", run);
	
	for(int r = 0; r < 1; r++){

	    b = -90;
	    printf("theta = %.2f\n", b);
	    initSys(&A, b, size+run+81, -1, 2);
	
	    A.max_voltage = dV;

	    int bnA, bnB;
	    int breaklimit = 4*size*size;
	    int updated_count = 0;

	    double **old_dV;
	    int **old_updated;
	    old_dV = calloc(size*size, sizeof(old_dV));
	    old_updated = calloc(size*size, sizeof(old_updated));
	    for(int i=0; i<size*size; i++){
		old_dV[i] = calloc(2, sizeof(old_dV[i]));
		old_updated[i] = calloc(2, sizeof(old_updated[i]));
	    }
		
	    //double voltage_step = dV/size;
	    /* double diff;	     */
	    /* //int voltage_updated = 0; */
	    /* int upd_coords, upd_bond; */
	    /* int min_diff_coords = -1; */
	    /* int min_diff_bond = -1; */
	    /* int hits = 0; */
	    /* int misses = 0; */
	    /* int old_updated_count = 0; */
	    bnA = (b > 50);
	    bnB = -1;
	    while( !bnA || !bnB ){

		CG(&A);

		updated_count = 0;

		if(!bnA){
		    while(!bnA && (updateConductancePlastic23(&A) || updateConductanceSingleElastic(&A))){
			
			/* upd_coords = -1; */
			/* upd_bond = -1; */
			/* for(int i=0; i<size*size; i++) */
			/*     for(int j=0; j<2; j++){ */
			/* 	if(A.updated[i][j] != old_updated[i][j]){ */
			/* 	    upd_coords = i; */
			/* 	    upd_bond = j; */
			/* 	    //voltage_updated = 1; */
			/* 	} */
			/* 	old_updated[i][j] = A.updated[i][j]; */
			/* 	old_dV[i][j] = A.max_dV[i][j]*(!A.updated[i][j]); */
			/*     } */
			
			/* if(A.num_updated_bonds - old_updated_count == 1){ */
			/*     old_updated_count = A.num_updated_bonds; */
			/*     //voltage_updated = 0; */
			/*     if(min_diff_coords == upd_coords && min_diff_bond == upd_bond) */
			/* 	hits++; */
			/*     else */
			/* 	misses++; */
			/* } */

			CG(&A);
			updateDV(&A);
			bnA = HKV(A.updated, A.L);
			
			updated_count++;
			
			if(updated_count > breaklimit){
			    printf("Breaklimit! Max voltage = %.4f\n", A.max_voltage);
			    break;
			}
			
			/* //double max_diff = 0; */
			/* double min_diff = DBL_MAX; */
			/* //int max_min_updated = 0; */

			/* for(int i=0; i<size*size; i++) */
			/*     for(int j=0; j<2; j++){ */
			/* 	if(!A.updated[i][j]){ */
			/* 	    //diff = fabs(A.max_dV[i][j]-old_dV[i][j]); */
			/* 	    diff = fabs(A.threshold[i][j]/A.max_dV[i][j]); */
			/* 	    min_diff_coords = diff < min_diff ? i : min_diff_coords; */
			/* 	    min_diff_bond = diff < min_diff ? j : min_diff_bond; */
			/* 	    min_diff = diff < min_diff ? diff : min_diff; */
			/* 	} */
			/*     } */



			/* if(voltage_updated){ */
			/*     voltage_updated = 0; */
			/*     for(int i=0; i<size*size; i++) */
			/* 	for(int j=0; j<2; j++){ */
			/* 	    diff = fabs(A.max_dV[i][j]-old_dV[i][j]); */
			/* 	    if(fabs(diff - voltage_step) > 1e-8 && !A.updated[i][j]  */
			/* 	       && A.num_updated_bonds == 60){ */
			/* 		//printf("Whaa! %d, %d, %.8f, %.8f\n", */
			/* 		//       i, j, diff, voltage_step); */
			/* 		max_diff = diff > max_diff ? diff : max_diff; */
			/* 		min_diff = diff < min_diff ? diff : min_diff; */
			/* 		max_min_updated = 1; */
			/* 	    } */
			/* 	} */
			/* } */
			/* if(max_min_updated) */
			/*     printf("max_diff: %.3e \t min_diff: %.3e \t max/min: %.3e \t vstep: %.3e\n", */
			/* 	   max_diff, min_diff, max_diff/min_diff, voltage_step); */
			
		    }
		}
		
		if(!bnA){
		    A.max_voltage += dV;
		    //voltage_updated = 1;
		}
	    }
	    
	    /* printf("Hits: %d \t Misses: %d \t %%Hits: %.3f\n\n",  */
	    /* 	   hits, misses, (double)hits/(hits+misses)); */
	    
	    for(int i=0; i<size*size; i++){
		free(old_dV[i]);
		free(old_updated[i]);
	    }
	    free(old_dV);
	    free(old_updated);
	    
	    breakBonds(&A);
	    
	    printf("    A:  Max voltage: %.4f, updated bonds: %d\n", A.max_voltage, A.num_updated_bonds);
	    
	    
	    int *nodes = calloc((size+1)*(size+1)+1, sizeof(nodes));  /*  is initialized in HK  */
	    
	    int **sc = getFracture(A.updated, size);
	    int **backbone = getBackbone(A.updated, size);
	    HK(A.updated, nodes, size);
	    
	    for(int i=0; i<size*size; i++){
		free(sc[i]);
		free(backbone[i]);
	    }
	    
	    free(sc);
	    free(backbone);
	    free(nodes);
	}	
    }

    double total_stop = walltime();

    //    freeSys(&B);
    freeSys(&A);
    
    printf("Finished run in %.4fs\n\n", total_stop-total_start);
   
}



void restrictTest(){

    struct System A;     
    int size = 64;
    int runs = 3;
    double start, stop, tot_start, tot_stop;
    allocSys(&A, size);

    printf("Running %d systems of size %d, with restrict\n", runs, size);
    tot_start = walltime();
    
    for(int run = 0; run < runs; run++){
	start = walltime();
	printf("Run %d ", run);
	initSys(&A, 0, size+run+81, -1, 1);
	optimalPathscape(&A);
	stop = walltime();
	printf("finished in %.6fs\n", stop-start);
    }

    tot_stop = walltime();
    printf("Total runtime: %.6f \nAvgerage runtime: %.6f\n", tot_stop-tot_start, (tot_stop-tot_start)/runs);

    freeSys(&A);
}
void pathscapeTest(int size, int runs){


    struct System A;
    
    /* int size = 32; */
    /* int runs = 20; */
    
    printf("\nSize: %d \t runs: %d\n\n", size, runs);

    allocSys(&A, size);

    int L = A.L, N = 2*L, l = L/2;
    int *conv_col;

    double **nodes, **nodes2, **temp_thresholds;
    int **bonds;

    conv_col = calloc(N, sizeof(conv_col));
    for(int i=0; i<l; i++)
	conv_col[i] = i+l;
    for(int i=l; i<l+L; i++)
	conv_col[i] = i-l;
    for(int i=l+L; i<N; i++)
	conv_col[i] = i-l-L;

    nodes = calloc(2*(L+1), sizeof(nodes));
    nodes2 = calloc(2*(L+1), sizeof(nodes2));
    bonds = calloc(2*(L+1), sizeof(bonds));
    temp_thresholds = calloc(N*L, sizeof(temp_thresholds));
    for(int i=0; i<2*(L+1); i++){
	nodes[i] = calloc(N*(L+1), sizeof(nodes[i]));
	nodes2[i] = calloc(N*(L+1), sizeof(nodes2[i]));
	bonds[i] = calloc(N*(L+1), sizeof(bonds[i]));
    
    }
    for(int i=0; i<L; i++){
	for(int j=0; j<N; j++){
	    temp_thresholds[ i*N+j ] = calloc(2, sizeof(temp_thresholds[i*N+j]));
	}
    }

    double total_start, total_stop, start, stop, 
	avg_old, avg_new, max_old, max_new, min_old, min_new;

    min_old = DBL_MAX;
    min_new = DBL_MAX;
    max_old = 0;
    max_new = 0;

    initSys(&A, 0, size+0+81, -1, 1);

    for(int i=0; i<L; i++){
	for(int j=0; j<N; j++){
	    temp_thresholds[ i*N+j ][0] = A.threshold[ i*L+conv_col[j] ][0];
	    temp_thresholds[ i*N+j ][1] = A.threshold[ i*L+conv_col[j] ][1];
	}
    }

    //start = walltime();
    //primeOptimalPathscape(nodes, temp_thresholds, bonds, L, 1e-10);   
    //calculateOptimalPathscape(nodes, temp_thresholds, bonds, L, 1e-10, 1);   
    //stop = walltime();

    //printf("Time used: %.6fs\n", stop-start);

    total_start = walltime();

    for(int run = 0; run < runs; run++){
    	//printf("Original primePathscape: \t run %d\n", run);
	
    	// Initialize A and write pathscape to file

    	initSys(&A, 0, size+run+81, -1, 1);

    	for(int i=0; i<L; i++){
    	    for(int j=0; j<N; j++){
    		temp_thresholds[ i*N+j ][0] = A.threshold[ i*L+conv_col[j] ][0];
    		temp_thresholds[ i*N+j ][1] = A.threshold[ i*L+conv_col[j] ][1];
    	    }
    	}

    	start = walltime();
	calculateOptimalPathscape(nodes, temp_thresholds, bonds, L, 1e-10, 1);
    	//primeOptimalPathscape(nodes, temp_thresholds, bonds, L, 1e-10);
    	stop = walltime();
	max_old = (stop-start > max_old) ? stop-start : max_old;
	min_old = (stop-start < min_old) ? stop-start : min_old;
    	//printf("\tTime to get pathscape: %.4fs\n", stop-start);

    }

    total_stop = walltime();
    avg_old = (total_stop-total_start)/runs;

    printf("Average time, original: %.6f\n", avg_old);
    printf("Maximum time, original: %.6f\n", max_old);
    printf("Minimum time, original: %.6f\n", min_old);

    //calculateOptimalPathscape(nodes, temp_thresholds, bonds, L, 1e-10, 0);
    total_start = walltime();

    for(int run = 0; run < runs; run++){
    	//printf("primePathscape, no sums: \t run %d\n", run);
	
    	// Initialize A and write pathscape to file

    	initSys(&A, 0, size+run+81, -1, 1);

    	for(int i=0; i<L; i++){
    	    for(int j=0; j<N; j++){
    		temp_thresholds[ i*N+j ] = calloc(2, sizeof(temp_thresholds[i*N+j]));
    		temp_thresholds[ i*N+j ][0] = A.threshold[ i*L+conv_col[j] ][0];
    		temp_thresholds[ i*N+j ][1] = A.threshold[ i*L+conv_col[j] ][1];
    	    }
    	}

    	start = walltime();
	calculateOptimalPathscape(nodes2, temp_thresholds, bonds, L, 1e-10, 0);
    	//primeOptimalPathscapeNoSum(nodes, temp_thresholds, bonds, L, 1e-10);
    	stop = walltime();
	max_new = (stop-start > max_new) ? stop-start : max_new;
	min_new = (stop-start < min_new) ? stop-start : min_new;

    	//printf("\tTime to get pathscape: %.4fs\n", stop-start);

   } 

    total_stop = walltime();
    avg_new = (total_stop-total_start)/runs;

    printf("Average time, improved: %.6f\n", avg_new);
    printf("Maximum time, improved: %.6f\n", max_new);
    printf("Minimum time, improved: %.6f\n", min_new);
    
    printf("\nAverage improvement: %.6f\n\n", avg_old/avg_new);

    int count = 0;

    double max_diff = 0;
    for(int i=0; i<2*(L+1); i++){
	for(int j=0; j<(L+1)*2*L; j++)
	    if(fabs(nodes[i][j] - nodes2[i][j]) > 1e-10){
		max_diff = fabs(nodes[i][j] - nodes2[i][j]) > max_diff ?
		    fabs(nodes[i][j] - nodes2[i][j]) : max_diff;
		count++;
	    }
    }

    printf("Count: %d \t max_diff: %.6f\n\n", count, max_diff);

    free(conv_col);
    for(int i=0; i<N*L; i++)
	free(temp_thresholds[i]);
    for(int i=0; i<2*(L+1); i++){
	free(bonds[i]);
	free(nodes[i]);
	free(nodes2[i]);
    }
    free(temp_thresholds);
    free(bonds);
    free(nodes);
    free(nodes2);

}

void fwriteBondThresholds(int size, int run){
    
    struct System A;
    FILE *outfile;
    char *txtfile;
    
    allocSys(&A, size);
    initSys(&A, 0, size+run+81, -1, 1);
    
    txtfile = calloc(60, sizeof(char));		
    sprintf(txtfile, "OutputPS/bond_test/bond_test_L%d_run%d.txt", size, run);
    outfile = fopen(txtfile, "w");
    
    for(int i=size-1; i >= 0; i--){
	for(int j=0; j < size; j++)
	    for(int k=0; k < 2; k++)
		fprintf(outfile, "%.6f ", A.threshold[i*size + j][k]);
	fprintf(outfile, "\n");
    }
    
    fclose(outfile);
    free(txtfile);
    freeSys(&A);    
	
}

void fwritePathscape(struct System *sys, int run){


    /*  Writes pathscape to one file, and the bonds connecting nodes to another */
    FILE *outfile;
    char *txtfile;
    int size = sys->L;
    txtfile = calloc(80, sizeof(char));
		
    sprintf(txtfile, "OutputPS/L%d/pathscape_run%d.txt", size, run);
    //sprintf(txtfile, "OutputPS/superconductingL%d/flowscape_run%d.txt", sys->L, run);
    //sprintf(txtfile, "OutputPS/test_run%d.txt", run);
    outfile = fopen(txtfile, "w");

    for(int i=size; i >= 0; i--){
	for(int j=0; j < size; j++){
	    fprintf(outfile, "%.6f ", sys->pathscape[i*size + j]);
	}
	fprintf(outfile, "\n");
    }

    fclose(outfile);
    free(txtfile);

    txtfile = calloc(80, sizeof(char));

    sprintf(txtfile, "OutputPS/L%d/pathscape_bonds_run%d.txt", size, run);
    //sprintf(txtfile, "OutputPS/superconductingL%d/flowscape_bonds_run%d.txt", sys->L, run);
    //sprintf(txtfile, "OutputPS/test_bonds_run%d.txt", run);
    outfile = fopen(txtfile, "w");
	
    for(int i=size; i >= 0; i--){
	for(int j=0; j < size; j++)
	    fprintf(outfile, "(%d,%d) ", sys->pathscape_bonds[i*size + j][0], sys->pathscape_bonds[i*size + j][1]);
	fprintf(outfile, "\n");
    }

    fclose(outfile);
    free(txtfile);


}

void fwriteBackbone(int size, int run, double theta,
		    int **backbone, int **sc, int **burnclass, double phi){

    /*  Writes backbone and spanning cluster to file  */

    FILE *outfile;
    char *txtfile;
    txtfile = calloc(50, sizeof(char));
    sprintf(txtfile, "OutputPS/L%d/scb_run%d_theta%d.txt", size, run, (int)theta);
    //sprintf(txtfile, "OutputPS/PhiL%d/scb_run%d_theta%.5f_phi%.5f.txt", size, run, theta, phi);
    outfile = fopen(txtfile, "w");
	
	
    fprintf(outfile, "# Cols 1-3 contain coords (x, y) and bond (0 or 1) of bonds belonging to the spanning cluster\n");
    fprintf(outfile, "# Cols 4-6 contain coords (x, y) and bond (0 or 1) of bonds belonging to the backbone\n");
    fprintf(outfile, "# Col 7 contains the burn class of the given bond (if applicable)\n");
    for(int i=0; i < size; i++){
	for(int j=0; j < size; j++){
	    for(int k=0; k<2; k++){
		if(sc[i*size+j][k]){
		    fprintf(outfile, "%d %d %d ", i, j, k);
		    if(backbone[i*size+j][k])
			fprintf(outfile, "%d %d %d", i, j, k);
		    else
			fprintf(outfile, "%d %d %d", -size, -size, size);
		    if(burnclass != NULL)
			fprintf(outfile, " %d", burnclass[i*size+j][k]);
		}
		if(sc[i*size+j][k] || backbone[i*size+j][k])
		    fprintf(outfile, "\n");
	    }
	}
    }


    fclose(outfile);
    free(txtfile);


}

void fwriteBackboneR(int size, int run, double r, double alpha, double beta,
		    int **backbone, int **sc, int **burnclass){

    /*  Writes backbone and spanning cluster to file  */

    FILE *outfile;
    char *txtfile;
    txtfile = calloc(80, sizeof(char));
    //sprintf(txtfile, "OutputPS/L%d/scb_run%d_theta%d.txt", size, run, (int)theta);
    sprintf(txtfile, "OutputPS/PhiL%d/scb_run%d_r%.3f_alpha%.3f_beta%.3f.txt", size, run, r, alpha, beta);
    outfile = fopen(txtfile, "w");
	
	
    fprintf(outfile, "# Cols 1-3 contain coords (x, y) and bond (0 or 1) of bonds belonging to the spanning cluster\n");
    fprintf(outfile, "# Cols 4-6 contain coords (x, y) and bond (0 or 1) of bonds belonging to the backbone\n");
    fprintf(outfile, "# Col 7 contains the burn class of the given bond (if applicable)\n");
    for(int i=0; i < size; i++){
	for(int j=0; j < size; j++){
	    for(int k=0; k<2; k++){
		if(sc[i*size+j][k]){
		    fprintf(outfile, "%d %d %d ", i, j, k);
		    if(backbone[i*size+j][k])
			fprintf(outfile, "%d %d %d", i, j, k);
		    else
			fprintf(outfile, "%d %d %d", -size, -size, size);
		    if(burnclass != NULL)
			fprintf(outfile, " %d", burnclass[i*size+j][k]);
		}
		if(sc[i*size+j][k] || backbone[i*size+j][k])
		    fprintf(outfile, "\n");
	    }
	}
    }


    fclose(outfile);
    free(txtfile);


}

void breakFuseNetwork(struct System *sys, int run, double theta){


    int L = sys->L;
    int N = L*L;
    int *ir = calloc(L, sizeof(ir));
    for(int i=0; i<L; i++)
	ir[i] = i+1;
    ir[L-1] = 0;

    int **burn_class, **old_updated;
    //int num_init_breaks = 2;
    int broken_network = 0;
    int cur_burn_class = 1;
    // burn_class denotes whether a bond is a primary, secondary, tertiary...
    sys->max_voltage = 1;

    burn_class = calloc(N, sizeof(burn_class));
    old_updated = calloc(N, sizeof(old_updated));
    for(int i=0; i<N; i++){
	burn_class[i] = calloc(2, sizeof(burn_class[i]));
	old_updated[i] = calloc(2, sizeof(old_updated[i]));
    }
   
    /* while( num_init_breaks > 1 ){ */

    /* 	double dV, ratio; */
    /* 	sys->max_voltage /= 2; */
    /* 	num_init_breaks = 0; */

    /* 	CG(sys); */

    /* 	for(int j=0; j<L; j++) */
    /* 	    for(int i=0; i<L; i++) */
    /* 		for(int k=0; k<2; k++){ */
    /* 		    if(j < L-1 && k == 0) */
    /* 			dV = fabs(sys->voltage[i+(j+1)*L] - sys->voltage[i+j*L]); */
    /* 		    else if(j < L-1 && k == 1) */
    /* 			dV = fabs(sys->voltage[ir[i] + (j+1)*L] - sys->voltage[i+j*L]); */
    /* 		    else */
    /* 			dV = fabs(sys->max_voltage - sys->voltage[i+j*L]); */
		    
    /* 		    ratio = sys->threshold[i+j*L][k]/dV; */
    /* 		    num_init_breaks += ratio < 1; */
    /* 		} */
    /* } */
    
    //printf("Starting voltage: %.4f\n", sys->max_voltage);
    /* double dV = 0.001; */
    /* sys->max_voltage = dV; */
    /* while( !broken_network ){ */

    /* 	CG(sys); */

    /* 	if(!broken_network){ */
    /* 	    while(!broken_network && updateConductancePlastic23(sys)){ */

    /* 		for(int i=0; i<L*L; i++) */
    /* 		    for(int k=0; k<2; k++) */
    /* 			if(old_updated[i][k] != sys->perm_updated[i][k]){ */
    /* 			    old_updated[i][k] = 1; */
    /* 			    burn_class[i][k] = cur_burn_class; */
    /* 			    cur_burn_class++; */
    /* 			} */
    /* 		CG(sys); */
    /* 		updateDV(sys); */
    /* 		broken_network = HKV(sys->updated, sys->L); */
    /* 	    } */
    /* 	} */
	
    /* 	if(!broken_network){ */
    /* 	    sys->max_voltage += dV; */
    /* 	    cur_burn_class = 1; */
    /* 	} */
    /* } */


    /* while( !broken_network ){ */


    /* 	double dV, ratio; */
    /* 	double min_ratio = DBL_MAX; */
    /* 	int min_ratio_coord = 0, min_ratio_bond = 0; */
	

    /* 	CG(sys); */


    /* 	for(int j=0; j<L; j++) */
    /* 	    for(int i=0; i<L; i++){ */
    /* 		for(int k=0; k<2; k++){ */
    /* 		    if(j < L-1 && k == 0) */
    /* 			dV = fabs(sys->voltage[i+(j+1)*L] - sys->voltage[i+j*L]); */
    /* 		    else if(j < L-1 && k == 1) */
    /* 			dV = fabs(sys->voltage[ir[i] + (j+1)*L] - sys->voltage[i+j*L]); */
    /* 		    else */
    /* 			dV = fabs(sys->max_voltage - sys->voltage[i+j*L]); */
		    
    /* 		    ratio = sys->threshold[i+j*L][k]/dV; */
		    
    /* 		    if(ratio < min_ratio && !sys->updated[i+j*L][k]){ */
    /* 			min_ratio = ratio; */
    /* 			min_ratio_coord = i+j*L; */
    /* 			min_ratio_bond = k; */
    /* 		    } */
    /* 		} */
    /* 	    } */
	
    /* 	sys->conductance[min_ratio_coord][min_ratio_bond] = 0; */
    /* 	sys->updated[min_ratio_coord][min_ratio_bond] = 1; */
    /* 	sys->perm_updated[min_ratio_coord][min_ratio_bond] = 1; */
	
    /* 	if(min_ratio < 1){ */
    /* 	    //printf("Burn class: %d\n", cur_burn_class); */
    /* 	    burn_class[min_ratio_coord][min_ratio_bond] = cur_burn_class; */
    /* 	    cur_burn_class++; */
    /* 	} */
    /* 	else{ */
    /* 	    burn_class[min_ratio_coord][min_ratio_bond] = 1; */
    /* 	    cur_burn_class = 2; */
    /* 	} */
	
    /* 	sys->num_updated_bonds++; */
    /* 	broken_network = HKV(sys->updated, sys->L); */
	
    /* } */


    int **backbone = getBackbone(sys->updated, L);
    int **sc = getFracture(sys->updated, L);

    //fwriteBackbone(L, run, theta, backbone, sc, burn_class);


    /* FILE *outfile; */
    /* char *txtfile = calloc(50, sizeof(char)); */
    /* //sprintf(txtfile, "OutputPS/L%d/scb_run%d_theta%d.txt", size, run, (int)b); */
    /* sprintf(txtfile, "OutputPS/L%d/burn_test_run%d.txt", L, run); */
    /* outfile = fopen(txtfile, "w"); */
	
    /* fprintf(outfile, "# Cols 1-3 contain coords (x, y) and bond (0 or 1), col 4 contains the burn class\n"); */
    /* for(int i=0; i < L; i++){ */
    /* 	for(int j=0; j < L; j++){ */
    /* 	    for(int k=0; k<2; k++){ */
    /* 		if(backbone[i*L+j][k]){ */
    /* 		    fprintf(outfile, "%d %d %d %d", i, j, k, burn_class[i*L+j][k]); */
    /* 		    /\* if(burn_class[i*L+j][k] == 0) *\/ */
    /* 		    /\* 	printf("Bah!\n"); *\/ */
    /* 		} */
    /* 		if(backbone[i*L+j][k]) */
    /* 		    fprintf(outfile, "\n"); */
    /* 	    } */
    /* 	} */
    /* } */
   
    /* fclose(outfile); */
    /* free(txtfile); */


    for(int i=0; i<N; i++){
	free(burn_class[i]);
	free(old_updated[i]);
	free(backbone[i]);
	free(sc[i]);
    }
    free(burn_class);
    free(old_updated);
    free(backbone);
    free(sc);
    free(ir);
}

    


void breakFuseFast(struct System *sys){

    int broken_network = 0;
    int L = sys->L;
    double dV, t, ratio, min_ratio;
    int min_coord, min_bond;
    

    int *ir = calloc(L, sizeof(ir));
    for(int i=0; i<L; i++)
	ir[i] = i+1;
    ir[L-1] = 0;

    sys->max_voltage = 1.0;
    
    while( !broken_network ){
	
	CG(sys);
	
	dV = 0;
	t = 0;
	ratio = 0;
	min_ratio = DBL_MAX;
	min_coord = 0;
	min_bond = 0;
	
	for(int j=0; j<L; j++)
	    for(int i=0; i<L; i++)
		for(int k=0; k<2; k++){
		    if(j < L-1 && k == 0)
			dV = fabs(sys->voltage[i+(j+1)*L] - sys->voltage[i+j*L]);
		    else if(j < L-1 && k == 1)
			dV = fabs(sys->voltage[ir[i] + (j+1)*L] - sys->voltage[i+j*L]);
		    else
			dV = fabs(sys->max_voltage - sys->voltage[i+j*L]);
		    
		    t = sys->threshold[i+j*L][k];
		    ratio = t/dV;
		    if(!sys->updated[j*L+i][k] && ratio < min_ratio){
			min_coord = j*L+i;
			min_bond  = k;
			min_ratio = ratio;
		    }
		}
	
	
	sys->conductance[min_coord][min_bond] = 0;
	sys->updated[min_coord][min_bond] = 1;
	sys->perm_updated[min_coord][min_bond] = 1;
	sys->num_updated_bonds++;
	
	updateDV(sys);
	broken_network = HKV(sys->updated, L);
    }
       
    
    breakBonds(sys);
    free(ir);
    
}


void breakNetwork(struct System *sys, int run){


    
    double dV = 0.001;
    sys->max_voltage = dV;
    double theta = round(atan(sys->b)*180/M_PI);
    int broken_network = (theta > 50);
    int breaklimit = 4*sys->L*sys->L;
    int updated_count = 0;
    //broken_network = (b > 50); // Ensures we're not moving into the superconducting regime

    if((int)theta == -90)
	breakFuseNetwork(sys, run, theta);
    
    else if(theta < 0)
	while( !broken_network ){

	    CG(sys);
	    updated_count = 0;

	    if(!broken_network){
		while(!broken_network && (updateConductancePlastic23(sys) || updateConductanceSingleElastic(sys))){
		    CG(sys);		
		    updateDV(sys);
		    broken_network = HKV(sys->updated, sys->L);

		    updated_count++;
		
		    if(updated_count > breaklimit){
			printf("Breaklimit! Max voltage = %.4f\n", sys->max_voltage);
			break;
		    }
		}
	    }

	    if(!broken_network)
		sys->max_voltage += dV;
	}

    else
	while( !broken_network ){

	    CG(sys);
	    updated_count = 0;

	    if(!broken_network){
		while(!broken_network && updateConductanceSingleElastic(sys)){
		    CG(sys);
		    updateDV(sys);
		    broken_network = HKV(sys->updated, sys->L);

		    updated_count++;
		
		    if(updated_count > breaklimit){
			printf("Breaklimit! Max voltage = %.4f\n", sys->max_voltage);
			break;
		    }
		}
	    }

	    if(!broken_network)
		sys->max_voltage += dV;
	}
    

    breakBonds(sys);
}

void benchmark(int nthreads){


    int thread_runs;
    int thread_list[4] = {1, 2, 4, 6};
    if(nthreads == 6)
	thread_runs = 4;
    else if(nthreads == 4)
	thread_runs = 3;
    else
	thread_runs = nthreads;
    


    double **start = calloc(thread_runs, sizeof(start));
    double **stop = calloc(thread_runs, sizeof(stop));
    double **t_time = calloc(thread_runs, sizeof(t_time));
    for(int i=0; i<thread_runs; i++){
    	start[i] = calloc(11, sizeof(start[i]));
	stop[i] = calloc(11, sizeof(stop[i]));
	t_time[i] = calloc(2, sizeof(t_time[i]));
    }

    int mruns = 4;
    int size;
    double theta;

    //Optimal pathscape, 1T, L = 32
    //Optimal pathscape, 1T, L = 64
    //Optimal pathscape, 2T, L = 32
    //Optimal pathscape, 2T, L = 64

    //Fuse Network, L = 32
    //Fuse Network, L = 64
    //Fuse Network, L = 128
    //Plastic network, L = 32
    //Plastic network, L = 64
    //Plasticelastic network, L = 32
    //Plasticelastic network, L = 64

    int par_threads, sys_threads;
    int size_array[11] = {32, 64, 32, 64, 32, 64, 128, 32, 64, 32, 64};
    double theta_array[11] = {0, 0, 0, 0, -90, -90, -90, 0, 0, 15, 15};
    int threads = 1;
    printf("Running benchmark...\n");
    for(int ind=0; ind < thread_runs; ind++){
	threads = thread_list[ind];
	printf("\nRunning with %d thread(s)\n", threads);
	t_time[ind][0] = walltime();


	par_threads = threads;
	sys_threads = 1;
	omp_set_dynamic(0);
	omp_set_nested(1);
#pragma omp parallel for num_threads(par_threads) schedule(dynamic)	\
    private(size, theta, par_threads, sys_threads)
	for(int type=0; type<11; type++){
	    
	    printf("    Running type %d\n", type);
	    size = size_array[type];
	    theta = theta_array[type];
	    par_threads = threads;
	    sys_threads = 1;
	    if(type == 2 || type == 3){
		par_threads = threads/2;
		sys_threads = 2;
	    }

	    start[ind][type] = walltime();
	    struct System A;
	    allocSys(&A, size);
	    for(int run=0; run<mruns; run++){

		initSys(&A, theta, size+run+81, -1, sys_threads);
		if(type < 4)
		    optimalPathscape(&A);
		else
		    breakNetwork(&A, run);
	    }	    
	    freeSys(&A);	    		
	    stop[ind][type] = walltime();

	}
	t_time[ind][1] = walltime();
    }
    
    printf("Finished benchmark, printing results to file\n");

    FILE *outfile;
    char *txtfile = calloc(50, sizeof(txtfile));
    sprintf(txtfile, "benchmark_%d_thr.txt", nthreads);
    outfile = fopen(txtfile, "w");
    fprintf(outfile, "Benchmark for %d threads, max runs = %d. Total time displayed, with average in parenthesis\n\n\n", nthreads, mruns);
    for(int ind=0; ind < thread_runs; ind++){
	threads = thread_list[ind];
	fprintf(outfile, "Threads: %d\n\n", threads);
	fprintf(outfile, "Pathscape,     1T,     L=32:    %.4f  ( %.4f )\n",
		stop[ind][0]-start[ind][0], (stop[ind][0]-start[ind][0])/mruns);
	fprintf(outfile, "Pathscape,     1T,     L=64:    %.4f  ( %.4f )\n",
		stop[ind][1]-start[ind][1], (stop[ind][1]-start[ind][1])/mruns);
	fprintf(outfile, "Pathscape,     2T,     L=32:    %.4f  ( %.4f )\n",
		stop[ind][2]-start[ind][2], (stop[ind][2]-start[ind][2])/mruns);
	fprintf(outfile, "Pathscape,     2T,     L=64:    %.4f  ( %.4f )\n",
		stop[ind][3]-start[ind][3], (stop[ind][3]-start[ind][3])/mruns);
	fprintf(outfile, "Fuse network,          L=32:    %.4f  ( %.4f )\n",
		stop[ind][4]-start[ind][4], (stop[ind][4]-start[ind][4])/mruns);
	fprintf(outfile, "Fuse network,          L=64:    %.4f  ( %.4f )\n",
		stop[ind][5]-start[ind][5], (stop[ind][5]-start[ind][5])/mruns);
	fprintf(outfile, "Fuse network,         L=128:    %.4f  ( %.4f )\n",
		stop[ind][6]-start[ind][6], (stop[ind][6]-start[ind][6])/mruns);
	fprintf(outfile, "Plastic network,       L=32:    %.4f  ( %.4f )\n",
		stop[ind][7]-start[ind][7], (stop[ind][7]-start[ind][7])/mruns);
	fprintf(outfile, "Plastic network,       L=64:    %.4f  ( %.4f )\n",
		stop[ind][8]-start[ind][8], (stop[ind][8]-start[ind][8])/mruns);
	fprintf(outfile, "Plastoelastic network, L=32:    %.4f  ( %.4f )\n",
		stop[ind][9]-start[ind][9], (stop[ind][9]-start[ind][9])/mruns);
	fprintf(outfile, "Plastoelastic network, L=64:    %.4f  ( %.4f )\n",
		stop[ind][10]-start[ind][10], (stop[ind][10]-start[ind][10])/mruns);
	fprintf(outfile, "\nTotal time used:             %.4f\n", t_time[ind][1]-t_time[ind][0]);
	fprintf(outfile, "Average time used per thread:  %.4f\n\n", (t_time[ind][1]-t_time[ind][0])/threads);
	for(int i=0; i<40; i++)
	    fprintf(outfile, "*");
	fprintf(outfile, "\n\n");
	ind++;
    }
    fclose(outfile);

    for(int i=0; i<nthreads; i++){
	free(start[i]);
	free(stop[i]);
	free(t_time[i]);
    }
    free(start);
    free(stop);
    free(t_time);
    free(txtfile);

}

double vtRatio(struct System *sys){

    int L = sys->L;
    double dV, t;
    double max_ratio = 0;
    
    int *ir = (int*)calloc(L, sizeof(int));
    for(int i=0; i<L; i++)
	ir[i] = i+1;
    ir[L-1] = 0;

    for(int j=0; j<L; j++)
	for(int i=0; i<L; i++)
	    for(int k=0; k<2; k++){
		if(j < L-1 && k == 0)
		    dV = fabs(sys->voltage[i+(j+1)*L] - sys->voltage[i+j*L]);
		else if(j < L-1 && k == 1)
		    dV = fabs(sys->voltage[ir[i] + (j+1)*L] - sys->voltage[i+j*L]);
		else
		    dV = fabs(sys->max_voltage - sys->voltage[i+j*L]);
		
		t = sys->threshold[i+j*L][k];
		max_ratio = dV/t > max_ratio ? dV/t : max_ratio;
	    }

    free(ir);
    return max_ratio;
}
double vtRatioAvg(struct System *sys){

    int L = sys->L;
    double dV, t;
    int count = 0;
    double ratio = 0;

    int *ir = (int*)calloc(L, sizeof(int));
    for(int i=0; i<L; i++)
	ir[i] = i+1;
    ir[L-1] = 0;
  
    for(int j=0; j<L; j++)
	for(int i=0; i<L; i++)
	    for(int k=0; k<2; k++){
		if(j < L-1 && k == 0)
		    dV = fabs(sys->voltage[i+(j+1)*L] - sys->voltage[i+j*L]);
		else if(j < L-1 && k == 1)
		    dV = fabs(sys->voltage[ir[i] + (j+1)*L] - sys->voltage[i+j*L]);
		else
		    dV = fabs(sys->max_voltage - sys->voltage[i+j*L]);
		
		t = sys->threshold[i+j*L][k];
		ratio += dV/t;
		count++;
	    }

    free(ir);
    return ratio/count;
}

void getVTRatio(struct System *sys, double **vt){

    // vt HAS TO BE OF SAME SIZE AS sys->threshold
    int L = sys->L;
    double dV, t;

    int *ir = (int*)calloc(L, sizeof(int));
    for(int i=0; i<L; i++)
	ir[i] = i+1;
    ir[L-1] = 0;

    for(int j=0; j<L; j++)
	for(int i=0; i<L; i++)
	    for(int k=0; k<2; k++){
		if(j < L-1 && k == 0)
		    dV = fabs(sys->voltage[i+(j+1)*L] - sys->voltage[i+j*L]);
		else if(j < L-1 && k == 1)
		    dV = fabs(sys->voltage[ir[i] + (j+1)*L] - sys->voltage[i+j*L]);
		else
		    dV = fabs(sys->max_voltage - sys->voltage[i+j*L]);
		
		t = sys->threshold[i+j*L][k];
		vt[i+j*L][k] = dV/t;
	    }
    
    free(ir);        
}
